:- module(bundle_config__ciaode, [], [bundleconfig]).

:- doc(title, "Configuration rules for CiaoDE").
:- doc(author, "Ciao Development Team").

:- use_module(library(bundletool/paths_extra), [fsR/2]).

% ---------------------------------------------------------------------------

:- doc(section, "General Build/Installation Options").

% (a parameter)
:- bundle_param('INTERACTIVE_CONFIG', [
    valid_values([false, true]),
    %
    hidden,
    noprevious,
    rule_default(false),
    %
    interactive([extended]),
    comment("Interactive configuration")
]).

:- bundle_param('INTERACTIVE_LEVEL', [
    valid_values(['1', '2', '3']),
    %
    hidden,
    noprevious,
    rule_default('1'),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Entering the interactive configuration.\n"||
      "You will now be asked some questions related to the configuration.\n"||
      "Hit [Enter] to accept the default values shown in square brackets.\n\n"||
      %
      "Please select level of interaction:\n\n"||
      "    1 --  Fully automatic (recommended).\n"||
      "    2 --  Manually configure just a minimum set of options.\n"||
      "    3 --  Manually configure an extended set of options."),
    comment("Interactive configuration level")
]).

:- bundle_param('STOP_IF_ERROR', [
    valid_values(['yes', 'no']),
    %
    rule_default('yes'),
    %
    interactive([extended],
      % .....................................................................
      "Stop installation if an error is found."),
    comment("Stop on build/installation errors")
]).

:- bundle_param('WITH_DOCS', [
    valid_values(['yes', 'no']),
    %
    rule_default('yes'),
    %
    interactive([minumum, extended],
      % .....................................................................
      "Generate documentation."),
    comment("Generate documentation")
]).

% ---------------------------------------------------------------------------

:- doc(section, "Installation Type, Directories, and Permissions").

:- bundle_param('INSTYPE', [
    valid_values(['global', 'local']),
    %
    rule_default('global'),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Select the type of installation:\n\n"||
      "    global -- Install the system in a separate location from the\n"||
      "              sources and set up things to use the installed version.\n"||
      "              The system will not require the sources to run, and \n"||
      "              they can be erased after installation.\n"||
      "    local  -- The system will be compiled in, and run from the \n"||
      "              sources (this is specially useful for developers)."),
    comment("Installation type")
]).

% TODO: make use of default value (simplify scripts)
% TODO: rename 'all' and 'user'? (e.g. global and local too?)
% TODO: include here other access methods like menus, and desktop icons 

:- bundle_param('REGISTRATION_TYPE', [
    valid_values(['all', 'user']),
    %
    rule_default(SysregType, (
      param('INSTYPE'(InsType)),
      def_registration_type(InsType, SysregType))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Registration type:\n\n"||
      "    all  --  Make the system available to all users. Typically you\n"||
      "             you will need to complete the installation as root.\n"||
      "    user --  Make the system available only for the current user\n"||
      "             (configure it in the user\'s home directory)."),
    comment("Registration type")
]).

def_registration_type(global, all).
def_registration_type(local, user).

% (Not configurable setting, necessary for CONFIG_input, input of config-and-dump.sh)
:- bundle_param('CIAODESRC', [
    rule_set_value(Value, fsR(bundle_src(ciaode), Value)),
    comment("Source directory for ciaode")
]).

% ---------------------------------------------------------------------------

:- doc(section, "Paths for installation").
% (binaries, libraries, documentation, etc.)

:- use_module(library(system_extra), [get_home/1]).

% TODO: (JFMC) A 'builddir' and a 'installdir' are very similar; are
%   they the same kind of objects? Of course, a 'installdir' must map
%   items to files in a system-specific way. Another idea is that each
%   installdir has a parent builddir.

:- bundle_param('INSTALLDIR_PREFIX', [
    rule_set_value(Value, (
      param('INSTYPE'(InsType)),
      InsType == 'local', build_dir(Value))),
    rule_default(DefValue, (
      param('INSTYPE'(InsType)),
      param('REGISTRATION_TYPE'(SysregType)),
      get_prefix(SysregType, InsType, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Specify the directory to perform the installation."),
    comment("Install prefix")
]).

build_dir := ~fsR(builddir(build)).

get_prefix(all,  global, '/usr/local').
get_prefix(user, global, ~get_home). % TODO: not valid combination!
get_prefix(all,  local, ~fsR(bundle_src(ciaode))). % TODO: unused?
get_prefix(user, local, ~fsR(bundle_src(ciaode))). % TODO: unused?

:- bundle_param('INSTALLDIR_BIN', [
    rule_set_value(Value, (
      param('INSTALLDIR_PREFIX'(Prefix)),
      fsR(Prefix/'bin', Value))),
    comment("Installation directory for executables")
]).

:- bundle_param('INSTALLDIR_LIBROOT', [
    rule_set_value(Value, (
      param('INSTALLDIR_PREFIX'(Prefix)),
      param('INSTYPE'(InsType)),
      get_final_lib_root(InsType, Prefix, Value))),
    comment("Installation directory for libraries")
]).

% TODO: strange! for 'global' it is a prefix, for 'local' it is not
get_final_lib_root(local, _Prefix) := ~fsR(bundle_src(ciao)). % TODO: use also Prefix; distinguish dir for source and binaries?
get_final_lib_root(global, Prefix) := ~fsR(Prefix/'lib').

% (Not configurable setting, necessary for CONFIG_input, input of config-and-dump.sh)
:- bundle_param('DEFAULTLIBDIR', [
    rule_set_value(Value, (
      param('INSTALLDIR_LIBROOT'(_)), % TODO: hidden dependency (due to fsR)
      get_installdir_lib(ciao, Value))),
    comment("Default directory for Ciao libraries")
]).

:- use_module(library(bundletool/config_common), [get_installdir_lib/2]).

% (also needed by config-and-dump.sh)
:- bundle_param('EXECMODE', [
    rule_default('775'),
    %
    interactive([extended]),
    comment("Permissions for installed execs/dirs")
]).

% (also needed by config-and-dump.sh)
:- bundle_param('DATAMODE', [
    rule_default('664'),
    %
    interactive([extended]),
    comment("Permissions for installed data files")
]).

% (also needed by config-and-dump.sh)
% TODO: Ignored by many of the installation code (thus, not working)
:- bundle_param('INSTALLGROUP', [
    rule_default(''),
    %
    interactive([extended],
      % .....................................................................
      "Group for the installed files (empty means use default)"),
    comment("Custom group for installed files")
]).

% ---------------------------------------------------------------------------

:- doc(section, "Optional external commands").
% Some custom non-configurable settings.

:- use_module(library(system_extra), [find_executable/3]).
:- use_module(library(lists), [append/3]).
:- use_module(library(messages), [error_message/1]).

:- discontiguous cmd_op/2.

% Locate Cmd or any of its variants (defined in
% @pred{cmd_op/2}). @var{CmdDir} and @var{CmdName} are the found path
% and command name, respectively.
get_cmd_op(Cmd, CmdDir, CmdName) :-
	( cmd_op(Cmd, CmdName),
	  find_executable(CmdName, Path, _) ->
	    atom_concat(Path, '/', CmdDir)
	; error_message("Cannot find any version of '" ||
		        (~append(~atom_codes(Cmd), "' command in the path."))),
	  fail
	).

% Detect GNU make
:- bundle_param('MAKEDIR', [
    rule_set_value(Value, get_cmd_op(gmake, Value, _)),
    comment("Directory of 'make' utility")
]).
:- bundle_param('MAKENAME', [
    rule_set_value(Value, get_cmd_op(gmake, _, Value)),
    comment("Name of 'make' utility")
]).

cmd_op(gmake) := gmake|make.

% Detect GNU tar
:- bundle_param('TARDIR', [
    rule_set_value(Value, get_cmd_op(gtar, Value, _)),
    comment("Directory of 'tar' utility")
]).
:- bundle_param('TARNAME', [
    rule_set_value(Value, get_cmd_op(gtar, _, Value)),
    comment("Name of 'tar' utility")
]).

cmd_op(gtar) := gnutar|gtar|tar.


