:- module(bundle_config__ciao, [], [bundleconfig]).

:- doc(title, "Configuration rules for Ciao").
:- doc(author, "Ciao Development Team").

:- use_module(library(system), [file_exists/1]).
:- use_module(library(process), [process_call/3]).
:- use_module(library(system_extra), [find_executable/2]).
:- use_module(library(system_extra), [get_home/1]).
:- use_module(library(lists), [append/3]).

:- use_module(library(bundletool/paths_extra), [fsR/2]).
:- use_module(library(bundletool/detcheader), [detect_c_headers/1]).

% ===========================================================================

:- doc(section, "Command-line Enviroment").

:- bundle_param('UPDATE_BASHRC', [
    valid_values(['yes', 'no']),
    %
    rule_default(DefValue, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      get_update_sh(SysregType, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Set to \"no\" if you do not wish to configure bash to work with Ciao or \n"||
      "if you wish to configure it by hand."),
    comment("Update bash init file")
]).

% ---------------------------------------------------------------------------

:- bundle_param('DOTBASHRC', [
    needed_if(param('UPDATE_BASHRC'(yes))),
    rule_default(DefValue, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      get_bashrc(SysregType, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "The bash initialization file where the Ciao variables are set."),
    comment("Bash initialization file")
]).

get_bashrc(all, F) :-
	( member(F, ['/etc/bash.bashrc', '/etc/bashrc']),
	    file_exists(F) ->
	    true
	; F = '/etc/bashrc'
	).
get_bashrc(user) := ~fsR(~get_home/'.bashrc').

% ---------------------------------------------------------------------------

:- bundle_param('UPDATE_CSHRC', [
    valid_values(['yes', 'no']),
    %
    rule_default(DefValue, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      get_update_sh(SysregType, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Set to \"no\" if you do not wish to configure csh/tcsh to work with\n"||
      "Ciao or if you wish to configure it by hand."),
    comment("Update csh init file")
]).

% ---------------------------------------------------------------------------

:- bundle_param('DOTCSHRC', [
    needed_if(param('UPDATE_CSHRC'(yes))),
    rule_default(DefValue, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      get_cshrc(SysregType, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "The csh/tcsh initialization file where the Ciao variables are set.\n"||
      "Note that on some systems tcsh reads \"~/.tcshrc\"."),
    comment("Csh/Tcsh initialization file")
]).

% by default, assume /etc/csh.cshrc
get_cshrc(all, F) :-
	( member(F, ['/etc/csh.cshrc', '/etc/tcsh.tcshrc']),
	    file_exists(F) ->
	    true
	; F = '/etc/csh.cshrc'
	).
get_cshrc(user) := ~get_cshrc_name.

% by default, assume .cshrc
get_cshrc_name(C) :-
	( ( member(F, ['.tcshrc', '.cshrc']),
	    C = ~fsR((~get_home)/F),
	    file_exists(C)
	  ) ->
	    true
	; F = '.cshrc',
	  C = ~fsR((~get_home)/F)
	).

get_update_sh('all',  'no').
get_update_sh('user', 'yes').

% ===========================================================================

:- doc(section, "Options for Ciao/Toplevel").

:- bundle_param('GEN_CIAO_ASR', [
    valid_values([yes, no]),
    %
    rule_default(yes),
    comment("Generate .asr files during Ciao compilation")
]).

:- bundle_param('INSTALL_PROLOG_NAME', [
    valid_values(['yes', 'no']),
    %
    rule_default('yes'),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Set to \"yes\" if you wish to create a link that brings up Ciao \n"||
      "when you type \"prolog\". You may want to say no if there are other\n"||
      "systems that support the Prolog language in your machine and you\n"||
      "do not want to make Ciao the default."),
    comment("Symbolic link from Ciao to 'prolog' executable")
]).

% ===========================================================================

:- doc(section, "Options for Emacs Mode").

:- bundle_param('WITH_EMACS_MODE', [
    valid_values(['yes', 'no']),
    %
    default_comment("Emacs available"),
    rule_default(VerifyEmacs, verify_emacs(VerifyEmacs)),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Set to \"yes\" if you wish to install the Ciao emacs libraries which\n"||
      "implement the Emacs-based IDE (integrated development environment)\n"||
      "(highly recommended).  It should be set to no if emacs is not\n"||
      "installed in the system.  It is safe to leave as \"yes\" otherwise."),
    comment("Install Emacs-based IDE")
]).

verify_emacs(Value) :-
	( emacs_installed -> Value = yes ; Value = no ).

emacs_installed :- find_emacs(_).

% TODO: it should consider auto_install option!
find_emacs(File) :- find_executable('emacs', File).

% ---------------------------------------------------------------------------

:- bundle_param('WITH_XEMACS_MODE', [
    valid_values(['yes', 'no']),
    %
    default_comment("XEmacs available"),
    rule_default(VerifyXEmacs, verify_xemacs(VerifyXEmacs)),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Set to \"yes\" if you wish to install the Ciao emacs libraries which\n"||
      "implement the XEmacs based IDE (integrated development environment).\n"||
      "It should be set to no if xemacs is not installed in the system. It\n"||
      "is safe to leave as \"yes\" otherwise."),
    comment("Install XEmacs based IDE")
]).

verify_xemacs(Value) :-
	( xemacs_installed -> Value = yes ; Value = no ).

xemacs_installed :- find_xemacs(_).

% TODO: it should consider auto_install option!
find_xemacs(File) :- find_executable('xemacs', File).

% ---------------------------------------------------------------------------

:- bundle_param('EMACS_FOR_CIAO', [
    needed_if(param('WITH_EMACS_MODE'(yes))),
    rule_default(DefValue, find_emacs(DefValue)),
    %
    interactive([minimum, extended],
      % .....................................................................
      "The version of emacs that you wish to use with Ciao. The development\n"||
      "environment will be compiled for use with this version."),
    comment("Emacs version to be used")
]).

% ---------------------------------------------------------------------------

% TODO: Change name
:- bundle_param('UPDATE_DOTEMACS', [
    valid_values(['yes', 'no']),
    %
    needed_if(param('WITH_EMACS_MODE'(yes))),
    % TODO: This set value means that no question is really asked
    rule_set_value(Value, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      update_dotemacs_(SysregType, Value))),
    rule_default('yes'),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Set to \"yes\" if you wish to configure emacs to work with Ciao\n"||
      "(modify emacs initialization file)."),
    comment("Modify emacs init file")
]).

% update_dotemacs_(InsType, VerifyEmacs, UpdateEmacs)

update_dotemacs_(all,  no) :- !.
update_dotemacs_(user, yes).

% ---------------------------------------------------------------------------

:- bundle_param('DOTEMACS', [
    needed_if(param('UPDATE_DOTEMACS'(yes))),
    rule_default(DefValue, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      get_dotemacs(SysregType, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Define the emacs initialization file where the Ciao settings will be\n"||
      "added."),
    comment("Emacs initialization file")
]).

get_dotemacs(user) := ~fsR(~get_home/'.emacs').

% ---------------------------------------------------------------------------

% TODO: Do not customize this
:- bundle_param('EMACSINITFILE', [
    needed_if(param('WITH_EMACS_MODE'(yes))),
    rule_set_value(Value, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      get_emacs_init_file(SysregType, Value))),
    interactive([], % TODO: configurable?
      % .....................................................................
      "Specify the name of the emacs lisp file defining the Ciao mode."),
    comment("Emacs init file")
]).
%% :- bundle_param('VERIFY_XEMACS', [
%%     rule_set_value(VerifyXEmacs, verify_xemacs(VerifyXEmacs))
%% ]).

get_emacs_init_file(all, '65ciao-mode-init.el') :-
	% Use if it is debian based
	get_os('LINUX'),
	file_exists('/etc/debian_version'),
	!.
get_emacs_init_file(_, 'ciao-mode-init.el').

% ---------------------------------------------------------------------------

:- bundle_param('XEMACS_FOR_CIAO', [
    needed_if(param('WITH_XEMACS_MODE'(yes))),
    rule_default(DefValue, find_xemacs(DefValue)),
    %
    interactive([minimum, extended],
      % .....................................................................
      "The version of xemacs that you wish to use with Ciao. The development\n"||
      "environment will be compiled for use with this version."),
    comment("XEmacs version to be used")
]).

% ---------------------------------------------------------------------------

% TODO: Change name
:- bundle_param('UPDATE_DOTXEMACS', [
    valid_values(['yes', 'no']),
    %
    needed_if(param('WITH_XEMACS_MODE'(yes))),
    % , ('VERIFY_XEMACS', yes)]),
    % This set value means that no question is really asked
    rule_set_value(Value, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      update_dotxemacs_(SysregType, Value))),
    rule_default('yes'),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Set to \"yes\" if you wish to configure XEmacs to work with Ciao\n"||
      "(modify xemacs initialization file)."),
    comment("Modify xemacs init file")
]).

update_dotxemacs_(all,  no) :- !.
update_dotxemacs_(user, yes) :-
	file_exists('~/.xemacs'),
	!.
update_dotxemacs_(user, no).

% ---------------------------------------------------------------------------

:- bundle_param('DOTXEMACS', [
    needed_if(param('UPDATE_DOTXEMACS'(yes))),
    rule_default(DefValue, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      get_dotxemacs(SysregType, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Define the xemacs initialization file where the Ciao settings will be\n"||
      "added."),
    comment("XEmacs initialization file")
]).

get_dotxemacs(user) := ~fsR(~get_home/'.xemacs'/'init.el').

% ---------------------------------------------------------------------------

:- bundle_param('EMACS_SITE_START', [
    needed_if(param('WITH_EMACS_MODE'(yes))),
    rule_default(Value, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      param(ciaode:'INSTYPE'(InsType)),
      get_emacs_site_start(emacs, SysregType, InsType, Value))),
    %
    interactive([extended],
      % .....................................................................
      "Specify in what file/directory you want to insert/copy the Ciao Emacs\n"||
      "Mode initialization code."),
    comment("Emacs site start")
]).
:- bundle_param('XEMACS_SITE_START', [
    needed_if(param('WITH_XEMACS_MODE'(yes))),
    rule_default(Value, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      param(ciaode:'INSTYPE'(InsType)),
      get_emacs_site_start(xemacs, SysregType, InsType, Value))),
    %
    interactive([extended],
      % .....................................................................
      "Specify in what file/directory you want to insert/copy the Ciao Emacs\n"||
      "Mode initialization code (XEmacs)."),
    comment("XEmacs site start")
]).

:- use_module(library(bundletool/config_common), [command_options/1]).
:- use_module(library(bundletool/config_common), [get_installdir_lib/2]).
:- use_module(library(system), [winpath/2]).

get_emacs_site_start(EmacsType, 'all', 'global', Value) :-
	emacs_site_start_(EmacsType, Value),
	!.
get_emacs_site_start(_, _, _, Value) :-
	Value0 = ~get_installdir_lib(ciao),
	winpath(Value, Value0). % (translate in Unix format, if necessary)

emacs_site_start_(EmacsType, SiteStart) :-
	possible_emacs_site_start(EmacsType, SiteStart),
	file_exists(SiteStart),
	!.

% Note: this returns files or directories, as follows:
%  .../site-lisp/site-start.d  ==> must put a script in that directory
%  .../site-lisp               ==> must use/create a site-start.el
possible_emacs_site_start(emacs) :=
	'/Applications/Emacs.app/Contents/Resources/site-lisp/site-start.d'|
	'/Applications/Emacs.app/Contents/Resources/site-lisp'|
	'/usr/share/emacs/site-lisp/site-start.d'|
	'/usr/share/emacs/site-lisp'|
	'/etc/emacs/site-start.d'.
possible_emacs_site_start(xemacs) :=
	'/usr/share/xemacs/site-packages/lisp/site-start.d'.

% ---------------------------------------------------------------------------
% TODO: Depends on LPdoc options!

:- doc(section, "Pillow Options").

:- bundle_param('WEB_IMAGES_PATH', [
    rule_default(DefValue, (
      param(lpdoc:'HTMLDIR'(HtmlDir)),
      get_web_images_path(HtmlDir, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "For the PiLLoW Web programming library, define the directory\n"||
      "(accessible via WWW) where the icons which come with PiLLoW\n"||
      "(in library/pillow/images) will go."),
    comment("Pillow images will go here")
]).

% TODO: trailing /?
% TODO: use 'pillow/images'? 'images' alone is a very common name
get_web_images_path(HtmlDir) := ~fsR(HtmlDir/images).

% ---------------------------------------------------------------------------

:- bundle_param('WEB_IMAGES_URL', [
    rule_default(DefValue, (
      param(lpdoc:'HTMLURL'(HtmlUrl)),
      get_web_images_url(HtmlUrl, DefValue))),
    %
    interactive([minimum, extended]),
    comment("URL for Pillow images")
]).

get_web_images_url(HtmlUrl) := ~fsR(HtmlUrl/'images').

% ---------------------------------------------------------------------------

:- doc(section, "Options for MySQL bindings").

:- bundle_param('WITH_MYSQL', [
    valid_values(['yes', 'no']),
    %
    default_comment("MySQL available"),
    default_value_comment(no,
        "MySQL has not been detected.  If you would like to use the\n"||
        "Ciao-MySQL interface it is highly recommended that you stop\n"||
        "the Ciao configuration now and install MySQL first."),
    rule_default(WithMySQL, verify_mysql(WithMySQL)),
    %
    interactive([extended],
      % .....................................................................
      "Set to \"yes\" if you wish to interface with the MySQL database.\n"||
      "If you choose to have the MySQL interface, you should have the MySQL\n"||
      "client part installed in the machine where you are compiling and using\n"||
      "it.  The MySQL daemon should also be up and running when using the\n"||
      "interface."),
    comment("Install MySQL support")
]).

verify_mysql(Value) :-
	( mysql_installed -> Value = yes ; Value = no ).

mysql_installed :-
	detect_c_headers(['mysql/mysql.h']).

% ---------------------------------------------------------------------------

:- use_module(library(messages), [warning_message/1]).

:- bundle_param('MYSQL_CLIENT_DIRECTORY', [
    needed_if(param('WITH_MYSQL'('yes'))),
    rule_default(MySQLDir, get_mysql_dir(MySQLDir)),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Directory where the MySQL client library is installed."),
    comment("MySQL client library path")
]).

% TODO: This is not a good way of detecting MySQL
get_mysql_dir(MySQL) :-
	locate_file('libmysqlclient.a', MySQL),
	!.
get_mysql_dir('/usr/lib/mysql') :-
	warning_message(
	    "Unable to determine where the MySQL client library is " ||
	    "installed.\nCurrent value (/usr/lib/mysql) is only a guess.").

% TODO: Use mysql_config
% TODO: REMOVE! 'locate' IS NOT available in many platforms (e.g., Mac OS X)
locate_file(FileName, FileDir) :-
	process_call(path(locate), [FileName],
	             [status(_), stdout(string(S))]),
	atom_codes(FileName, SFileName),
	append(SFileName, "\n", SFileNameN),
	append(SFileDir,  "/" || SFileNameN, S),
	atom_codes(FileDir, SFileDir),
	!.

% Note: libmysqlclient-dev in Debian

% ---------------------------------------------------------------------------

:- discontiguous(m_bundle_foreign_config_tool/3).
	
:- doc(section, "Options for GSL Library").

:- bundle_param('WITH_GSL', [
    valid_values(['yes', 'no']),
    %
    default_comment("GSL available"),
    default_value_comment(no,
        "GSL has not been detected.  If you want to use the math\n"||
        "library it is highly recommended that you stop the Ciao\n"||
        "configuration and install the GSL library first."),
    rule_default(WithGSL, verify_gsl(WithGSL)),
    % rule_default('no'),
    %
    interactive([extended],
      % .....................................................................
      "Set to \"yes\" if you wish to interface with the GSL (GNU Scientific\n"||
      "Library). If you choose to have the GSL interface, you should have the\n"||
      "GSL development library installed in the machine where you are\n"||
      "compiling and using it."),
    comment("Install GSL support")
]).

m_bundle_foreign_config_tool(ciao, gsl, 'gsl-config').

% TODO: it should consider auto_install option!
gsl_installed :-
	find_executable(~m_bundle_foreign_config_tool(ciao, gsl), _),
	% TODO: Next literal required because now GSL 32 bits is not available
	% TODO: in Linux 64 bits -- EMM.
	\+ get_platform('LINUXi86_64').

verify_gsl(Value) :-
	( gsl_installed -> Value = yes ; Value = no ).

:- bundle_param('AUTO_INSTALL_GSL', [
    valid_values(['yes', 'no']),
    %
    rule_default('no'),
    %
    interactive([extended],
      % .....................................................................
      "Set to \"yes\" if you want to auto-install GSL (third party)"),
    comment("Auto-install GSL (third party)")
]).

% ---------------------------------------------------------------------------

:- doc(section, "Options for PPL Library").

:- bundle_param('WITH_PPL', [
    valid_values(['yes', 'no']),
    %
    rule_default('no'),
    %
    interactive([extended],
      % .....................................................................
      "Set to \"yes\" if you wish to interface with PPL"),
    comment("Enable PPL bindings")
]).
		% rule_default(verify_ppl(WithPPL), WithPPL),
%  default_comment("PPL >= 0.9 available"),
%  default_value_comment("PPL has not been detected."),

m_bundle_foreign_config_tool(ciao, ppl, 'ppl-config').

% TODO: it should consider auto_install option!
%% ppl_installed :-
%% 	find_executable(~m_bundle_foreign_config_tool(ciao, ppl), _).
%% 
%% % TODO: not used... reactivate?
%% verify_ppl(Value) :-
%% %	( ppl_installed, ppl_version(V) ->
%% %	    Value = yes
%% %	    message([V])
%% 	( ppl_installed, ppl_version(_) ->
%% 	    Value = yes
%% 	; Value = no
%% 	).
%% 
%% ppl_version(Version, Str) :-
%% 	foreign_config_version(ppl, Str).

:- bundle_param('AUTO_INSTALL_PPL', [
    valid_values(['yes', 'no']),
    %
    rule_default('no'),
    %
    interactive([extended],
      % .....................................................................
      "Set to \"yes\" if you want to auto-install PPL (third party)"),
    comment("Auto-install PPL (third party)")
]).

% ---------------------------------------------------------------------------

:- doc(section, "Options for Mathematica Support").

:- bundle_param('WITH_MATHEMATICA', [
    valid_values(['yes', 'no']),
    %
    default_comment("Mathematica available"),
    default_value_comment(no,
      % .....................................................................
      "Mathematica has not been detected.  If you want to use Mathematica\n" ||
      "interface it is highly recommended that you stop the Ciao\n" || 
      "configuration, install the Mathematica first and/or add to your PATH\n" ||
      "variable the directory of the MathKernel command line."),
    rule_default(WithMathematica, verify_mathematica(WithMathematica)),
    %
    interactive([extended],
      % .....................................................................
      "Set to \"yes\" if you wish to interface with the Mathematica.\n" ||
      "If you choose to have the Mathematica interface, you should have\n"||
      "Mathemtical installed in the machine where you are compiling and\n" || 
      "using it, then and/or add to your PATH variable the directory of\n"|| 
      "the MathKernel command line."),
    comment("Install Mathematica support")
]).

% TODO: remove fsR ? 
m_bundle_foreign_config_tool(ciao, mathematica, ~fsR(bundle_src(ciao)/contrib/mathematica/'mathematica-config.sh')).

mathematica_installed :-
	find_executable(~m_bundle_foreign_config_tool(ciao, mathematica), Exec), 
	process_call(path(Exec), ['--mathlink'],
	             [stdout(null), stderr(stdout), status(0)]).

verify_mathematica(Value) :-
	( mathematica_installed -> Value = yes ; Value = no ).

% ---------------------------------------------------------------------------

:- doc(section, "Options for Java Interface").

:- bundle_param('WITH_JAVA_INTERFACE', [
    valid_values(['yes', 'no']),
    %
    default_comment("Javac and Javadoc available"),
    default_value_comment(no,
        "Java has not been detected. If you would like to use the\n"||
        "utilities for the Java interface it is highly recommended that\n"||
        "you stop the Ciao configuration now and install Java first."),
% 	    "If Java is already installed and you use Debian/Ubuntu perhaps\n"||
% 	    "you forgot to run: sudo update-java-alternatives --set java-6-sun."
    rule_default(VerifyJava, verify_java(VerifyJava)),
    %
    interactive([extended],
      % .....................................................................
      "Whether you have a reasonably recent version of Java.\n"||
      "If so the utilities for the Java interface under\n"||
      "$(CIAODESRC)/ciao/library/javall will be compiled , along with\n"||
      "examples and documentation."),
    comment("Use Java interface")
]).

verify_java(Value) :-
	( javac_installed, javadoc_installed -> Value = yes ; Value = no ).

javac_installed :-
	find_executable('javac', _),
	is_sun_javac.

% TODO: Any better way to detect Sun Java? (EMM)
is_sun_javac :-
	process_call(path(javac), ['-version'],
	             [stderr(stdout), stdout(string(String)), status(_)]),
	append(_, "javac 1."||_, String),
	process_call(path(java), ['-version'],
	             [stderr(stdout), stdout(string(SJava)), status(_)]),
	% Kludge: In linux 64, you have to use the 64-bit Server VM --EMM
	( get_platform('LINUXi86_64') -> append(_, "64-Bit"||_, SJava)
	; true
	),
	!.

javadoc_installed :- find_executable('javadoc', _).

% ---------------------------------------------------------------------------

:- doc(section, "Ant Options").

:- bundle_param('WITH_ANT', [
    valid_values(['yes', 'no']),
    %
    default_comment("Ant is available"),
    default_value_comment(no,
        "Ant has not been detected. If you would like to use the\n"||
        "resource analysis for java, which requires ant, it is\n"||
        "highly recommended that you stop the Ciao configuration\n"||
        "now and install Ant first."),
    rule_default(VerifyAnt, (
      param('WITH_JAVA_INTERFACE'(WithJavaInterface)),
      verify_ant(WithJavaInterface, VerifyAnt))),
    %
    % TODO: this should not be configurable...
    interactive([extended]),
    comment("Ant is available")
]).

verify_ant(no,  no).
verify_ant(yes, VerifyAnt) :-
	verify_ant_yes(VerifyAnt).

verify_ant_yes(Value) :-
	( ant_installed -> Value = yes ; Value = no ).

ant_installed :- find_executable('ant', _).

% ---------------------------------------------------------------------------

:- doc(section, "Engine and C Compilation Options").

% ---------------------------------------------------------------------------
% TODO: Those flags are set from sh code in ciaosetup (see
%       do__bootstrap_settings), but those definitions are necessary to
%       write the final configuration files from Prolog.

% (also needed by config-and-dump.sh)
:- bundle_param('CUSTOM_CC', [
    rule_default(''),
    %
    interactive([extended]),
    comment("Custom C compiler")
]).

% (also needed by config-and-dump.sh)
:- bundle_param('CUSTOM_LD', [
    rule_default(''),
    %
    interactive([extended]),
    comment("Custom C linker")
]).

% (also needed by config-and-dump.sh)
:- bundle_param('EXTRA_CFLAGS', [
    rule_default(''),
    %
    interactive([extended]),
    comment("Additional C compiler flags")
]).

% (also needed by config-and-dump.sh)
:- bundle_param('EXTRA_LDFLAGS', [
    rule_default(''),
    %
    interactive([extended]),
    comment("Specify additional C linker flags")
]).

% ---------------------------------------------------------------------------

:- bundle_param('OPTIMIZING_COMPILER', [
    valid_values(['yes', 'no']),
    %
    rule_set_value(no, (
      param(ciaode:'INSTYPE'(InsType)),
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      \+((InsType == 'local', SysregType == 'user')))),
    rule_default('no'),
    %
    interactive([minimum, extended]),
    comment("Install optimizing compiler")
]).

% ---------------------------------------------------------------------------

% (also needed by config-and-dump.sh)
:- bundle_param('USE_THREADS', [
    valid_values(['yes', 'no']),
    %
    rule_default('yes'),
    %
    interactive([extended],
      % .....................................................................
      "If you wish to compile an engine with threads capability\n"||
      "(concurrency), set the following variable to \"yes\".  Otherwise, set\n"||
      "it to \"no\".  If the architecture does not support threads (or\n"||
      "thread support has not yet been added to Ciao for this\n"||
      "architecture), this will be automatically disabled at compile time.\n"||
      "Concurrency support does not cause any appreciable runtime overhead\n"||
      "for non-concurrent programs, so it is safe to leave it as \"yes\"."),
    comment("Enable threads in engine")
]).

% ---------------------------------------------------------------------------

% (also needed by config-and-dump.sh)
:- bundle_param('AND_PARALLEL_EXECUTION', [
    valid_values(['yes', 'visandor', 'no']),
    %
    rule_default('no'),
    %
    interactive([extended],
      % .....................................................................
      "Set the following variable to \"yes\" if you wish to compile an\n"||
      "engine with support for and-parallel execution of goals in\n"||
      "(Herbrand-)independent fashion or to \"visandor\" if you wish also\n"||
      "support for VisAndOr's events. Choose one of:\n\n"||
      "        yes             -- Support for and-parallel execution.\n"||
      "        visandor        -- Support for and-parallel execution and\n"||
      "                           VisAndOr's events.\n"||
      "        no              -- No support."),
    comment("Enable and-parallel execution")
]).

% ---------------------------------------------------------------------------

% (also needed by config-and-dump.sh)
:- bundle_param('PAR_BACK', [
    valid_values(['yes', 'no']),
    %
    rule_default('no'),
    %
    interactive([extended],
      % .....................................................................
      "Set the following variable to \"yes\" if you wish to compile an\n"||
      "engine with support for parallel backtracking execution of goals.\n"||
      "This feature is experimental and may not be available in all releases."),
    comment("Enable parallel backtracking")
]).

% ---------------------------------------------------------------------------

% TODO: should be: with-...?
% (also needed by config-and-dump.sh)
:- bundle_param('TABLED_EXECUTION', [
    valid_values(['yes', 'no']),
    %
    rule_default('yes'),
    %
    interactive([extended],
      % .....................................................................
      "Set the following variable to \"yes\" if you wish to compile an engine\n"||
      "with support for tabled execution of goals."),
    comment("Enable tabled execution")
]).

% ---------------------------------------------------------------------------

% (also needed by config-and-dump.sh)
:- bundle_param('OPTIM_LEVEL', [
    valid_values(['optimized', 'normal']),
    %
    rule_default('optimized'),
    %
    interactive([extended],
      % .....................................................................
      "Optimization level used when compiling the bytecode emulator. Choose\n"||
      "one of:\n"||
      "\n"||
      "   optimized       -- Turn on optimization flags\n"||
      "   normal          -- Normal emulator (non-optimized code)\n"||
      "\n"||
      "For normal use, we recommend leaving it as \"optimized\".  But if you\n"||
      "suspect that your C compiler performs buggy optimizations (which\n"||
      "should not be the case), turn optimization off.  This may happen more\n"||
      "easily in concurrent applicacions: if you write any thread-based\n"||
      "program and unexpected results appear, try recompiling Ciao without\n"||
      "optimization options first."),
   comment("Optimization level")
]).

% ---------------------------------------------------------------------------

% TODO: I'd say... mostly obsolete
:- bundle_param('CROSS_COMPILER_HOST', [
    rule_default('none'),
    %
    interactive([extended],
      % .....................................................................
      "If you will cross-compile the engine later, please enter the user\n"||
      "name and address of the target machine to extract run-time\n"||
      "characteristics from -- e.g., \"root@my.other.host.com\".  If you\n"||
      "are not going to crosscompile, leave the default value.\n"||
      "Cross-compiling is at the moment done with \"make build crossengine\"\n"||
      "once the regular compilation is completed."),
   comment("Host for cross compilation")
]).

% ---------------------------------------------------------------------------

:- bundle_param('DEBUG_LEVEL', [
    valid_values(['nodebug', 'debug', 'profile', 'profile-debug',
    	'paranoid-debug']),
    %
    rule_default('nodebug'),
    %
    interactive([extended],
      % .....................................................................
      "You only want to change this if you are a developer.  Additionally,\n"||
      "setting the environment variable CIAODEBUG to the value \'-debug\'\n"||
      "at the time of compiling the engine will override the OPTIM_LEVEL\n"||
      "and DEBUG_LEVEL flags, and produce non optimized emulator code with\n"||
      "debugging information.\n"||
      "\n"||
      "Level of debugging built into the bytecode emulator. Choose one of:\n"||
      "\n"||
      "   nodebug         -- Do not include debug information or messages\n"||
      "   debug           -- Emulator with C level debugging info available\n"||
      "                      plus extended C compilation warnings\n"||
      "   profile         -- Include profiling options for the emulator\n"||
      "   profile-debug   -- Include profiling and debug options for the\n"||
      "                      emulator\n"||
      "   paranoid-debug  -- Emulator with C level debugging info available\n"||
      "                      plus paranoid C compilation warnings."),
   comment("Engine debug level")
]).

% ===========================================================================

:- doc(section, "Configurable Prolog flags").
% TODO: declaration in modules should be extracted from this configuration

:- bundle_param('UNUSED_PRED_WARNINGS', [
    config_prolog_flag(unused_pred_warnings,[yes,no],no),
    interactive([extended],
      % .....................................................................
      "If you wish to show warnings about unused predicates, set this\n"||
      "variable to \"yes\"."),
    comment("Enable unused predicate warnings")
]).
:- bundle_param('COMPRESS_EXEC', [
    config_prolog_flag(compress_exec,[yes,no],no)
]).
:- bundle_param('ITF_FORMAT', [
    config_prolog_flag(itf_format,[f,r],f)
]).
:- bundle_param('COMPRESS_LIB', [
    config_prolog_flag(compress_lib,[yes,no],no),
    interactive([extended],
      % .....................................................................
      "If you wish to compile the Ciao libraries with their bytecode\n"||
      "compressed then set the following variable to \"yes\". Libraries\n"||
      "generated this way will be smaller at the cost of a slightly slower\n"||
      "usage, both in their load as when used to create an executable."),
    comment("Enable compressed bytecode")
]).
:- bundle_param('RUNTIME_CHECKS', [
    config_prolog_flag(runtime_checks,[yes,no],no),
    interactive([extended],
      % .....................................................................
      "If you wish to compile the Ciao libraries with runtime checks enabled\n"||
      "then set the following variable to \"yes\". This of course reduces\n"||
      "performance."),
    comment("Enable (optional) runtime checks")
]).
:- bundle_param('RTCHECKS_LEVEL', [
    config_prolog_flag(rtchecks_level,[inner,exports],inner)
]).
:- bundle_param('RTCHECKS_TRUST', [
    config_prolog_flag(rtchecks_trust,[yes,no],yes)
]).
:- bundle_param('RTCHECKS_ENTRY', [
    config_prolog_flag(rtchecks_entry,[yes,no],yes)
]).
:- bundle_param('RTCHECKS_EXIT', [
    config_prolog_flag(rtchecks_exit,[yes,no],yes)
]).
:- bundle_param('RTCHECKS_TEST', [
    config_prolog_flag(rtchecks_test,[yes,no],no)
]).
:- bundle_param('RTCHECKS_INLINE', [
    config_prolog_flag(rtchecks_inline,[yes,no],no)
]).
:- bundle_param('RTCHECKS_ASRLOC', [
    config_prolog_flag(rtchecks_asrloc,[yes,no],yes)
]).
:- bundle_param('RTCHECKS_PREDLOC', [
    config_prolog_flag(rtchecks_predloc,[yes,no],yes)
]).
:- bundle_param('RTCHECKS_CALLLOC', [
    config_prolog_flag(rtchecks_callloc,[no,literal,predicate],predicate)
]).
:- bundle_param('RTCHECKS_NAMEFMT', [
    config_prolog_flag(rtchecks_namefmt,[short,long],long)
]).
:- bundle_param('RTCHECKS_ABORT_ON_ERROR', [
    config_prolog_flag(rtchecks_abort_on_error,[yes,no],no)
]).






