#!/bin/sh
#
# Generation of setting options for the engine compilation and
# internal script installation commands (that are necessarily not
# encodable in Prolog, since that would not allow us to bootstrap the
# system).
#
# Parameters: CIAOBUILDDIR
#
# Jose F. Morales (based on previous code)
# ----------------------------------------------------------------------------

# Get CIAOBUILDDIR
CIAOBUILDDIR="${1}"

# Include ${CIAOBUILDDIR}/CONFIG_input (contains configuration variables)
. "${CIAOBUILDDIR}/CONFIG_input"

# Name used for the engine lib
ENGINE_LIB_NAME="ciao"

SRC="${ciaode__CIAODESRC}/ciao"

# ===========================================================================
SHELL="/bin/sh"
SYSDEP_FILES="${ciaode__CIAODESRC}/ciao/config-sysdep"
CIAOARCH="`"${ciaode__CIAODESRC}/ciao/etc/ciao_get_arch"`"

if test x"${CIAOARCH}" = x""; then
    echo "No CIAOARCH" >& 2
    exit 1
fi

# Number of processors: use the real number
PROCESSORS="`"${SYSDEP_FILES}/available_processors" ${CIAOARCH}`"

# TODO: Only for engine/Makefile
ETAGS="etags --ignore-indentation"

# TODO: duplicated in bundle_config__ciao.pl
if test x"${ciao__DEBUG_LEVEL}" != x"nodebug"; then
    CIAODEBUG="-${ciao__DEBUG_LEVEL}"
else
    CIAODEBUG=
fi

if test x"${ciao__AND_PARALLEL_EXECUTION}" = x"yes"; then
    ANDPARALLEL="-DANDPARALLEL"
elif test x"${ciao__AND_PARALLEL_EXECUTION}" = x"visandor"; then
    ANDPARALLEL="-DANDPARALLEL -DVISANDOR"
elif test x"${ciao__AND_PARALLEL_EXECUTION}" = x"no"; then
    ANDPARALLEL=""
fi

if test x"${ciao__PAR_BACK}" = x"yes"; then
    PARBACK="-DPARBACK"
fi

if test x"${ciao__TABLED_EXECUTION}" = x"yes"; then
    TABLING="-DTABLING"
fi

# General optimization flags; can be overriden in specific architectures
if test x"${CIAODEBUG}" = x"-debug"; then
    # TODO: override?
    ciao__OPTIM_LEVEL="normal"
    if test x"${ciao__DEBUG_LEVEL}" = x"nodebug"; then
	# TODO: override?
	ciao__DEBUG_LEVEL="debug"
    fi
fi

if test x"${ciao__OPTIM_LEVEL}" = x"optimized"; then
    OPTIM_FLAGS="-O2"
elif test x"${ciao__OPTIM_LEVEL}" = x"normal"; then
    OPTIM_FLAGS=""
fi

case "${ciao__DEBUG_LEVEL}" in
    paranoid-debug)
	DEBUG_FLAGS="-g -Wall -W -Wtraditional -Wshadow -Wpointer-arith -Wbad-function-cast -Wcast-qual -Wcast-align -Wconversion -Waggregate-return -Wstrict-prototypes -Wmissing-prototypes -Wmissing-declarations -Wredundant-decls -Wnested-externs -Winline -DDEBUG"
	;;
    debug)
	DEBUG_FLAGS="-g -Wall -DDEBUG -DGLOBVARS -DINTERNAL_CALLING"
	;;
    nodebug)
	DEBUG_FLAGS=""
	;;
    profile)
	DEBUG_FLAGS="-DDEBUG -DPROFILE"
#    DEBUG_FLAGS="-DPROFILE"
	;;
    profile-debug)
	DEBUG_FLAGS="-g -Wall -DDEBUG -DGLOBVARS -DINTERNAL_CALLING -DPROFILE"
	;;
esac

# ===========================================================================
# Do the actual configuration
# TODO: include this file here? merge with previous code?

. "${SYSDEP_FILES}/config-sysdep.sh"

CFLAGS="${OPTIM_FLAGS} ${DEBUG_FLAGS} ${THREAD_FLAG} ${FOREIGN_FILES_FLAG} ${ANDPARALLEL} ${VISANDOR} ${PARBACK} ${TABLING} -D${OSNAME} -D${ARCHNAME} ${ARCHFLAGS} ${MEM_MNG_FLAG} ${ciao__EXTRA_CFLAGS}"

LDFLAGS="${LINK_FLAGS} ${ciao__EXTRA_LDFLAGS}"

# ===========================================================================

# in ciao/ciaoc/Makefile, ciao/Makefile, ciao/shell/Makefile
VERSION="`cat "${ciaode__CIAODESRC}/ciao/version/GlobalVersion"`"
PATCH="`cat "${ciaode__CIAODESRC}/ciao/version/GlobalPatch"`"

case "`uname`" in
    *CYGWIN*)
	ALT="win32"
	EXECSUFFIX=".exe"
	;;
esac

case "${OSTYPE}" in
    cygwin)
	ALT="win32"
	EXECSUFFIX=".exe"
	;;
    cygwin32)
	ALT="win32"
	EXECSUFFIX=".exe"
	;;
esac

# The next is a kludge to use the .cpx extension in Windows
if test x"${ALT}" = x"win32"; then
    CIAOSUFFIX=".cpx"
fi

# ciao/Makefile ciao/ciaoc/Makefile ciao/shell/Makefile
CIAOC="${CIAOBUILDDIR}/bin/ciaoc-${VERSION}${CIAOSUFFIX}"

# ===========================================================================

BUILD_OBJDIRROOT="${CIAOBUILDDIR}/objs"
BUILD_OBJDIR="${BUILD_OBJDIRROOT}/${CIAOARCH}${CIAODEBUG}"

# Relative path from ${BUILD_OBJDIR} to the ${ciaode__CIAODESRC}/ciao/engine directory
ENGSRC_FROM_BUILD_OBJDIR="../../../ciao/engine"

# ciao/Makefile ciao/engine/Makefile
BUILD_HDIRROOT="${CIAOBUILDDIR}/include"
BUILD_HDIR="${BUILD_HDIRROOT}/${CIAOARCH}${CIAODEBUG}"

# TODO: duplicated in name_ciaoengine/1
ENGINEBASE="ciaoengine"
ENGINENAME="${ENGINEBASE}${EXECSUFFIX}" # may be patched by config-sysdep.sh

# Custom options for cross Win32
# TODO: move seamlessly into config-sysdep.sh? (by setting the engine extension and adding cc flags in the right variable)
case "${CIAOARCH}" in
    crossWin32i86)
       # TODO: why?
       ENGINENAME="${ENGINENAME}.exe"
       # TODO: why?
       # Standard libraries
       CFLAGS="${CFLAGS} -DSTATICENG -DcrossWin32i86 -I/usr/i386-mingw32/include"
       ;;
    *) true
esac    

# ===========================================================================

# Export settings as Makefile/shell definitions

# The first argument defines the string quoting character for the
# specified output kind. That ensures that values consisting on
# several words are correctly assigned (i.e., V="foo bar" or V=foo
# bar).
emit_CONFIG() { # ${1} is the quotation char
    Q="${1}"
    # TODO: Sure? Configuration of GSL seems to be done too late...
    # TODO: Generalize for other libraries
    CONFIG_GSL_FILE="${CIAOBUILDDIR}/CONFIG_GSL"
    if test -r "${CONFIG_GSL_FILE}"; then
	. "${CONFIG_GSL_FILE}"
	STAT_LIBS="${STAT_LIBS} ${GSL_STAT_LIBS}"
    fi
    echo
    # TODO: Prune options (most of them are not needed here now that the use of Makefile is very limited)
    cat <<EOF
# (previously named SETTINGS)
# -- values provided by bundle_config__*.pl
ciaode__CIAODESRC=${Q}${ciaode__CIAODESRC}${Q}
ciaode__DEFAULTLIBDIR=${Q}${ciaode__DEFAULTLIBDIR}${Q}
ciaode__EXECMODE=${Q}${ciaode__EXECMODE}${Q}
ciaode__DATAMODE=${Q}${ciaode__DATAMODE}${Q}
ciaode__INSTALLGROUP=${Q}${ciaode__INSTALLGROUP}${Q}
ciao__CUSTOM_CC=${Q}${ciao__CUSTOM_CC}${Q}
ciao__CUSTOM_LD=${Q}${ciao__CUSTOM_LD}${Q}
ciao__EXTRA_CFLAGS=${Q}${ciao__EXTRA_CFLAGS}${Q}
ciao__EXTRA_LDFLAGS=${Q}${ciao__EXTRA_LDFLAGS}${Q}
ciao__USE_THREADS=${Q}${ciao__USE_THREADS}${Q}
ciao__AND_PARALLEL_EXECUTION=${Q}${ciao__AND_PARALLEL_EXECUTION}${Q}
ciao__PAR_BACK=${Q}${ciao__PAR_BACK}${Q}
ciao__TABLED_EXECUTION=${Q}${ciao__TABLED_EXECUTION}${Q}
ciao__OPTIM_LEVEL=${Q}${ciao__OPTIM_LEVEL}${Q}
ciao__CROSS_COMPILER_HOST=${Q}${ciao__CROSS_COMPILER_HOST}${Q}
ciao__DEBUG_LEVEL=${Q}${ciao__DEBUG_LEVEL}${Q}
# -- values defined in config-and-dump.sh
# Define this to be the main file name (with suffix if there is one).
ENGINE_LIB_NAME=${Q}${ENGINE_LIB_NAME}${Q}
VERSION=${Q}${VERSION}${Q}
PATCH=${Q}${PATCH}${Q}
# (parameter of config-and-dump.sh)
CIAOBUILDDIR=${Q}${CIAOBUILDDIR}${Q}
ALT=${Q}${ALT}${Q}
EXECSUFFIX=${Q}${EXECSUFFIX}${Q}
# The next is a kludge to use the .cpx extension in Windows
CIAOSUFFIX=${Q}${CIAOSUFFIX}${Q}
CIAOC=${Q}${CIAOC}${Q}
CIAOCOPTS=${Q}${CIAOCOPTS}${Q}
# ===========================================================================
# TODO: unused later? (redefined in engine/Makefile)
SHELL=${Q}${SHELL}${Q}
# TODO: unused later?
SYSDEP_FILES=${Q}${SYSDEP_FILES}${Q}
CIAOARCH=${Q}${CIAOARCH}${Q}
# Number of processors: use the real number
PROCESSORS=${Q}${PROCESSORS}${Q}
# TODO: unused later? (used in engine/Makefile for 'create_tags' target)
ETAGS=${Q}${ETAGS}${Q}
# TODO: unused later?
CIAODEBUG=${Q}${CIAODEBUG}${Q}
# TODO: unused later?
ANDPARALLEL=${Q}${ANDPARALLEL}${Q}
# TODO: unused later?
PARBACK=${Q}${PARBACK}${Q}
# TODO: unused later?
TABLING=${Q}${TABLING}${Q}
# TODO: unused later?
OPTIM_FLAGS=${Q}${OPTIM_FLAGS}${Q}
# TODO: unused later?
DEBUG_FLAGS=${Q}${DEBUG_FLAGS}${Q}
# TODO: unused later? (used in engine/Makefile)
CFLAGS=${Q}${CFLAGS}${Q}
# TODO: unused later? (used in engine/Makefile)
LDFLAGS=${Q}${LDFLAGS}${Q}
# ===========================================================================
# From config-sysdep.sh
# TODO: refine
OSNAME=${Q}${OSNAME}${Q}
ARCHNAME=${Q}${ARCHNAME}${Q}
# TODO: unused later? (used in engine/Makefile)
CC=${Q}${CC}${Q}
# TODO: unused later? (used in engine/Makefile)
LD=${Q}${LD}${Q}
# TODO: unused later?
FOREIGN_FILES_FLAG=${Q}${FOREIGN_FILES_FLAG}${Q}
# TODO: unused later?
CCSHARED=${Q}${CCSHARED}${Q}
# TODO: unused later?
LDSHARED=${Q}${LDSHARED}${Q}
SOSUFFIX=${Q}${SOSUFFIX}${Q}
STAT_LIBS=${Q}${STAT_LIBS}${Q}
LCC=${Q}${LCC}${Q}
LLD=${Q}${LLD}${Q}
INSTALL=${Q}${INSTALL}${Q}
ENGINE_LINK_TYPE=${Q}${ENGINE_LINK_TYPE}${Q}
LIBS=${Q}${LIBS}${Q}
ENGLOCATION=${Q}${ENGLOCATION}${Q}
LDRPATH=${Q}${LDRPATH}${Q}
NOCONSOLEFLAG=${Q}${NOCONSOLEFLAG}${Q}
SDBEGIN=${Q}${SDBEGIN}${Q}
OSREV=${Q}${OSREV}${Q}
# ===========================================================================
BUILD_OBJDIRROOT=${Q}${BUILD_OBJDIRROOT}${Q}
BUILD_OBJDIR=${Q}${BUILD_OBJDIR}${Q}
# Relative path from ${BUILD_OBJDIR} to the ${ciaode__CIAODESRC}/ciao/engine directory
ENGSRC_FROM_BUILD_OBJDIR=${Q}${ENGSRC_FROM_BUILD_OBJDIR}${Q}
# ciao/Makefile ciao/engine/Makefile
BUILD_HDIRROOT=${Q}${BUILD_HDIRROOT}${Q}
BUILD_HDIR=${Q}${BUILD_HDIR}${Q}
ENGINEBASE=${Q}${ENGINEBASE}${Q}
ENGINENAME=${Q}${ENGINENAME}${Q}
EOF
}

emit_as_atmlist() {
    LST="["
    if [ $# -ge 1 ]; then
	LST=${LST}"'""$1""'"
	shift
	while [ $# -ge 1 ]; do
	    LST=${LST}", '""$1""'"
	    shift
	done
    fi
    LST=${LST}"]"
    echo "${LST}"
}

# Emit compile_options_auto, for foreign compilation
# TODO: simplify, move outside config-and-dump.sh
emit_compile_options_auto() {
    CCFORCOMP="`emit_as_atmlist ${CCSHARED} ${ciao__EXTRA_CFLAGS}`"
    LDFORCOMP="`emit_as_atmlist ${LDSHARED} ${ciao__EXTRA_LDFLAGS}`"

    # TODO: This is generated in the lib/ directory. I would like to
    # generate it in the build directory instead, but it would be
    # useless now since there is no alias path for that directory, and
    # that file is included from Prolog.
    OPTFILE="${ciaode__CIAODESRC}/ciao/lib/compile_options_auto.pl"
    cat > "${OPTFILE}" <<EOF
%%% DO NOT EDIT THIS FILE!
%%% IT HAS BEEN AUTOMATICALLY GENERATED DURING COMPILATION!
foreign_compiler_options('${OSNAME}','${ARCHNAME}','${CC}',${CCFORCOMP}).
foreign_linker_options('${OSNAME}','${ARCHNAME}','${LD}',${LDFORCOMP}).
EOF
    chmod "${ciaode__DATAMODE}" "${OPTFILE}"
}

# ===========================================================================

# Export config for 'sh' scripts
emit_CONFIG "'" > "${CIAOBUILDDIR}/CONFIG_sh"
# Export config for Makefiles
emit_CONFIG "" > "${CIAOBUILDDIR}/CONFIG_mkf"
# Export config (just compile options) for Prolog (the foreign interface)
emit_compile_options_auto

