/*
 *  alloc.c
 *
 *  Allocation of principal WAM areas.
 *
 *  Copyright (C) 1996,1997,1998, 1999, 2000, 2001, 2002 UPM-CLIP
 */

#if defined(Solaris) || defined(LINUX) || defined(DARWIN) || defined(BSD)
#include <string.h>
#else
#include <memory.h>
#endif
#include <unistd.h>

#include "datadefs.h"
#include "support_macros.h"
#include "own_malloc.h"
#include "alloc.h"

char *tryalloc_errstring;

#define MEMORY_FAULT(MESSAGE) {			\
    tryalloc_errstring = (MESSAGE);		\
    return NULL;				\
  }

#define CHECK_FOR_MEMORY_FAULT(ALLOC_CALL)  {	   \
    tagged_t *__ptr = (ALLOC_CALL);		   \
    if (!__ptr) SERIOUS_FAULT(tryalloc_errstring); \
    return __ptr;				   \
  }

/* segfault patch -- jf */
#if defined(MallocBase)     
#define ENSURE_ADDRESSABLE(P, SIZE) \
  (((tagged_t)(P) >= (tagged_t)MallocBase) && \
  (((tagged_t)(P) - (tagged_t)MallocBase) < POINTERMASK) && \
   (((tagged_t)(P) - (tagged_t)MallocBase + (SIZE)) < POINTERMASK))
#else
#define ENSURE_ADDRESSABLE(P, SIZE) \
  ((tagged_t)(P) < POINTERMASK)
#endif

static tagged_t *get_tiny_blocks(void);
/*static char *mem_start; */  /* beginning of our virtual memory -- Shared */
extern int end;                             /* Does it exist in NeXT? (MCL) */

 /* total # bytes grabbed by malloc/realloc/free . Not accurately measured
    -- it depends on the mem. mang. own data structures (MCL) */
intmach_t total_mem_count = 0;                                 /* Shared */

 /* locks for accessing the shared memory.  Although malloc() is MT-safe in
    Solaris (how 'bout Linux?), the engine uses several internal data
    structures for its own memory management which should be correctly
    updated.  So we use our own lock for providing an atomic access to
    tryalloc/tryrealloc/checkdealloc (MCL).
 */

extern SLOCK    mem_mng_l;

           /* From an execution profile, 24 seems a good threshhold (MCL) */
#define THRESHHOLD 24
#define NTINY 682

static tagged_t *tiny_blocks = NULL;                     /* Shared & locked */

static tagged_t *get_tiny_blocks(void)
{
  CIAO_REGISTER int i;
  tagged_t *ptr;
  CIAO_REGISTER tagged_t *p;
  tagged_t tail = 0;

 /* prt was a call to checkalloc, but I made a direct call to Malloc()
    because of recursive locks stopping the engine.  Thus, the total amount
    of memory is also increased here. */

  ptr = Malloc(NTINY*THRESHHOLD);

  if (!ptr) {
    Release_slock(mem_mng_l);
    MEMORY_FAULT("Memory allocation failed [Malloc()]");
  }
  if (!ENSURE_ADDRESSABLE(ptr, NTINY*THRESHHOLD)) {
    Release_slock(mem_mng_l);
    MEMORY_FAULT("Memory out of addressable bounds! [Malloc()]");
  }

  p = ptr+(NTINY*THRESHHOLD>>2);
  total_mem_count += NTINY*THRESHHOLD;

  for (i=NTINY; i>0; i--)
    p -= THRESHHOLD>>2, p[0] = tail, tail = (tagged_t)p;
  tiny_blocks = ptr+(THRESHHOLD>>2);
  return ptr;
}

tagged_t *tryalloc(int size)
{
  tagged_t *p;
  Wait_Acquire_slock(mem_mng_l);

  if (size<=THRESHHOLD) {
    if ((p=tiny_blocks))
      tiny_blocks = (tagged_t *)p[0];
    else {
      p=get_tiny_blocks();
      if (!p) {                                      /* get_tiny_block fails */
	Release_slock(mem_mng_l);
	return NULL;
      }
    }
  } else {                                           /* size > THRESHHOLD */
    p = (tagged_t *)Malloc(size);

    if (!p) {
      Release_slock(mem_mng_l);
      MEMORY_FAULT("Memory allocation failed [Malloc()]");
    }
    /* segfault patch -- jf */
    if (!ENSURE_ADDRESSABLE(p, size)) {
      Release_slock(mem_mng_l);
      MEMORY_FAULT("Memory out of addressable bounds! [Malloc()]");
    }

    total_mem_count += size;
  }
#if defined(DEBUG)
  if (debug_mem)
    printf("tryalloc returned %x, %d chars\n", (unsigned int)p, size);
#endif
  Release_slock(mem_mng_l);
  return p;
}

tagged_t *checkalloc(int size)
{
  CHECK_FOR_MEMORY_FAULT(tryalloc(size));
}

void checkdealloc(tagged_t *ptr, int decr)
{
  Wait_Acquire_slock(mem_mng_l);

  if (decr<=THRESHHOLD) {
    ptr[0] = (tagged_t)tiny_blocks;
    tiny_blocks = ptr;
  } else {
    total_mem_count -= decr;
    Free(ptr);
  }
#if defined(DEBUG)
  if (debug_mem)
    printf("checkdealloc freed %x, %d chars\n", (unsigned int)ptr, decr);
#endif
  Release_slock(mem_mng_l);
}


tagged_t *tryrealloc(tagged_t *ptr, int decr, int size)
{
  tagged_t *p;

  Wait_Acquire_slock(mem_mng_l);

  if (decr<=THRESHHOLD) {
    if (size<=THRESHHOLD)
      p = ptr;
    else {
      p = Malloc(size);                       /* Was a call to checkalloc */
       if (!p) {
        Release_slock(mem_mng_l);
        MEMORY_FAULT("Memory allocation failed [Malloc()]");
      }
      /* segfault patch -- jf */
      if (!ENSURE_ADDRESSABLE(p, size)) {
	Release_slock(mem_mng_l);
	MEMORY_FAULT("Memory out of addressable bounds [Malloc()]");
      }
      total_mem_count += size;
      memcpy(p, ptr, ((decr-1) & -4)+4);
      ptr[0] = (tagged_t)tiny_blocks;
      tiny_blocks = ptr;
    }
  } else                                             /* decr > THRESHHOLD */
    if (size<=THRESHHOLD) {
      if (!(p=tiny_blocks)) {
	p=get_tiny_blocks(); 
	if (!p) {                                    /* get_tiny_block fails */
	  Release_slock(mem_mng_l);
	  return NULL;
	}
      }
      memcpy(p, ptr, ((size-1) & -4)+4);
    } else {
      p = (tagged_t *)Realloc(ptr,size);
      if (!p) {
        Release_slock(mem_mng_l);
        MEMORY_FAULT("Memory allocation failed [in Realloc()]");
      }
      /* segfault patch -- jf */
      if (!ENSURE_ADDRESSABLE(p, size)) {
	Release_slock(mem_mng_l);
	MEMORY_FAULT("Memory out of addressable bounds [Realloc()]");
      }
      total_mem_count += (size-decr);
    }
#if defined(DEBUG)
  if (debug_mem)
  printf("tryrealloc returned %x, %d chars\n",
         (unsigned int)p, size);
#endif
  Release_slock(mem_mng_l);
  return p;
}

tagged_t *checkrealloc(tagged_t *ptr, int decr, int size)
{
  CHECK_FOR_MEMORY_FAULT(tryrealloc(ptr,decr,size));
}

