:- module(bundle, [], [assertions, fsyntax, dcg]).

:- doc(title, "Bundle Support").
:- doc(author, "Jose F. Morales").

:- doc(module, "A bundle is a collection of modules together with
meta-information about namespaces (such as alias paths). This module
defines operations on source bundles and bundle-aware wrappers of
module loading code.

@begin{alert}
DOCUMENTATION MISSING
@end{alert}

@section{Manifest vs. file search paths}

We discurage the direct use of @pred{file_search_path/1} for modifying
the library namespace, except for library names that are manipulated
at runtime. The reason is that there may be circularity problems if
both @tt{:- use_module(foo(bar))} and an alias path definition for
@tt{foo} appears in the code (even in the same module).  The bundle
system solves the problem by moving those paths to a higher meta-level
(the @tt{Manifest.pl} file).

").

:- doc(bug, "Complete documentation").
:- doc(bug, "Merge with bundletool code").
:- doc(bug, "Unify Manifest.pl syntax").

:- use_module(library(terms), [atom_concat/2]).
:- use_module(library(lists), [append/3]).
:- use_module(library(strings), [write_string/1]).
:- use_module(library(streams)).
:- use_module(library(write), [portray_clause/1]).
:- use_module(library(system), [file_exists/1, working_directory/2]).
:- use_module(library(system_extra), [mkpath/1]).
:- use_module(library(process), [process_call/3]).

% ---------------------------------------------------------------------------
% Loading of bundles (just enable namespaces, does not load any module)

:- use_module(library(compiler), [use_module/3]).

:- export(use_bundle/1).
:- export(use_bundle/2).
:- meta_predicate use_bundle(addmodule).
% Use a bundle (from the toplevel)
% TODO: this needs many improvements: config, reload, visibility of the bundle, etc.
use_bundle(BundleDir,This) :-
	ensure_bundle_manifest(BundleDir),
	ensure_bundle_paths(BundleDir),
	F = ~bundle_paths_file(BundleDir),
	compiler:use_module(F,all,This).

% ---------------------------------------------------------------------------
% Loading of modules with automatic bundle load

:- export(use_module/1).
:- export(use_module/2).
:- meta_predicate use_module(addmodule).

use_module(Mod,This) :- bundle:use_module(Mod,all,This).

:- export(use_module/3).
:- meta_predicate use_module(+,addmodule).

% TODO: optimize and remove overheads
use_module(Mod, Imports, This) :-
	% If Mod is a file, load its bundle automatically
	atom(Mod),
	lookup_bundle_root(Mod, BundleDir),
	!,
	use_bundle(BundleDir),
	compiler:use_module(Mod, Imports, This).
use_module(Mod, Imports, This) :-
	compiler:use_module(Mod, Imports, This).

% ---------------------------------------------------------------------------

% TODO: BundleDir should be BundleBase or BundleId

% TODO: see 'bundlectl' script (used in external code)

:- export(bundle_configure/1).
% Configure the bundle
bundle_configure(BundleDir) :-
	ensure_bundle_manifest(BundleDir),
	ensure_bundle_paths(BundleDir),
	BundleName = ~bundle_name(BundleDir),
	message(['Bundle \'', BundleName, '\' successfully configured']).

:- export(bundle_info/1).
% Show bundle info
bundle_info(BundleDir) :-
	BundleDir = ~current_bundle_root,
	ensure_bundle_manifest(BundleDir),
	check_bundle_paths(BundleDir),
	show_info(BundleDir).

:- export(bundle_ciaoc_opts/1).
% Print options for bundle usage in 'ciaoc'
bundle_ciaoc_opts(BundleDir) :-
	ensure_bundle_manifest(BundleDir),
	ciaoc_bundle_opts(BundleDir).

:- export(bundle_ciaoc_mods/1).
% Print extra modules for bundle usage in 'ciaoc' (for executable creation)
bundle_ciaoc_mods(BundleDir) :-
	ensure_bundle_manifest(BundleDir),
	ciaoc_bundle_mods(BundleDir).

:- export(bundle_clean/1).
% Clean the whole bundle
bundle_clean(BundleDir) :-
	ensure_bundle_manifest(BundleDir),
	exec_aux(BundleDir, ['clean']).

:- export(bundle_test/2).
% Perform the bundle tests
bundle_test(BundleDir, Action) :- !,
	ensure_bundle_manifest(BundleDir),
	( test_action(Action) ->
	    true
	; throw(error(['Unknown test action ', Action]))
	),
	exec_aux(BundleDir, [Action]).

:- export(bundle_bench/1).
% Perform the bundle benchmarks
bundle_bench(BundleDir) :-
	ensure_bundle_manifest(BundleDir),
	exec_aux(BundleDir, ['bench']).

test_action(check).
test_action(compare).
test_action(save).
test_action(bench).

exec_aux(BundleDir, Args) :-
	atom_concat([BundleDir, '/bundlectl_aux.sh'], Exec),
	process_call(Exec, Args, [cwd(BundleDir), verbose_error, status(0)]).

% ---------------------------------------------------------------------------

show_info(BundleDir) :-
	info_string(BundleDir, Text, []),
	write_string(Text).

info_string(BundleDir) -->
	{ BundleName = ~bundle_name(BundleDir) },
	{ F = ~bundle_paths_file(BundleDir) },
	"Recipes for '", atom(BundleName), "' bundle usage:\n",
	"\n",
	"  Executable creation (paths file must also be included in executable):\n",
	"    ciaoc -u ", atom(F), " PARAMS ", atom(F), "\n",
	% TODO: Make executables relocatable without recompilation
	"  NOTE: executables cannot currently be relocated without recompilation\n",
	"\n",
	"  Compilation (no executable):\n",
	"    ciaoc -u ", atom(F), " PARAMS\n",
	"\n",
	"  Top-level usage:\n",
	"    ?- ['", atom(F), "'].\n".

% Options for using this bundle in ciaoc (just compiling)
ciaoc_bundle_opts(BundleDir) :-
	ciaoc_bundle_opts_string(BundleDir, Text, []),
	write_string(Text).
	
ciaoc_bundle_opts_string(BundleDir) -->
	{ F = ~bundle_paths_file(BundleDir) },
	"-u ", atom(F).

% Modules for using this bundle in ciaoc (for linking)
ciaoc_bundle_mods(BundleDir) :-
	ciaoc_bundle_mods_string(BundleDir, Text, []),
	write_string(Text).

ciaoc_bundle_mods_string(BundleDir) -->
	{ F = ~bundle_paths_file(BundleDir) },
	atom(F).

% ---------------------------------------------------------------------------

% Default directory for the bundle state directory at the bundle root
bundle_state_dir('.ciao').

% TODO: memoize
bundle_state_base(BundleDir) := D :-
	BundleStateDir = ~bundle_state_dir,
	atom_concat([BundleDir, '/', BundleStateDir], D).

% TODO: memoize
bundle_paths_file(BundleDir) := F :-
	D = ~bundle_state_base(BundleDir),
	M = ~bundle_paths_module(BundleDir),
	atom_concat([D, '/', M], F).

% TODO: memoize
bundle_paths_module(BundleDir) := M :-
	BundleName = ~bundle_name(BundleDir),
	atom_concat(BundleName, '__paths', M).

% bundle_paths file (with .pl suffix)
bundle_paths_file_pl(BundleDir) := F_pl :-
	F_pl = ~atom_concat(~bundle_paths_file(BundleDir), '.pl').

% ---------------------------------------------------------------------------

% Create the bundle state base (if it does not exists)
ensure_bundle_state_dir(BundleDir) :-
	D = ~bundle_state_base(BundleDir),
	mkpath(D).

% ---------------------------------------------------------------------------

:- use_module(library(bundle/bundle_scan),
	[is_bundle_dir/1, locate_manifest_file/2]).
:- use_module(library(pathnames), [path_split/3]).

:- export(current_bundle_root/1).
% Lookup the bundle root for the working directory
current_bundle_root(BundleDir) :-
	working_directory(D, D),
	( lookup_bundle_root(D, BundleDir0) ->
	    BundleDir = BundleDir0
	; throw(error(['Not a bundle \'', D, '\'']))
	).

:- export(lookup_bundle_root/2).
% Detect the bundle root dir for the given File (a directory or normal
% file)
lookup_bundle_root(File, BundleDir) :-
	is_bundle_dir(File),
	!,
	% Found a bundle dir
	BundleDir = File.
lookup_bundle_root(File, BundleDir) :-
	% Not a bundle dir, visit the parent
	path_split(File, Base, _Name),
	Base \== '/', % TODO: is it possible?
	atom_concat(Base0, '/', Base),
	lookup_bundle_root(Base0, BundleDir).

% ---------------------------------------------------------------------------

:- use_module(library(pathnames), [path_concat/3]).
:- use_module(library(aggregates), [findall/3]).

% Ensure that BundleDir contains the bundle_paths file
ensure_bundle_paths(BundleDir) :-
	has_bundle_paths(BundleDir),
	!.
ensure_bundle_paths(BundleDir) :-
	generate_bundle_paths(BundleDir).

% Generate the bundle_paths file for BundleDir
generate_bundle_paths(BundleDir) :-
	ensure_bundle_state_dir(BundleDir),
	F_pl = ~bundle_paths_file_pl(BundleDir),
	open_output(F_pl, Out),
	M = ~bundle_paths_module(BundleDir),
	portray_clause((:- module(M, [], []))),
	portray_clause((:- multifile file_search_path/2)),
	portray_clause((:- dynamic file_search_path/2)),
	portray_clause((:- multifile library_directory/1)),
	portray_clause((:- dynamic library_directory/1)),
	bundle_paths(BundleDir),
	close_output(Out).

% BundleDir contains the bundle_paths file
has_bundle_paths(BundleDir) :-
	ensure_bundle_state_dir(BundleDir),
	F_pl = ~bundle_paths_file_pl(BundleDir),
	file_exists(F_pl),
	!.

% Check that BundleDir contains the bundle_paths file
check_bundle_paths(BundleDir) :-
	has_bundle_paths(BundleDir),
	!.
check_bundle_paths(BundleDir) :-
	BundleName = ~bundle_name(BundleDir),
	throw(error(['Bundle \'', BundleName, '\' is not configured'])).

% Contents of the bundle_paths file (based on the Manifest information for BundleDir)
bundle_paths(BundleDir) :-
	As = ~get_aliases(BundleDir),
	alias_clauses(As, BundleDir).

get_aliases(BundleDir) := As :-
	findall(alias(Alias, RelPath),
                bundle_alias_path(BundleDir, Alias, RelPath),
		As).

alias_clauses([], _BundleDir).
alias_clauses([A|As], BundleDir) :-
	alias_clause(A, BundleDir),
	alias_clauses(As, BundleDir).

alias_clause(alias(Alias, RelPath), BundleDir) :-
	path_concat(BundleDir, RelPath, AbsLibDir),
	( Alias = library ->
	    C = library_directory(AbsLibDir)
	; C = file_search_path(Alias, AbsLibDir)
	),
	portray_clause(C).

% ---------------------------------------------------------------------------
% TODO: Merge with modules in library(bundle/bundle_scan)

check_bundle_dir(BundleDir) :-
	is_bundle_dir(BundleDir), !.
check_bundle_dir(BundleDir) :-
	throw(error(['\'', BundleDir, '\' does not seem to contain a Ciao bundle'])).

% ---------------------------------------------------------------------------
% Database of bundle Manifest

:- use_module(library(read)).

:- data bundle_name/2.
:- data bundle_alias_path/3.

% Ensure that the manifest for BundleDir has been loaded
ensure_bundle_manifest(BundleDir) :-
	bundle_name(BundleDir, _),
	!.
ensure_bundle_manifest(BundleDir) :-
	bundle_read_manifest(BundleDir).

% Load a Manifest.pl file
bundle_read_manifest(BundleDir) :-
	check_bundle_dir(BundleDir),
	locate_manifest_file(BundleDir, ManifestFile),
	%
	open(ManifestFile, read, In),
	read_manifest_(In, BundleDir),
	close(In).

read_manifest_(In, BundleDir) :-
	read(In, T),
	T \== end_of_file,
	!,
	treat_manifest_sent(T, BundleDir),
	read_manifest_(In, BundleDir).
read_manifest_(_In, _BundleDir).

treat_manifest_sent(bundle(N), BundleDir) :- !,
	assertz_fact(bundle_name(BundleDir, N)).
treat_manifest_sent(alias_path(Alias, RelPath), BundleDir) :- !,
	assertz_fact(bundle_alias_path(BundleDir, Alias, RelPath)).
treat_manifest_sent(Sent, BundleDir) :-
	throw(error(['Unknown sentence ', Sent, ' in Manifest at bundle ', BundleDir])).

% ---------------------------------------------------------------------------
% Auxiliary

atom(Atm) --> { atom_codes(Atm, Str) }, string(Str).

string(Str, Xs, Xs0) :- append(Str, Xs0, Xs).
