:- module(bundle_config, [], [assertions, dcg, fsyntax]).

:- doc(title, "Bundle Configuration Values (Persistent)").
:- doc(author, "Jose F. Morales (new version for bundles)").
:- doc(author, "Edison Mera (first version)").

:- use_module(library(system), [file_exists/1]).
:- use_module(library(streams), [open_output/2, close_output/1]).
:- use_module(library(read), [read/2]).

% ===========================================================================
% Configuration values

% config_value_(Name,Bundle,Value)
:- data config_value_/3.

% ===========================================================================
% Get/set configuration values

% Configuration value
:- export(current_config_value/3).
current_config_value(Bundle, Name, Value) :-
	config_value_(Name, Bundle, Value).

% Set a configuration value
:- export(set_bundle_config_value/3).
set_bundle_config_value(Bundle, Name, Value) :-
	retractall_fact(config_value_(Name, Bundle, _)),
	assertz_fact(config_value_(Name, Bundle, Value)).

% Get a configuration value (emit error message if not found)
:- export(bundle_config_value/3).
% TODO: Throw exception if the value is not found...
bundle_config_value(Bundle, Name, Value) :-
	config_value_(Name, Bundle, Value0),
	!,
	Value = Value0.
bundle_config_value(Bundle, Name, _Value) :-
	throw(unknown_bundle_config_variable(Bundle, Name)).

% Clean current config values (useful for fixpoint computation)
:- export(clean_config_values/0).
clean_config_values :-
	retractall_fact(config_value_(_, _, _)).

:- use_module(library(bundletool/bundle_params), [
	root_bundle_name/1, bundle_param_value/3]).

:- export(get_bundle_config_value/3).
% TODO: GIT_REPO_DIR should be an option of the bundle registry, not configuration!
% TODO: Hack for ciaobot, see COMMIT_INFO_OPTS at ciaobot/ciaobot_bundlesrc.sh
%   (specify where the Git repo is stored for this bundle)
get_bundle_config_value(_Bundle, git_repo_dir, V) :- !, % TODO: _Bundle not used
	root_bundle_name(Root),
	( bundle_param_value(Root, 'GIT_REPO_DIR', V) ->
	    true
	; fail
	).

% ===========================================================================
% Restore the bundle configuration values from persistent store (filesystem)

% TODO: Need file locks

% :- use_module(engine(internals), [reload_bundleregs/0]).
:- use_module(library(bundletool/paths_extra), [fsR/2]).

% TODO: use one file per bundle
% TODO: use BuildDirId
bundle_config_file := ~fsR(builddir(build)/'CONFIG_saved').

% Restore the saved configuration (if the file exists)
:- initialization(restore_all_bundle_config).

:- export(restore_all_bundle_config/0).
restore_all_bundle_config :-
	% reload_bundleregs, % TODO: Hack! make initialization explicit and use the right order
	restore_bundle_config.
 
restore_bundle_config :-
	clean_config_values,
	%
	ConfigFile = ~bundle_config_file,
	( file_exists(ConfigFile) ->
	    open(ConfigFile, read, Stream),
	    restore_bundle_config_(Stream),
	    close(Stream)
	; % Do nothing if no configuration exists yet
	  true
	).

restore_bundle_config_(Stream) :-
	read(Stream, R),
	( R = end_of_file ->
	    true
	; nonvar(R), R = config_value(Bundle, Name, Value) ->
%	    display(user_error, read_config_value(Name, Value)), nl(user_error),
	    assertz_fact(config_value_(Name, Bundle, Value)),
	    restore_bundle_config_(Stream)
	; throw(bad_config_entry(R))
	).

:- use_module(library(system), [delete_file/1]).

:- export(reset_all_bundle_config/0).
% Reset current configuration values (useful to ignore saved values)
reset_all_bundle_config :-
	clean_config_values,
	% Delete config file (so that it is not reloaded later)
	ConfigFile = ~bundle_config_file,
	( file_exists(ConfigFile) -> delete_file(ConfigFile) ; true ).

% ===========================================================================
% Save the bundle configuration values to persistent store (filesystem)

:- export(save_bundle_config/0).
save_bundle_config :-
	ConfigFile = ~bundle_config_file,
	open_output(ConfigFile, Output),
	display_config_values,
	close_output(Output).

display_config_values :-
	( % (failure-driven loop)
	  current_config_value(Bundle, Name, Value),
	    displayq(config_value(Bundle, Name, Value)),
	    display('.'),
	    nl,
	    fail
	; true
	).


