:- module(bundle_scan, [], [dcg, assertions]).

:- doc(title, "Scanning and Registering of Bundles").
:- doc(author, "Jose F. Morales").
:- doc(author, "Edison Mera (original author)").

:- doc(module, "Scanning source for bundles and processing of
   @tt{Manifest.pl} files.

   A registered bundle can be viewed as the result of the compilation
   or processing of a @tt{Manifest.pl} (it should contain plain facts
   and no syntactic sugar).

@begin{alert}
  Make sure that this code does not end in executables unless
  necessary (it should not be in the dependencies of usual user
  programs).
@end{alert}").

:- use_module(library(read)).
:- use_module(library(system)).
:- use_module(library(streams)).
:- use_module(library(write)).
:- use_module(library(messages)).

% TODO: be careful with fsR/2 (bundles may not be loaded yet)
:- use_module(library(bundletool/paths_extra), [fsR/2]).
:- use_module(engine(internals), [current_bundle_reg_dir/1]).

:- use_module(library(pathnames), [path_concat/3]).
:- use_module(library(system_extra)).

% ----------------------------------------------------------------------------

% TODO: Duplicated in config_common.pl / bundle_config__ciaode.pl (what would be the right place to put it?)
perms(perms(rwX, rwX, rX)).

:- export(bundle_scan/1).
:- pred bundle_scan(Src) # "Scan all the bundles under the @var{Src}
   directory and annotate the results in the bundle registry (creating
   it if missing)".

bundle_scan(Src) :-
	ensure_current_bundle_reg_dir,
	%
	bundledirs_at_dir(Src, BundleDirs),
	create_bundleregs(BundleDirs, local).

% Make sure that the directory for the bundle database exists
ensure_current_bundle_reg_dir :-
	current_bundle_reg_dir(BundleRegDir),
	perms(Perms),
	mkpath(BundleRegDir, Perms), % TODO: owner?
 	mark_directory(noinstall, BundleRegDir).

% ---------------------------------------------------------------------------

:- use_module(engine(system_info), [ciao_lib_dir/1]).
:- use_module(engine(internals), [bundle_reg_dir/2, bundlereg_filename/3]).
:- use_module(engine(internals), [bundlereg_version/1]).

% Like current_bundle_reg_dir/1, but for the given installation type
% InsType and prefixed with rootprefix if needed.

rootprefix_bundle_reg_dir(InsType, BundleRegDir) :-
	rootprefix_ciao_lib_dir(InsType, BundleRegDirBase),
	bundle_reg_dir(BundleRegDirBase, BundleRegDir).

% TODO: see get_installdir_lib
rootprefix_ciao_lib_dir(InsType, Dir) :-
	( InsType = local ->
	    ciao_lib_dir(Dir)
	; InsType = global ->
	    fsR(rootprefix(installdir_lib(ciao)), Dir)
	; fail
	).

% ---------------------------------------------------------------------------

% List of all bundle directories (absolute path) under @var{Src} (non recursively)
%:- export(bundledirs_at_dir/2).
bundledirs_at_dir(Src, BundleDirs) :-
	directory_files(Src, Files),
	find_bundles(Files, Src, BundleDirs).

find_bundles([],           _,   []).
find_bundles([File|Files], Src, [BundleDir|BundleDirs]) :-
	nonvar(File),
	\+ File = '..',
	\+ File = 'Manifest', % not the manifest directory
	( File = '.' ->
	    C1 = Src
	; path_concat(Src, File, C1)
	),
	\+ directory_has_mark(nocompile, C1),
	is_bundle_dir(C1),
	!,
	BundleDir = C1,
	find_bundles(Files, Src, BundleDirs).
find_bundles([_File|Files], Src, BundleDirs) :-
	find_bundles(Files, Src, BundleDirs).

% ---------------------------------------------------------------------------

:- export(is_bundle_dir/1).
% @var{BundleDir} is the directory for a bundle (i.e., contains a
% Manifest.pl file)
is_bundle_dir(BundleDir) :-
	locate_manifest_file(BundleDir, _).

:- export(locate_manifest_file/2).
% Given @var{BundleDir} a bundle directory, @var{ManifestFile}
% is unified with the name of the manifest file.
locate_manifest_file(BundleDir, ManifestFile) :-
	( BundleDir2 = BundleDir
	; path_concat(BundleDir, 'Manifest', BundleDir2)
	),
	path_concat(BundleDir2, 'Manifest.pl', ManifestFile),
	file_exists(ManifestFile),
	!.

% ---------------------------------------------------------------------------

:- doc(section, "Create/destroy bundleregs").
% A bundlereg contains the processed manifest information and resolved
% paths.

% NOTE: This code uses library(fastrw) so that bundleregs can be
%   efficiently read from internals.pl during the boot process.

:- use_module(library(fastrw), [fast_write/1]).

% TODO: allow relative paths? (with some relocation rules)

create_bundleregs([], _InsType).
create_bundleregs([BundleDir|BundleDirs], InsType) :-
	create_bundlereg(BundleDir, InsType),
	create_bundleregs(BundleDirs, InsType).

:- export(create_bundlereg/2).
% Create a bundlereg for bundle at @var{BundleDir} and installation
% type @var{InsType} (which determines the location of the bundle
% registry and the absolute directories for alias paths).

create_bundlereg(BundleDir, InsType) :-
	% Load Manifest.pl (and optionally version from version/)
	locate_manifest_file(BundleDir, ManifestFile),
	loop_read_file(ManifestFile, ManifestSents0),
	read_bundle_version(BundleDir, ManifestSents0, ManifestSents),
	%
	member_chk(bundle_name(BundleName), BundleDir, ManifestSents),
	%
	bundle_installdir_lib(InsType, BundleName, BundleDir, AliasBase),
	%
	( member_chk(bundle_alias_paths(RelAliasPaths), BundleDir, ManifestSents) ->
	    true
	; RelAliasPaths = []
	),
	%
	abs_alias_paths(RelAliasPaths, AliasBase, AliasPaths),
	%
	bundle_reg_file(InsType, BundleName, RegFile),
	%
	open_output(RegFile, UO),
	%
	% Version of bundlereg
	bundlereg_version(V),
	fast_write(bundlereg_version(V)),
        % Alias paths
	write_alias_paths(AliasPaths, BundleName),
	% TODO: write extra info in a separate file?
	member_chk(bundle_packname(Packname), BundleDir, ManifestSents),
	member_chk(bundle_type(BundleType), BundleDir, ManifestSents),
	fast_write(bundle_id(BundleName)),
	fast_write(bundle_prop(BundleName, packname(Packname))),
	fast_write(bundle_prop(BundleName, kind(BundleType))),
	( member(bundle_version(Version), ManifestSents) ->
	    fast_write(bundle_prop(BundleName, version(Version)))
	; true
	),
	( member(bundle_patch(Patch), ManifestSents) ->
	    fast_write(bundle_prop(BundleName, patch(Patch)))
	; true
	),
	% TODO: relocate BundleDir for InsType=global?
	fast_write(bundle_srcdir(BundleName, BundleDir)),
	%
	close_output(UO),
	perms(Perms),
	set_file_perms(RegFile, Perms).

write_alias_paths([], _Bundle).
write_alias_paths([X|Xs], Bundle) :-
	write_alias_path(X, Bundle),
	write_alias_paths(Xs, Bundle).

write_alias_path(AliasName=AliasPath, Bundle) :- !,
	fast_write(bundle_alias_path(AliasName, Bundle, AliasPath)).
write_alias_path(AliasPath, Bundle) :-
	fast_write(bundle_alias_path(library, Bundle, AliasPath)).

:- export(remove_bundlereg/2).
% Remove the bundlereg for bundle @var{BundleName} and installation
% type @var{InsType} (which determines the location of the bundle
% registry).

remove_bundlereg(BundleName, InsType) :-
	bundle_reg_file(InsType, BundleName, RegFile),
	my_del_file_nofail(RegFile).

% TODO: merge with system_extra:del_file_nofail?
my_del_file_nofail(FileName) :-
	( file_exists(FileName) -> delete_file(FileName) ; true ).

% File is the registry file for the BundleName bundle
bundle_reg_file(InsType, BundleName, RegFile) :-
	% TODO: use rootprefix because we may be installing, prefix outside?
	rootprefix_bundle_reg_dir(InsType, BundleRegDir),
	bundlereg_filename(BundleName, BundleRegDir, RegFile).

% Compute absolute paths names for alias paths (add @var{Base} to the
% beginning of each relative alias path).
abs_alias_paths([], _, []).
abs_alias_paths([RelAliasPath|RelAliasPaths], Base, [AliasPath|AliasPaths]) :-
	abs_alias_path(RelAliasPath, Base, AliasPath),
	abs_alias_paths(RelAliasPaths, Base, AliasPaths).

abs_alias_path(X, Base, Y) :-
	( X = (AliasName=RelAliasPath) -> Y = (AliasName=AliasPath)
	; X = RelAliasPath, Y = AliasPath
	),
	( RelAliasPath = '.' ->
	    AliasPath = Base
	; path_concat(Base, RelAliasPath, AliasPath)
	).

% ---------------------------------------------------------------------------

% TODO: see get_installdir_lib
bundle_installdir_lib(InsType, BundleName, BundleDir, Dir) :-
	( InsType = local ->
	    Dir = BundleDir
	; InsType = global ->
	    fsR(installdir_lib(BundleName), Dir)
	; fail
	).

% ---------------------------------------------------------------------------

% Load a Manifest.pl file
loop_read_file(FileName, Xs) :-
	open(FileName, read, S),
	loop_read(S, Xs),
	close(S).

loop_read(S, [X|Xs]) :-
	read(S, X),
	X \== end_of_file,
	!,
	loop_read(S, Xs).
loop_read(_S, []).

member_chk(Item, BundleDir, Sents) :-
	( member(Item, Sents) -> true
	; warning_message("Item ~w not found in bundle Manifest at ~w",
            [Item, BundleDir])
	).

% ---------------------------------------------------------------------------
% Load bundle version/patch from version/ files if not provided in Manifest
% TODO: Deprecate in the future?

:- use_module(library(system_extra), [file_to_line/2]).

read_bundle_version(_BundleDir, ManifestSents0, ManifestSents) :-
	( member(bundle_version(_), ManifestSents0)
	; member(bundle_patch(_), ManifestSents0)
	),
	!,
	ManifestSents = ManifestSents0.
read_bundle_version(BundleDir, ManifestSents0, ManifestSents) :-
        path_concat(BundleDir, 'version/GlobalVersion', FileV),
        path_concat(BundleDir, 'version/GlobalPatch', FileP),
	( file_to_atom(FileV, Version),
	  file_to_atom(FileP, Patch) ->
	    ManifestSents = [
              bundle_version(Version), bundle_patch(Patch)|ManifestSents0
            ]
	; ManifestSents = ManifestSents0
	).

file_to_atom(File, X) :-
	file_exists(File),
	file_to_line(File, Str),
	atom_codes(X, Str).

% ---------------------------------------------------------------------------
% Directory marks

% TODO: Complete missing cases, implement unmark? interact with SVN?

% TODO: Directory marks (move to a module)
mark_directory(noinstall, Dir) :-
 	% creates empty NOINSTALL file (which marks that this
 	% directory will not be installed)
 	path_concat(Dir, 'NOINSTALL', F),
 	touch(F).

directory_has_mark(nocompile, Dir) :-
	path_concat(Dir, 'NOCOMPILE', F),
	file_exists(F).

