# ===========================================================================
# autoboot.sh: automatic build of bootstrappable executables
#
# Author: Jose F. Morales
# ===========================================================================
#
# Usage: include this code in your script ('.' or 'source')
#
# Requires:
#   CIAODESRC: environment variable pointing to the root of Ciao sources
#   BOOTSTRAP_CIAOC: location of static ciaoc.sta for bootstrap
#   no_bootstrap_error: function invoked when bootstrap is missing
#   failed_bootstrap_error: invoked when bootstrap is could not be built 
#
# Exports:
#   autoboot_autobuild
#   autoboot_build
#   autoboot_exec
# ---------------------------------------------------------------------------

# Imports
. "${CIAODESRC}/ciao/lib/compiler/scripts/messages_aux.sh"
. "${CIAODESRC}/ciao/lib/compiler/scripts/builddir_aux.sh"

# Location of (static) executable for MAINMODULE in BUILDDIR
get_mainexec() { # BUILDDIR MAINMODULE
    echo "${1}/bin/`basename "${2}"`.sta"
}

# Build or check if MAINMODULE is already compiled, and depending on
# the value of AUTOBUILD:
#
#  - AUTOBUILD=no:      prompt an error if it does not exists
#  - AUTOBUILD=yes:     build if it does not exists
#  - AUTOBUILD=rebuild: rebuild (incrementally), even if it exists
#
autoboot_autobuild() { # BUILDDIR MAINMODULE (env: AUTOBUILD)
    MAINEXEC=`get_mainexec "${1}" "${2}"`
    if [ ! x"${AUTOBUILD}" = x"rebuild" ] && \
       [ -x "${MAINEXEC}" ] && \
       builddir_do_engine "${1}" engine_is_ok; then
	# Bootstrap is ready
	true
    else
	if [ x"${AUTOBUILD}" = x"no" ]; then
            # TODO: detect (mark) in builddir if bootstrap
            #   compilation failed or it is just missing
	    no_bootstrap_error
	else # AUTOBUILD=yes or AUTOBUILD=rebuild
	    builddir_configure_boot "${1}"
	    if autoboot_build "${1}" "${2}"; then
		true
	    else
		failed_bootstrap_error
	    fi
	fi
    fi
}

# Build MAINMODULE
autoboot_build() { # BUILDDIR MAINMODULE
    MAINNAME=`basename "${2}"`
    boot_message "Compiling native code (engine) for `basename "${BOOTSTRAP_CIAOC}"` (using '`basename ${1}`' config)..."
    MAINEXEC=`get_mainexec "${1}" "${2}"`
    BOOTSTRAP_C_DIR=`dirname "${BOOTSTRAP_CIAOC}"`
    if builddir_do_engine "${1}" build "${BOOTSTRAP_C_DIR}"; then
	true
    else
	return 1
    fi
    do_exec_header_for_builddir "${1}" # TODO: this should not be necessary...
    #
    boot_message "Compiling bytecode for ${MAINNAME} (using `basename "${BOOTSTRAP_CIAOC}"` and '`basename ${1}`' config)..."
    #
    # Build a static executable using the bootstrap 'ciaoc'.
    # It must be static to avoid bootstrapping problems (the compiler
    # compiling its own libraries).
    #
    # TODO: CIAOCACHEDIR may allow non-static exec here
    #
    builddir_set_local_ciaocopts "${1}"
    mkdir -p "${1}/bin"
    builddir_exec "${1}" "${BOOTSTRAP_CIAOC}" ${LOCAL_CIAOCOPTS} -s -x -o "${MAINEXEC}" "${2}"
    #
    # NOTE: We assume that the builddir for 'ciaoc.sta' (which
    #   contains the engine) is the same than the builddir for this
    #   executable (we share the engine). Sometimes we may need a
    #   different engine.
    #
    boot_message "Reusing engine from `basename "${BOOTSTRAP_CIAOC}"`"
}

# Do autobuild and execute the compiled MAINMODULE
autoboot_exec() { # BUILDDIR MAINMODULE [ARGS]
    BUILDDIR="${1}"
    MAINMODULE="${2}"
    MAINEXEC=`get_mainexec "${1}" "${2}"`
    shift
    shift
    autoboot_autobuild "${BUILDDIR}" "${MAINMODULE}"
    builddir_exec "${BUILDDIR}" "${MAINEXEC}" "$@"
}

