#!/bin/sh
# ===========================================================================
# build_engine.sh: Auxiliary code to build native code in builddir
#
# Author: Jose F. Morales (based on previous code)
# ===========================================================================
#
# Usage: execute as a different sh process (otherwise it will pollute
#   the environment with many local variable definitions)
#
# Requires:
#   CIAODESRC: root of the Ciao system
#   CURRBUILDDIR: current builddir
#   
# ---------------------------------------------------------------------------

# Exit immediately if a simple command exits with a non-zero status
set -e 

# ---------------------------------------------------------------------------
# Imports

. "${CIAODESRC}/ciao/lib/compiler/scripts/messages_aux.sh"
. "${CIAODESRC}/ciao/lib/compiler/scripts/get_commit_info.sh"

# ---------------------------------------------------------------------------

MAKE="make -s SHELL=/bin/bash"

if test x"${CURRBUILDDIR}" = x""; then
    echo "ERROR: CURRBUILDDIR is undefined" >&2
    exit 1
fi

not_configured() {
    [ ! -f "${CURRBUILDDIR}/CONFIG_sh" ]
}

CONFIGLOADED=""
ensure_config_loaded() {
    if [ x"${CONFIGLOADED}" != x"" ]; then
	# Exit, bootstrap settings already loaded
	return
    fi
    if not_configured; then
	echo "bug: no configuration found" >&2
        exit 1
    fi
    # TODO: make sure that it is generated (or exit graciously)
    # TODO: during a installation there exist two different configurations here..
    #       one for the bootstrap system and other for the final one
    . "${CURRBUILDDIR}/CONFIG_sh"

    CONFIGLOADED=yes
}

# ---------------------------------------------------------------------------
# Engine build/bootstrap

# Is there a compiled engine?
engine_is_ok() {
    if not_configured; then
	return 1
    else
	ensure_config_loaded
	[ -f "${BUILD_OBJDIR}/${ENGINENAME}" ]
    fi
}

# (called from 'ciaosetup')
do_build_engine() { # AUTOGEN_C_DIR
    ensure_config_loaded
    do_engine_srcdir_and_links "${1}"
    do_build_engine__2
    do_engine_links
}

# TODO: This command is not symmetric to build_engine! why?
# (called from 'ciaosetup')
do_build_static_engine() { # AUTOGEN_C_DIR
    ensure_config_loaded
    # TODO: Static engine seem to be required for some profiler code
    ENGINE_LINK_TYPE=dostat do_build_engine__2
}

# These links are needed to run ciao from the sources
# Note: The links are a bit trickly because this should work with the ciao header.
do_engine_links() {
    OLDPWD=`pwd`
    cd "${BUILD_OBJDIRROOT}"
    rm -f "${ENGINEBASE}.${CIAOARCH}${EXECSUFFIX}"
    ln -s "${CIAOARCH}${CIAODEBUG}/${ENGINEBASE}" "${ENGINEBASE}.${CIAOARCH}${EXECSUFFIX}"
    rm -f "${ENGINEBASE}.lnk" "${ENGINEBASE}"
    ln -s "${ENGINEBASE}.${CIAOARCH}${EXECSUFFIX}" "${ENGINEBASE}"
    cd "${OLDPWD}"
}

ENGINEBASESTAT="${ENGINEBASE}.sta"
ENGINENAMESTAT="${ENGINEBASESTAT}${EXECSUFFIX}"

MFLAGS="-j${PROCESSORS}"

GMAKEOPTS="--no-print-directory"
GMAKE="${MAKE} ${GMAKEOPTS}"
#GMAKE=${MAKE}

do_gmake() {
    ${GMAKE} "$@"
}

do_build_engine__2() {
    OLDPWD=`pwd`
    cd "${CIAODESRC}/ciao"
    gen_engine_build_info
    umask 002
    case "${ENGINE_LINK_TYPE}" in
	dyn)
#	    normal_message "Building dynamic engine for ${OSNAME}${ARCHNAME}"
	    do_gmake -C "${BUILD_OBJDIR}" configure.h
	    do_gmake_currlibs -C "${BUILD_OBJDIR}" ${MFLAGS} "${ENGINENAME}"
	    ;;
	dostat)
#	    normal_message "Building static engine for ${OSNAME}${ARCHNAME}"
	    # Nasty trick: rename the old engine
	    if [ -f "${BUILD_OBJDIR}/${ENGINENAME}" ] ; then
		mv "${BUILD_OBJDIR}/${ENGINENAME}" "${BUILD_OBJDIR}/${ENGINENAME}.tmp" ;
	    fi
	    do_gmake -C "${BUILD_OBJDIR}" configure.h
	    do_gmake_currlibs_stat -C "${BUILD_OBJDIR}" ${MFLAGS} "${ENGINENAME}"
	    # Nasty trick: rename the engine to ENGINENAMESTAT
	    mv "${BUILD_OBJDIR}/${ENGINENAME}" "${BUILD_OBJDIR}/${ENGINENAMESTAT}"
	    chmod ${ciaode__EXECMODE} "${BUILD_OBJDIR}/${ENGINENAMESTAT}"
	    # Nasty trick: move saved engine to the original place
	    mv "${BUILD_OBJDIR}/${ENGINENAME}.tmp" "${BUILD_OBJDIR}/${ENGINENAME}"
	    ;;
    esac
    cd "${OLDPWD}"
}

#[REMOVED]
# # TODO: Necessary for Windows?
# do_build_libciao() {
#     normal_message "Building 'libciao'"
#     OLDPWD=`pwd`
#     cd ${BUILD_OBJDIR}
#     do_gmake_currlibs libciao
#     do_gmake_currlibs ciaoobject
#     cd ${OLDPWD}
# }

# TODO: for build_static_engine... useful?
# In Windows we need an executable built without console support 
# ifeq (${OSNAME},Win32)
# 	cd ${BUILD_OBJDIR} &&	do_gmake clean
# 	do_gmake ${MFLAGS} CONSOLEFLAG="${NOCONSOLEFLAG}" dostateng
# 	/bin/mv ${BUILD_OBJDIR}/${ENGINAME}.sta ${BUILD_OBJDIR}/${ENGINAME}_nc.sta
# 	cd ${BUILD_OBJDIR} &&	do_gmake clean
# endif

do_gmake_currlibs() {
    # make with CURRLIBS set to LIBS
    # TODO: this could be inside the makefile itself...
    ${GMAKE} "$@" CURRLIBS="${LIBS}"
}

do_gmake_currlibs_stat() {
    # make with CURRLIBS set to LIBS
    # TODO: this could be inside the makefile itself...
    ${GMAKE} "$@" CURRLIBS="${LIBS} ${STAT_LIBS}" ADDOBJ="`statobj`"
}

# TODO: This implementation is incorrect.
#       Not all C libraries should be part of the static engine.
#         random.o sockets_c.o fastrw.o, ...
statobj() {
    find "${CIAODESRC}/ciao" -name "*${CIAOARCH}.o" -o -name "*${CIAOARCH}_glue.o" \
    | tr '\n' ' '
}

do_engine_srcdir_and_links() { # AUTOGEN_C_DIR
    USE_SYMLINKS=yes
    #
    # TODO: USE_SYMLINKS=no could be useful for platforms or
    #       compilation targets where symbolic links are not
    #       appropiate.
    #
    AUTOGEN_C_DIR="${1}"
#    normal_message "Creating engine source directory"
    if [ x"${USE_SYMLINKS}" = x"yes" ]; then
	if [ -e "${BUILD_OBJDIR}"/Makefile ]; then
	    return
	fi
	create_engdir "${BUILD_OBJDIRROOT}" "${BUILD_OBJDIR}"
	OLDPWD=`pwd`
	cd "${BUILD_OBJDIR}"
	ln -s "${ENGSRC_FROM_BUILD_OBJDIR}"/*.[ch] .
	ln -s "${ENGSRC_FROM_BUILD_OBJDIR}"/Makefile .
	ln -s "${ENGSRC_FROM_BUILD_OBJDIR}"/Makefile-configure .
	if [ -d "${AUTOGEN_C_DIR}" ]; then
	    # Getting files generated by emugen
	    ln -s "${AUTOGEN_C_DIR}"/*.[ch] .
	    ln -s "${AUTOGEN_C_DIR}"/Makefile-absmach_def .
	else
	    # TODO: We assume that the missing files has been generated by emugen
	    true
	fi
	rm -f configure.h
	cd "${OLDPWD}"
    else
	create_engdir "${BUILD_OBJDIRROOT}" "${BUILD_OBJDIR}"
	# TODO: This code is not used (and probably wrong)
	OLDPWD=`pwd`
	cd engine
	for File in *.[ch] Makefile Makefile-configure Makefile-absmach_def ; do
	    if [ ! -f "${BUILD_OBJDIR}/${File}" -o "${File}" -nt "${BUILD_OBJDIR}/${File}" ]; then
		rm -f "${BUILD_OBJDIR}/${File}"
		cp "${File}" "${BUILD_OBJDIR}/${File}"
	    fi
	done
	cd "${OLDPWD}"
    fi
    # TODO: check if the directories are correct
#    normal_message "Creating engine include directory"
    create_engdir "${BUILD_HDIRROOT}" "${BUILD_HDIR}"
}

create_engdir() { # DIRROOT DIR
    if test ! -d "${1}"; then
	mkdir "${1}"
	chmod "${ciaode__EXECMODE}" "${1}"
    fi
    if test ! -d "${2}"; then
	mkdir -p "${2}"
	chmod "${ciaode__EXECMODE}" "${2}"
    fi
}

# Writes the input to FILE. If the input is exactly the same than the
# output, the old file is preserved (thus the modification time is not
# affected)
# TODO: Use 'update_file' in other parts of the build process (for some _auto.pl files and configuration options)
update_file() { # FILE
    TMP="${1}-tmp"
    cat > "${TMP}"
    if cmp -s "${1}" "${TMP}"; then
	# identical (keep the original, thus preserving its
	# creation date)
	rm "${TMP}"
    else
	# different, or ${1} does not exist (replace by new file)
	rm -f "${1}"
	mv "${TMP}" "${1}"
	chmod "${ciaode__DATAMODE}" "${1}"
    fi
}

# TODO: Make this part of the configuration process? (not the build
#       process) This would make 'update_file' unnecessary, unless you
#       want to reuse that to detect when the configuration changed in
#       some way. (JFMC)

# Generate the build info for the engine:
#   - versions, gcc options, OS suffixes, etc.

# NOTE: See generated CONFIG_sh for variable definitions (such as ciaode__DEFAULTLIBDIR)
# TODO: Simplify bootstrap engine config by ignoring some defs (or passing them in env vars)?
gen_engine_build_info() {
    get_commit_info # get COMMIT_*

    if test x"${ciao__DEBUG_LEVEL}" = x"nodebug"; then
	SHOW_DEBUG_LEVEL=""
    else
	SHOW_DEBUG_LEVEL=" [${ciao__DEBUG_LEVEL}]"
    fi

    # version.h:
    MAJORVER=`echo "${VERSION}" | cut -f1 -d'.'`
    MINORVER=`echo "${VERSION}" | cut -f2 -d'.'`
    update_file "${BUILD_OBJDIR}/version.h" <<EOF
#define CIAO_MAJOR_VERSION ${MAJORVER}
#define CIAO_MINOR_VERSION ${MINORVER}
#define CIAO_PATCH_VERSION ${PATCH}
#define CIAO_COMMIT_BRANCH "${COMMIT_BRANCH}"
#define CIAO_COMMIT_ID "${COMMIT_ID}"
#define CIAO_COMMIT_DATE "${COMMIT_DATE}"
#define CIAO_COMMIT_DESC "${COMMIT_DESC}"
EOF

    # version.c:
    update_file "${BUILD_OBJDIR}/version.c" <<EOF
char *emulator_version = "Ciao ${COMMIT_DESC} (${COMMIT_DATE}) [${OSNAME}${ARCHNAME}] ${SHOW_DEBUG_LEVEL}";
char *emulator_architecture = "${ARCHNAME}";
char *emulator_os =    "${OSNAME}";
char *emulator_debug = "${CIAODEBUG}";
char *emulator_location = "${ENGLOCATION}";
char *ciao_version =   "${VERSION}";
char *ciao_patch =     "${PATCH}";
char *ciao_commit_branch = "${COMMIT_BRANCH}";
char *ciao_commit_id = "${COMMIT_ID}";
char *ciao_commit_date = "${COMMIT_DATE}";
char *ciao_commit_desc = "${COMMIT_DESC}";
char *ciao_suffix =    "${CIAOSUFFIX}";
char *exec_suffix =    "${EXECSUFFIX}";
char *so_suffix =      "${SOSUFFIX}";
char *default_lib_dir = "${ciaode__DEFAULTLIBDIR}";
char *default_c_headers_dir = "${BUILD_HDIR}";
EOF
}

do_clean_engine() {
    if not_configured; then
	# assume that if there is no configuration, then there is no
	# engine to be clean
	true
    else
	ensure_config_loaded
#    normal_message "Removing engine for all architectures"
	rm -rf "${BUILD_OBJDIRROOT}"
	rm -rf "${BUILD_HDIRROOT}"
    fi
}

case "${1}" in
    build) shift; do_build_engine "${1}" ;;
    build_static) shift; do_build_static_engine "${1}" ;;
    clean) do_clean_engine ;;
    engine_is_ok) engine_is_ok ;;
    *)
	echo "Unknown target '${1}' in build_engine.sh" >&2
	exit 1
	;;
esac
