# ===========================================================================
# builddir_aux.sh: Auxiliary code to manage building areas (builddir)
#
# Author: Jose F. Morales
# ===========================================================================
#
# Usage: include this code in your script ('.' or 'source')
#
# Requires:
#   loaded clean_aux.sh
#   DEFAULT_CIAOLIB: default value for CIAOLIB
#   
# Exports:
#   do_exec_header_for_builddir
#   builddir_configure_boot
#   builddir_do_engine
#   builddir_clean
#   builddir_set_local_ciaocopts
#   builddir_exec
#
# ---------------------------------------------------------------------------

exit_on_error() { # MSG
    echo "ERROR: ${@}" >&2
    exit 1
}

# TODO: create from Prolog; put here just a dummy version for bootstrapping

# TODO: Executables build in this way cannot be moved to machines
#       where Ciao has been installed in other place, unless the CIAO*
#       environment variables are set. Can a nicer solution be found?
# TODO: use bundle_actions__ciao.pl:make_exec_header
do_exec_header_for_builddir() { # CURRBUILDDIR
    CURRBUILDDIR="${1}"
    if [ ! -f "${CURRBUILDDIR}/CONFIG_sh" ] ; then
	exit_on_error "No configuration found at ${CURRBUILDDIR}"
    fi

    # (local)
    BUILD_OBJDIRROOT=`. "${CURRBUILDDIR}/CONFIG_sh"; echo "${BUILD_OBJDIRROOT}"`
    # (local)
    ENGINENAME=`. "${CURRBUILDDIR}/CONFIG_sh"; echo "${ENGINENAME}"`

    # TODO: write in another place (builddir?)
    # (local)
    HEADER="${CIAODESRC}/ciao/lib/compiler/header"

    # Override shell clobbering
    do_delete_file "${HEADER}" # TODO: why?
    # TODO: Find a better way which do not require /bin/sh (just /bin/env)
    #       This can point to a small executable that does this logic with CIAOARCH!
    cat > "${HEADER}" <<EOF
#!/bin/sh
INSTENGINE="${BUILD_OBJDIRROOT}/${ENGINENAME}"\${CIAOARCH:+".\$CIAOARCH"}
ENGINE=\${CIAOENGINE:-\${INSTENGINE}}
exec "\$ENGINE" "\$@" -C -b "\$0"

EOF
}

# Create a default (bootstrap) configuration for the specified CURRBUILDDIR.
#
# This create a CONFIG_mkf and CONFIG_sh (without using an already
# compiled Prolog engine).
builddir_configure_boot() { # CURRBUILDDIR
    CURRBUILDDIR="${1}"
    test x"${CURRBUILDDIR}" = x"" && exit_on_error "No builddir in builddir_configure_boot" # safety check...
    test -d "${CURRBUILDDIR}" || mkdir "${CURRBUILDDIR}"
    # Erase generated configuration files # TODO: necessary?
    builddir_clean "${CURRBUILDDIR}" config
    # TODO: DEBUG_LEVEL=install is wrong! That is not a debug-level.
    cat > "${CURRBUILDDIR}/CONFIG_input" <<EOF
ciaode__INSTYPE=local
ciaode__EXECMODE=775
ciaode__DATAMODE=664
ciaode__INSTALLGROUP=
ciao__USE_THREADS=yes
ciao__AND_PARALLEL_EXECUTION=no
ciao__PAR_BACK=no
ciao__TABLED_EXECUTION=no
ciao__OPTIM_LEVEL=optimized
ciao__DEBUG_LEVEL=install
#
ciaode__CIAODESRC="${CIAODESRC}"
ciaode__DEFAULTLIBDIR="${CIAODESRC}/ciao"
ciao__CUSTOM_CC="${ciao__CUSTOM_CC}"
ciao__CUSTOM_LD="${ciao__CUSTOM_LD}"
ciao__EXTRA_CFLAGS="${ciao__EXTRA_CFLAGS}"
ciao__EXTRA_LDFLAGS="${ciao__EXTRA_LDFLAGS}"
#
CIAOCOPTS="${CIAOCOPTS}"
EOF
    # Create CONFIG_mkf and CONFIG_sh from CONFIG_input
    # TODO: Call duplicated in some bundle_actions__ciao.pl (in Prolog)
    "${CIAODESRC}"/ciao/config-sysdep/config-and-dump.sh "${CURRBUILDDIR}"
}

# TODO: 'all' does not clean bundlereg, is it ok?
builddir_clean() { # CURRBUILDDIR BUILDDIR_ITEM
    CURRBUILDDIR="${1}"
    shift
    test x"${CURRBUILDDIR}" = x"" && exit_on_error "No builddir in builddir_clean" # safety check...
    case ${1} in
	bin)
            # Clean binaries
	    do_remove_dir "${CURRBUILDDIR}"/bin
            # TODO: not here?
	    do_delete_file "${CIAODESRC}"/ciao/lib/compiler/header
	    do_delete_file "${CIAODESRC}"/ciao/lib/compiler/bat_skel
	    do_delete_file "${CIAODESRC}"/ciao/Win32/setup_bat.bat
	    ;;
	pbundle)
	    do_remove_dir "${CURRBUILDDIR}"/pbundle
	    ;;
	bundlereg)
            # TODO: move bundlereg to builddir?
	    # TODO: use bundle_reg_dir/2
	    do_remove_dir "${CIAODESRC}"/ciao/lib/bundlereg__auto
	    ;;
	config)
	    do_delete_file "${CURRBUILDDIR}/CONFIG_input"
	    do_delete_file "${CURRBUILDDIR}/CONFIG_mkf"
	    do_delete_file "${CURRBUILDDIR}/CONFIG_sh"
	    do_delete_file "${CURRBUILDDIR}/CONFIG_saved"
	    do_delete_file "${CURRBUILDDIR}/CONFIG_GSL"
	    ;;
	all)
	    do_remove_dir "${CURRBUILDDIR}"
	    ;;
    esac
}

# Set LOCAL_CIAOCOPTS
builddir_set_local_ciaocopts() { # CURRBUILDDIR
    CURRBUILDDIR="${1}"
    if [ ! -f "${CURRBUILDDIR}/CONFIG_sh" ] ; then
	exit_on_error "No configuration found at ${CURRBUILDDIR}"
    fi
    # (default 'ciaoc' options)
    LOCAL_CIAOCOPTS=`. "${CURRBUILDDIR}/CONFIG_sh"; echo "${CIAOCOPTS}"`
}

# Execute CURREXEC using the given CURRBUILDDIR
builddir_exec() { # CURRBUILDDIR CURREXEC [ARGS]
    CURRBUILDDIR="${1}"
    CURREXEC="${2}"
    shift
    shift

    if [ ! -f "${CURRBUILDDIR}/CONFIG_sh" ] ; then
	exit_on_error "No configuration found at ${CURRBUILDDIR}"
    fi

    # TODO: find a better way to load those variables
    # (local)
    BUILD_HDIR=`. "${CURRBUILDDIR}/CONFIG_sh"; echo "${BUILD_HDIR}"`
    # (local)
    BUILD_OBJDIR=`. "${CURRBUILDDIR}/CONFIG_sh"; echo "${BUILD_OBJDIR}"`
    # (local)
    ENGINENAME=`. "${CURRBUILDDIR}/CONFIG_sh"; echo "${ENGINENAME}"`

    CIAOALIASPATH="" \
	CIAOLIB="${DEFAULT_CIAOLIB}" \
	CIAOHDIR="${BUILD_HDIR}" \
	CIAOENGINE="${BUILD_OBJDIR}/${ENGINENAME}" \
	"${CURREXEC}" "$@"
}

# Invoke the engine build operation on CURRBUILDDIR
builddir_do_engine() { # CURRBUILDDIR
    CURRBUILDDIR="${1}"
    shift
    CIAODESRC="${CIAODESRC}" \
	CURRBUILDDIR="${CURRBUILDDIR}" \
	"${CIAODESRC}"/ciao/lib/compiler/scripts/build_engine.sh "$@"
}


