# ===========================================================================
# get_commit_info.sh: Get commit information from the control version
# system (Git or SVN)
#
# Author: Jose F. Morales
# ===========================================================================
# 
# Usage: include this code in your script ('.' or 'source')
# 
# Requires:
#   CIAODESRC: environment variable pointing to the root of Ciao sources
#
# Exports:
#   get_commit_info
# ---------------------------------------------------------------------------

# TODO: Combine with 'ciaobot' and 'bundletool' code (probably remove from
# this part).
# TODO: 'branch, id, and date' definitions for SVN are missing

get_commit_info() {
    if `which git > /dev/null 2>&1` && \
       [ -d "${CIAODESRC}/.git" ] ; then
	REPO_KIND=git
    elif `which svnversion > /dev/null 2>&1` && \
       [ -d "${CIAODESRC}/.svn" ] ; then
	REPO_KIND=svn
    else
	REPO_KIND=none
    fi

    set +e # allow errors here since 'git' or 'svnversion' may fail

    # COMMIT_ID: Git commit id or SVN revision number (0 if everything else fails)
    # COMMIT_BRANCH: branch name (ignored in SVN at this moment)
    # COMMIT_DATE: Git commit date (ignored for SVN)
    # COMMIT_DESC: human-readable description of the commit (including
    #   version, patch, branch, etc. if available)
    if [ x"${REPO_KIND}" = x"svn" ]; then
	COMMIT_BRANCH=""
	COMMIT_ID="`svnversion "${CIAODESRC}/Manifest" 2>/dev/null`"
	if test x"${COMMIT_ID}" = x"exported"; then
	    COMMIT_ID=""
	fi
	COMMIT_DATE=""
	if [ x"${COMMIT_ID}" = x"" ] ; then
	    COMMIT_DESC="${VERSION}.${PATCH}"
	else
	    COMMIT_DESC="${VERSION}.${PATCH}-${COMMIT_ID}"
	fi
    elif [ x"${REPO_KIND}" = x"git" ]; then
	COMMIT_BRANCH="`cd "${CIAODESRC}"; git rev-parse --abbrev-ref HEAD 2>/dev/null`"
	if [ x"${COMMIT_BRANCH}" = x"HEAD" ] ; then
	    COMMIT_BRANCH="" # Detached HEAD!
	fi
	COMMIT_ID="`cd "${CIAODESRC}"; git log -1 --format='%H' 2>/dev/null`"
	COMMIT_DATE="`cd "${CIAODESRC}"; git log -1 --format='%ci' 2>/dev/null`"
	COMMIT_DESC="`cd "${CIAODESRC}"; git describe --tags 2>/dev/null`"
	if [ x"${COMMIT_DESC}" = x"" ] ; then
	    # Create our own human-readable commit description using the
	    # branch name (this should not be needed if we have proper
	    # tags in our commit graph).
	    COMMIT_ID_SHORT="`cd "${CIAODESRC}"; git log -1 --format='%h' 2>/dev/null`"
	    COMMIT_DESC="${COMMIT_BRANCH}-g${COMMIT_ID_SHORT}"
	fi
        # Fix COMMIT_DESC (this assumes that we may include version and
        # patch in tag and branch names, and removes redundant
        # information)
	case ${COMMIT_DESC} in
	    # Version and patch in COMMIT_DESC, let us assume that is a
	    # particular code release (a commit with a tag, not a branch).
	    v${VERSION}${PATCH}*) COMMIT_DESC="${VERSION}${PATCH}" ;;
	    ${VERSION}${PATCH}*)  COMMIT_DESC="${VERSION}${PATCH}" ;;
	    # Version but not patch in COMMIT_DESC, just use it (it is a
	    # develoment release).
	    v${VERSION}*)
		# (remove 'v' from the string)
		COMMIT_DESC="`portable_substr "${COMMIT_DESC}" 1 ${#COMMIT_DESC}`"
		;;
	    ${VERSION}*)  true ;;
	    # No version info in commit desc, just append it (also a
	    # development release).
	    *) COMMIT_DESC="${VERSION}-${COMMIT_DESC}" ;;
	esac
    else
	COMMIT_BRANCH=""
	COMMIT_ID=""
	COMMIT_DATE=""
	COMMIT_DESC=""
    fi

    # Finally, get from file if empty
    if test x"${COMMIT_BRANCH}" = x""; then
	if test -f "${CURRBUILDDIR}/COMMIT_BRANCH"; then \
	    COMMIT_BRANCH=`cat "${CURRBUILDDIR}/COMMIT_BRANCH"`
	else
	    COMMIT_BRANCH="0"
	fi
    fi
    if test x"${COMMIT_ID}" = x""; then
	if test -f "${CURRBUILDDIR}/COMMIT_ID"; then \
	    COMMIT_ID=`cat "${CURRBUILDDIR}/COMMIT_ID"`
	else
	    COMMIT_ID="0"
	fi
    fi
    if test x"${COMMIT_DATE}" = x""; then
	if test -f "${CURRBUILDDIR}/COMMIT_DATE"; then \
	    COMMIT_DATE=`cat "${CURRBUILDDIR}/COMMIT_DATE"`
	else
	    COMMIT_DATE="0"
	fi
    fi
    if test x"${COMMIT_DESC}" = x""; then
	if test -f "${CURRBUILDDIR}/COMMIT_DESC"; then \
	    COMMIT_DESC=`cat "${CURRBUILDDIR}/COMMIT_DESC"`
	else
	    COMMIT_DESC="0"
	fi
    fi
    set -e
}

# The expression ${V:Offset:Len} is a bashism. Try 'awk' when available.
portable_substr() { # STRING OFFSET LEN
    local STRING
    local OFFSET
    local LEN
    STRING="${1}"
    OFFSET=${2}
    LEN=${3}
    if which awk > /dev/null 2>&1; then
	echo ${STRING} | awk -v offset=${OFFSET} -v len=${LEN} '{ string=substr($0, offset+1, len); print string; }'
    else
	echo "${STRING:${OFFSET}:${LEN}}"
    fi
}

