:- module(_, [], [assertions, fsyntax, hiord]).

:- doc(title,  "Auxiliary Definitions for Bundle Actions").

:- doc(author, "Edison Mera (original author)").
:- doc(author, "Jos@'{e} F. Morales (redesign, new code)").

:- use_module(library(pathnames), [path_concat/3]).

:- use_module(library(bundletool/paths_extra), [fsR/2]).
:- use_module(library(bundletool/messages_aux), [normal_message/2]).
:- use_module(library(bundletool/messages_aux), [bold_message/2]).

% ===========================================================================

% TODO: move to internals.pl?

:- use_module(library(system_extra), [fixed_absolute_file_name/2]).

% Source directory for CiaoDE at boot time (based on ciao_lib_dir/1
% for the bootstrap system).
% TODO: Simplify; add an environment variable during bootstrap?
:- export(root_bundle_source_dir/1).
root_bundle_source_dir(Dir) :-
	ciao_lib_dir(CiaoLibDir),
	atom_concat(CiaoLibDir, '/..', Dir0),
	fixed_absolute_file_name(Dir0, Dir).

% ===========================================================================

:- doc(section, "Invokation of external tools").
% TODO: make verbose messages optional

:- use_module(library(process), [process_call/3]).
:- use_module(library(bundletool/config_common), [
	gmake/1, command_options/1]).

:- export(invoke_gmake/2).
invoke_gmake(Dir, Args) :-
	bold_message("Executing `make' on `~w' with arguments ~w", [Dir, Args]),
	Options = [cwd(Dir) | ~command_options],
	process_call(~gmake, Args, Options).

:- export(invoke_ant/2).
% Compilation of foreign Java code through Apache Ant (http://ant.apache.org/)
% TODO: Make Ant command configurable?
invoke_ant(Dir, Args) :-
	bold_message("Executing `ant' on `~w' with arguments ~w", [Dir, Args]),
	Options = [cwd(Dir) | ~command_options],
	process_call(path(ant), Args, Options).

% ---------------------------------------------------------------------------

:- export(clean_log/2).
% Clean log files
clean_log(Dir, LogName) :-
	del_file_nofail(~fsR(concat_k(ext('.log'), Dir/LogName))),
	del_file_nofail(~fsR(concat_k(ext('.err'), Dir/LogName))).

% ===========================================================================

:- doc(section, "Generation of TAGS (for find-tag emacs command)").

:- use_module(library(system), [copy_file/2, file_exists/1]).
:- use_module(library(source_tree), [current_file_find/3]).
:- use_module(library(aggregates), [findall/3]).

% TAGS file for the given Bundle
tags_file(Bundle) := ~fsR(bundle_src(Bundle)/'TAGS').

% Creation of TAGS for use with the find-tag command (ESC-.) in emacs
:- export(create_tags_file/1).
create_tags_file(Bundle) :-
	TagsFile = ~tags_file(Bundle),
	del_file_nofail(TagsFile),
	% Combine the TAGS file with that of other bundles
	% TODO: why is necessary? should this be part of the emacs mode?
	% TODO: generalize to follow bundle dependencies?
	( Bundle = ciaopp ->
	    % (for Ciaopp, copy tags from Ciao)
	    CiaoTags = ~tags_file(ciao),
            ( file_exists(CiaoTags) -> (-(copy_file(CiaoTags, TagsFile))) ; true )
        ; true
        ),
	%
 	BundleSrc = ~fsR(bundle_src(Bundle)),
	path_concat(BundleSrc, '', BaseDir),
	findall(FileName,
	        current_file_find([untainted,
		                   srctype([module,package,include])],
		                  BaseDir, FileName),
		List),
	etags(List, TagsFile).

:- export(delete_tags_file/1).
delete_tags_file(Bundle) :-
	del_file_nofail(~tags_file(Bundle)).

etags([], _TagFile).
etags([AbsFile|AbsFileList], TagFile) :-
	etag(AbsFile, TagFile),
	etags(AbsFileList, TagFile).

etag(AbsFile, TagFile) :-
	process_call(path(etags),
	       ['-a', '-l', prolog, AbsFile, '-o', TagFile],
	       [verbose_error, status(_)]).

% ===========================================================================

:- use_module(library(system_extra), [mkpath/2]).
:- use_module(library(bundletool/config_common), [perms/1]).

:- export(ensure_build_doc_dirs/0).
% Prepare the build subdirectory where docs are placed during the
% build process.
ensure_build_doc_dirs :-
 	mkpath(~fsR(builddir_doc(build)), ~perms). % owner?

:- export(ensure_builddir/0).
% Prepare the build directory
ensure_builddir :-
 	mkpath(~fsR(builddir(build)), ~perms). % owner?

% ===========================================================================

:- doc(section, "Operations on Build/Installation Files").

:- use_module(library(system), [copy_file/3]).
:- use_module(library(system_extra),
	[del_file_nofail/1, '-'/1, '--'/1]).
:- use_module(library(source_tree), [remove_dir/1]).

% ---------------------------------------------------------------------------

:- doc(subsection, "Auxiliary for build and installation").

:- export(b_copy/4).
b_copy(Kind, Bundle, FromDir, File) :-
	copy_file(~fsR(FromDir/File),
	          ~fsR(concat_verk(Bundle, Kind, builddir_bin(build)/(File))),
		  [overwrite]).

:- export(b_filebuild/4).
b_filebuild(Kind, Bundle, Base, FileBuild) :-
	FileBuild = ~fsR(concat_verk(Bundle, Kind, builddir_bin(build)/(Base))).

:- export(b_link/3).
% 'File' will point to 'File-Ver'
b_link(Kind, Bundle, File) :-
	b_link_as(Kind, Bundle, File, File).

:- export(b_link_as/4).
% 'Dest' will point to 'Src-Ver'
b_link_as(Kind, Bundle, Src, Dest) :-
	From = ~fsR(concat_verk(Bundle, Kind, Src)),
	% TODO: previously Kind was ignored here... verify that it works (i.e., in Win)
	To = ~fsR(concat_k(Kind, builddir_bin(build)/(Dest))),
	create_link(From, To).

% TODO: If installation type is 'src', b_clean_* will also remove the
%       binary files! Are we sure about it?

:- export(b_clean_link/2).
b_clean_link(Kind, File) :-
	del_file_nofail(~fsR(concat_k(Kind, builddir_bin(build)/(File)))).

:- export(b_clean_copy/3).
b_clean_copy(Kind, Bundle, File) :-
	del_file_nofail(~fsR(concat_verk(Bundle, Kind, builddir_bin(build)/(File)))).

% ---------------------------------------------------------------------------

:- doc(subsection, "Lib Installation").

:- use_module(library(bundletool/config_common),
	[installdirbase_lib/2, relinstalldir_lib/2]).

:- export(rl_install_copy_and_link/2).
rl_install_copy_and_link(FromDir, File) :-
	rl_install_copy(FromDir, File),
	rl_install_link(File).

% Install a copy of File in ~fsR(rootprefix(installdir_lib(ciao)))
% TODO: use bundle
:- export(rl_install_copy/2).
rl_install_copy(FromDir, File) :-
	From = ~fsR(FromDir/File),
	To = ~fsR(rootprefix(installdir_lib(ciao)/(File))),
	-copy_file(From, To, [overwrite]),
	-set_file_perms(To, ~perms).

% Install a link to File in rootprefix(~installdirbase_lib(Bundle))
% TODO: use bundle
:- export(rl_install_link/1).
rl_install_link(File) :-
	rl_install_link_as(File, File).

% Install a link to Src as Dest in rootprefix(~installdirbase_lib(Bundle))
% TODO: see lpdoc symlink, it seems the same thing
% TODO: This is just for a symbolic link; find a better way (e.g.,
%       computing the difference of absolute path names)
:- export(rl_install_link_as/2).
% TODO: use bundle
rl_install_link_as(Src, Dest) :-
	Bundle = ciao,
	Rel = ~relinstalldir_lib(Bundle),
	( Rel = '' ->
	    From = Src
	; From = ~fsR(Rel/Src)
	),
	To = ~fsR(rootprefix(~installdirbase_lib(Bundle)/(Dest))),
	create_link(From, To).

:- export(rl_uninstall_link/1).
% TODO: use bundle
rl_uninstall_link(File) :-
	b_uninstall_file(~installdirbase_lib(ciao)/(File)).

:- export(rl_uninstall_copy/1).
rl_uninstall_copy(File) :-
	b_uninstall_file(~fsR(installdir_lib(ciao)/(File))).

:- export(rl_uninstall_copy_and_link/1).
rl_uninstall_copy_and_link(File) :-
	rl_uninstall_link(File),
	rl_uninstall_copy(File).

:- export(r_install_as/2).
r_install_as(File0, Dest0) :-
	File = ~fsR(File0),
	Dest = ~fsR(rootprefix(Dest0)),
	-copy_file(File, Dest, [overwrite]),
	-set_file_perms(Dest, ~perms).

:- export(r_abslink_as/2).
r_abslink_as(Src0, Dest0) :-
	% TODO: remove line if correct, share
%	Src = ~fsR(rootprefix(Src0)),
	Src = ~fsR(Src0),
	Dest = ~fsR(rootprefix(Dest0)),
	create_link(Src, Dest).

% (shared)
%:- export(create_link/2).
create_link(From, To) :-
	del_file_nofail(To),
	--copy_file(From, To, [overwrite, symlink]).
        % TODO: do not set perms on a symbolic link (the source may
        %       not exist, at it happens in RPM generation)
%	-set_file_perms(To, ~perms).

% ---------------------------------------------------------------------------

:- doc(subsection, "Installation from the Build Staging Area").

:- use_module(library(system), [delete_directory/1]).
:- use_module(library(source_tree), [copy_file_tree/4]).
:- use_module(library(messages), [show_message/3]).

% Creates a directory in the installation area
:- export(b_install_dir/1).
b_install_dir(Dir0) :-
	Dir = ~fsR(rootprefix(Dir0)),
	( mkpath(Dir, ~perms) -> % TODO: owner?
	    true
	; show_message(error, "Could not create ~w", [Dir]),
	  fail
	).

% (copy all)
:- export(b_install_dir_rec/2).
b_install_dir_rec(FromDir, ToDir) :-
	copy_file_tree(installable_precomp(full),
	               ~fsR(FromDir), ~fsR(rootprefix(ToDir)), ~perms).

% (copy all except .po and .itf)
:- export(b_installsrc_dir_rec/2).
b_installsrc_dir_rec(FromDir, ToDir) :-
	copy_file_tree(installable_precomp(src),
	               ~fsR(FromDir), ~fsR(rootprefix(ToDir)), ~perms).

:- export(b_install_copy_and_link/3).
b_install_copy_and_link(Kind, Bundle, File) :-
	b_install_copy(Kind, Bundle, File),
	b_install_link_as(Kind, Bundle, File, File).

:- export(b_install_copy/3).
b_install_copy(Kind, Bundle, File) :-
	From = ~fsR(concat_verk(Bundle, Kind, builddir_bin(build)/(File))),
	To = ~fsR(rootprefix(concat_verk(Bundle, Kind, installdir_bin(build)/(File)))),
	b_install_dir(~fsR(installdir_bin(build))),
	del_file_nofail(To),
	copy_file(From, To, [overwrite]),
	-set_exec_perms(To, ~perms).

:- export(b_install_file_exec/2).
b_install_file_exec(From0, To0) :-
	From = ~fsR(From0),
	To = ~fsR(rootprefix(To0)),
	del_file_nofail(To),
	copy_file(From, To, [overwrite]),
	-set_exec_perms(To, ~perms).

:- export(b_install_file_noexec/2).
b_install_file_noexec(From0, To0) :- % TODO: add Kind...
	From = ~fsR(From0),
	To = ~fsR(rootprefix(To0)),
	del_file_nofail(To),
	copy_file(From, To, [overwrite]),
	-set_file_perms(To, ~perms).

:- export(b_install_link_as/4).
b_install_link_as(Kind, Bundle, Src, Dest) :-
	From = ~fsR(concat_verk(Bundle, Kind, Src)),
	% TODO: previously Kind was ignored in @var{To}, verify that it works (i.e., in Win)
	To = ~fsR(rootprefix(concat_k(Kind, installdir_bin(build)/(Dest)))),
	create_link(From, To).

:- export(b_install_file_link_as/2).
b_install_file_link_as(From0, To0) :-
	From = ~fsR(From0),
	To = ~fsR(rootprefix(To0)),
	create_link(From, To).

:- export(b_uninstall_link/2).
b_uninstall_link(Kind, File) :-
	b_uninstall_file(concat_k(Kind, installdir_bin(build)/(File))).

:- export(b_uninstall_copy/3).
b_uninstall_copy(Kind, Bundle, File) :-
	b_uninstall_file(concat_verk(Bundle, Kind, installdir_bin(build)/(File))).

:- export(b_uninstall_copy_and_link/3).
b_uninstall_copy_and_link(Kind, Bundle, File) :-
	b_uninstall_link(Kind, File),
	b_uninstall_copy(Kind, Bundle, File).

:- export(b_uninstall_file/1).
b_uninstall_file(File) :-
	del_file_nofail(~fsR(rootprefix(File))).

:- export(b_uninstall_dir_rec/1).
b_uninstall_dir_rec(Dir) :-
	% TODO: Add some sanity check here to avoid filesystem havoc
	remove_dir(~fsR(rootprefix(Dir))).

:- export(b_uninstallsrc_dir_rec/1).
b_uninstallsrc_dir_rec(Dir) :-
	% TODO: Add some sanity check here to avoid filesystem havoc
	remove_dir(~fsR(rootprefix(Dir))).

:- export(b_uninstall_dir/1).
b_uninstall_dir(Dir) :-
	-delete_directory(~fsR(rootprefix(Dir))).

:- export(b_uninstall_dir_if_empty/1).
b_uninstall_dir_if_empty(Dir) :-
	--delete_directory(~fsR(rootprefix(Dir))).

% ---------------------------------------------------------------------------

:- doc(subsection, "(Un)Install standalone utilities").

% TODO: show the same kind of messages that are used when compiling libraries
:- export(install_standalone_list/2).
install_standalone_list(Bundle, Ps) :-
	( % (failure-driven loop)
	  member(P0, Ps),
	    install_standalone(Bundle, P0),
	    fail
	; true
	).

install_standalone(Bundle, P0) :-
	n_and_props(P0, P, Props),
	n_name(Props, Name),
	n_kind(Props, K),
	n_output(P, Props, Output),
	normal_message("Installing '~w' (~s)", [Output, Name]),
	b_install_copy_and_link(K, Bundle, P).

% TODO: show the same kind of messages that are used when compiling libraries
:- export(uninstall_standalone_list/2).
uninstall_standalone_list(Bundle, Ps) :-
	( % (failure-driven loop)
	  member(P0, Ps),
	    uninstall_standalone(Bundle, P0),
	    fail
	; true
	).

uninstall_standalone(Bundle, P0) :-
	n_and_props(P0, P, Props),
	n_name(Props, Name),
	n_kind(Props, K),
	n_output(P, Props, Output),
	normal_message("Uninstalling '~w' (~s)", [Output, Name]),
	b_uninstall_copy_and_link(K, Bundle, P).

% ---------------------------------------------------------------------------
% Descriptions for standalone targets
% TODO: improve

:- export(n_and_props/3).
n_and_props(P0, P, Props) :-
	( P0 = P-Props -> true ; P = P0, Props = [] ).

:- export(n_name/2).
n_name(Props, Name) :-
	( member(name=Name, Props) ->
	    true
	; Name = "stand-alone utility" % default name
	).

:- export(n_output/3).
n_output(P, Props, Name) :-
	( member(output=Name0, Props) ->
	    Name = Name0
	; Name = P
	).

n_kind(Props, Kind) :-
	( member(kind=Kind, Props) -> true ; Kind=plexe ).

% ===========================================================================

:- doc(section, "Instantiating Template Files with Parameters").

:- use_module(library(file_utils), [file_to_string/3]).
:- use_module(library(system_extra),
	[replace_params_in_file/3, del_file_nofail/1, '-'/1]).
:- use_module(library(system_extra), [set_file_perms/2, set_exec_perms/2]).

% (like write_template, but load in a string)
:- export(load_template/4).
load_template(File, Repl, String, String0) :-
	Temp = ~fsR(concat_k(ext('.tmp'), File)), % TODO: Use a temporary file
	replace_params_in_file(
               Repl,
	       ~fsR(concat_k(ext('.skel'), File)),
	       Temp),
	file_to_string(Temp, String, String0),
	del_file_nofail(Temp).

:- export(write_template/2).
write_template(File, Repl) :-
	replace_params_in_file(Repl,
	    ~fsR(concat_k(ext('.skel'), File)),
	    ~fsR(File)),
	set_file_perms(~fsR(File), ~perms).

:- export(wr_template/4).
% TODO: reorder the arguments of this predicate
wr_template(cwd, Dir, File, Repl) :- % TODO: correct?
	replace_params_in_file(Repl,
	    ~fsR(concat_k(ext('.skel'), Dir/File)),
	    File).
wr_template(at(OutDir), Dir, File, Repl) :-
	replace_params_in_file(Repl,
	    ~fsR(concat_k(ext('.skel'), Dir/File)),
	    ~fsR(OutDir/File)).
wr_template(origin, Dir, File, Repl) :-
	replace_params_in_file(Repl,
	    ~fsR(concat_k(ext('.skel'), Dir/File)),
	    ~fsR(Dir/File)).
wr_template(kver(Kind, Bundle), Dir, File, Repl) :-
	replace_params_in_file(Repl,
	    ~fsR(concat_k(ext('.skel'), Dir/File)),
	    ~fsR(concat_verk(Bundle, Kind, builddir_bin(build)/(File)))),
	( kind_exec_perms(Kind) ->
	    -set_exec_perms(~fsR(concat_verk(Bundle, Kind, builddir_bin(build)/(File))), ~perms)
	; true
	).
wr_template(inslib(Bundle), Dir, File, Repl) :-
	replace_params_in_file(Repl,
	    ~fsR(concat_k(ext('.skel'), Dir/File)),
	    ~fsR(rootprefix(installdir_lib(Bundle)/File))),
	-set_file_perms(~fsR(rootprefix(installdir_lib(Bundle)/File)), ~perms).

kind_exec_perms(shscript).

% ===========================================================================

:- use_module(engine(internals), ['$bundle_prop'/2]).	
:- use_module(library(bundle/bundle_info),
	[bundle_version/2, bundle_patch/2]).

:- use_module(library(terms), [atom_concat/2]).
:- use_module(library(format), [format/3]).
:- use_module(library(write), [portray_clause/2]).
:- use_module(library(system), [file_exists/1]).
:- use_module(library(system_extra), [datime_string/1]).

:- export(generate_version_auto/2).
% Create a file (File) with a version/1 fact indicating the current
% version of Bundle, build time, and compiler version.

% TODO: generalize for all bundles; change modiftime only if there are changes
%       move, include other options (for runtime)?

generate_version_auto(_Bundle, File) :-
	% TODO: Update file if version changed?
	file_exists(File),
	!.
generate_version_auto(Bundle, File) :-
	Version = ~bundle_version(Bundle),
	Patch = ~bundle_patch(Bundle),
	atom_codes(Date, ~datime_string),
	%
	'$bundle_prop'(ciao, packname(Ciao)),
	CVersion = ~bundle_version(ciao),
	CPatch = ~bundle_patch(ciao),
	%
	VersionAtm = ~atom_concat([
	  Version, '.', Patch, ': ', Date, ' (compiled with ', Ciao, ' ', CVersion, '.', CPatch, ')'
        ]),
	open(File, write, O),
	format(O, "%% Do not edit - automatically generated!\n", []),
        portray_clause(O, version(VersionAtm)),
	close(O),
	-set_file_perms(File, ~perms).

% ===========================================================================

:- export(stop_command_options/2).
% TODO: Document and rename if necessary
% (REALLY EXPORTED)
stop_command_options(yes, []).
stop_command_options(no,  [verbose_error, status(_)]).



