:- module(bundle_configure, [], [assertions, basicmodes, fsyntax, hiord]).

:- doc(title, "Bundle Configuration Process").
:- doc(author, "Jose F. Morales (major rewrite)").
:- doc(author, "Edison Mera (original author)").

:- doc(module, "This module computes bundle configurations values
   (@lib{bundle_config}) based on bundle parameters
   (@lib{bundle_params}) and configuration rules
   (@tt{bundle_config__*} files).

   The configuration process may invoke external configuration tools
   if required, including the @tt{config-sysdep/config-and-dump.sh}
   script.").

:- use_module(library(lists)).
:- use_module(library(dynamic)).
:- use_module(library(process), [process_call/3]).
:- use_module(library(messages)).
:- use_module(library(bundletool/messages_aux), [bold_message/1]).

% ---------------------------------------------------------------------------

% Configuration parameters
:- use_module(library(bundletool/bundle_params),
	[root_bundle_name/1,
	 bundle_param_value/3,
	 normalize_param_name/2, toupper/2, tolower/2]).

% Configuration values (persistent)
:- use_module(library(bundle/bundle_config), [
	clean_config_values/0,
	current_config_value/3,
	set_bundle_config_value/3,
	save_bundle_config/0]).

% ---------------------------------------------------------------------------

:- include(library(bundleconfig/bundleconfig_defs)).

:- use_module(library(lists), [append/3]).

param_def(Bundle, Name, X) :-
	m_bundle_config_entry(Bundle, Name, Def),
	member(X, Def),
	!.

% TODO: Document bundle configuration parameters. Document
%   config_prolog_flag/3 special case.

% TODO: Split default value in a different kind of rules?
%
%  rule_set_value ---> there is a unique value
%  rule_default   ---> there are multiple values
param_body(Bundle, Name,
	   NeededIfGoal,
	   SetValueGoal, Value,
	   DefaultValueGoal, DefaultValue) :-
	m_bundle_config_entry(Bundle, Name, Def0),
	!,
	% expand config entry for prolog_flags
	( member(config_prolog_flag(_FlagNameL, Values, Default), Def0) ->
%	    toupper(FlagNameL, FlagName),
	    ( list(Values) ->
	        NDef = [valid_values(Values)|NDef0]
	    ; NDef = NDef0
	    ),
	    NDef0 = [rule_default(Default)
		     % interactive("Set the given Prolog Flag.", [extended])
		     |NDef1],
	    NDef1 = [],
	    append(Def0, NDef, Def)
	; Def = Def0
	),
	%
	( member(needed_if(NeededIfGoal), Def) ->
	    true
	; NeededIfGoal = true
	),
	%
	( member(rule_set_value(Value), Def) ->
	    SetValueGoal = true % TODO: not used?
	; member(rule_set_value(Value, SetValueGoal), Def) ->
	    true
	; SetValueGoal = fail
	),
	%
	( member(rule_default(DefaultValue), Def) ->
	    DefaultValueGoal = true
	; member(rule_default(DefaultValue, DefaultValueGoal), Def) ->
	    true
	; DefaultValueGoal = fail
	).
	  
% ---------------------------------------------------------------------------

:- use_module(library(bundletool/paths_extra), [fsR/2]).
:- use_module(library(bundletool/interactive_aux)).

% Be careful with get_vers and get_patch: when uninstalling, the patch
% version may differ with the version that we try to uninstall.

% ----------------------------------------------------------------------------

get_config_level(Name, ConfigLevel) :-
	root_bundle_name(Root),
	bundle_param_value(Root, 'INTERACTIVE_CONFIG', 'true'),
	( Name = 'INTERACTIVE_LEVEL' ->
	    NConfigLevel = '3' % (for self-configure this option)
	; current_config_value(Root, 'INTERACTIVE_LEVEL', NConfigLevel)
	),
	!,
	configlevel(NConfigLevel, ConfigLevel).
get_config_level(_Name, default).

% TODO: document, this seems to be the level for *interactive* configuration
configlevel('1', default). % TODO: This is not an interactive configuration!!!
configlevel('2', minimum).
configlevel('3', extended).

% ---------------------------------------------------------------------------

% The full configuration is stored in ~fsR(builddir(build)), in different formats
% for different tools:
%
%   - CONFIG_saved: internal format (saved, no sysdep)
%   - CONFIG_mkf: makefile format, includes sysdep
%   - CONFIG_sh: SH format, includes sysdep
%   - CONFIG_input: input for sysdep configuration script (a
%       projection of CONFIG_saved in sh format)

:- export(config_noscan/0).
:- pred config_noscan # "Configure the system. If
   @tt{INTERACTIVE_CONFIG} is specified, the process is
   interactive. Otherwise, values for configuration are detected
   automatically.".

config_noscan :-
	% First part: materialize configuration values (based on user
	% preferences)
	check_bundle_params, % (for user prefs)
	eval_config_rules, % (can be interactive)
	save_bundle_config, % (save CONFIG_save)
	% TODO: Move to the engine bundle
	% Second part: sysdep configuration for low-level foreign code
	config_sysdep_and_dump,
	%
	bold_message("Configuration complete").

% ---------------------------------------------------------------------------
% Execute sysdep configuration step (low-level foreign code) 

% NOTE: This step is also done during bootstrap compilation, but with
%   default values for CONFIG_input

% Precondition: the configuration values are loaded
config_sysdep_and_dump :-
	BuildDir = ~fsR(builddir(build)),
	CfgInput = ~fsR(builddir(build)/'CONFIG_input'),
	export_config_as_sh(CfgInput),
	process_call('./config-sysdep/config-and-dump.sh',
	       [BuildDir],
	       [cwd(~fsR(bundle_src(ciao)))]).

:- use_module(library(file_utils), [string_to_file/2]).
:- use_module(library(aggregates), [findall/3]).

export_config_as_sh(FileName) :-
	Text = ~list_concat([
	    "# Warning: This file has been created automatically\n\n",
	    ~options_to_sh_string]),
	string_to_file(Text, FileName).

options_to_sh_string(String) :-
	findall(Line, option_to_sh_assign(Line), L),
	list_concat(L, String).

% (nondet)
option_to_sh_assign(Line) :-
	current_config_value(Bundle, Name, Value),
	  atom_codes(Bundle, BundleS),
	  atom_codes(Name, NameS),
	  atom_codes(Value, ValueS),
	  list_concat([BundleS, "__", NameS, "=\"", ValueS, "\"\n"], Line).

% ---------------------------------------------------------------------------
% Verify that all user provided options (from the command-line)
% corresponds to configuration settings (m_bundle_config_entry/3)

check_bundle_params :-
	( bundle_param_value(Bundle, Name, _),
	    ( m_bundle_config_entry(Bundle, Name, _ParamDef) ->
		true
	    ; show_message(warning, "Invalid configuration parameter `~w'", [Name])
	    ),
	    fail
	; true
	).

% ---------------------------------------------------------------------------
% Display the configuration options (for the command-line interface)

:- export(config_list_options/1).
% TODO: use Bundle
config_list_options(Bundle) :-
	display_list(['Available configuration options for bundle `', Bundle, '\'\n\n']),
	config_list_custom_options(Bundle),
	display_list(['\n',
		      'Use --describe-option OPT for help about a particular option.\n',
		      '\n']).

config_list_custom_options(Bundle) :-
	( m_bundle_config_entry(Bundle, Name, _),
	    \+ param_def(Bundle, Name, hidden),
	    show_config_opt(Bundle, Name),
	    fail
	; true
	).

show_config_opt(Bundle, Name) :-
	tolower(Name, Lower),
	( param_def(Bundle, Name, valid_values(ValidValues)) ->
	    true
	; ValidValues = 'VALUE'
	),
	display_list(['  ', '--', Bundle, ':', Lower, '=', ValidValues, '\n']).

:- export(config_describe_option/1).
config_describe_option(Bundle) :-
	root_bundle_name(Root),
	( bundle_param_value(Root, 'DESCRIBE_OPTION', Opt) ->
	    true
	; throw(bug_in_config_describe_option)
	),
	toupper(Opt, Name),
	%
	( m_bundle_config_entry(Bundle, Name, _),
	  \+ param_def(Bundle, Name, hidden) ->
	    describe_config_opt(Bundle, Name)
	; display_list(['Unknown option `', Bundle, ':', Opt, '\'\n\n'])
	).

describe_config_opt(Bundle, Name) :-
	show_config_opt(Bundle, Name),
	nl,
	param_help(Bundle, Name, Help),
	display_string(Help), nl.

% Extract help text from the interactive message or comment.
param_help(Bundle, Name, Help) :-
	( param_def(Bundle, Name, interactive(_ConfigLevels, Help0)) ->
	    Help = Help0
	; param_def(Bundle, Name, comment(Comment)) ->
	    Help = Comment
	; Help = "(No description available)"
	).

% ===========================================================================

% TODO: Define more precisely what is the semantics of configuration
%   parameters TODO: allow structures in own_make targets (fix spawn)
% TODO: access config values in condcomp conditions (or export them to
%   condcomp declarations)

% Determine the values for the configuration entries
% (automatically or via user interaction, depending on settings
% INTERACTIVE_CONFIG)
eval_config_rules :-
	reset_config_fixpo, % (just in case)
	catch(eval_config_rules_, E0, E=yes(E0)),
	reset_config_fixpo,
	( nonvar(E), E = yes(E0) -> throw(E0) ; true ).

eval_config_rules_ :-
	save_config_values,
	clean_config_values,
	eval_config_level,
	( % (failure-driven loop)
	  m_bundle_config_entry(Bundle, Name, _),
	    eval_config_rule(Bundle, Name, []),
	    fail
	; true
	).

% Evaluate rule for configuration level
eval_config_level :-
	root_bundle_name(Root),
	eval_config_rule(Root, 'INTERACTIVE_LEVEL', []).

% prev_config_value(Name, Bundle, Value).
:- data prev_config_value/3.

% Save current config values in prev_config_value 
save_config_values :-
	( % (failure-driven loop)
	  current_config_value(Bundle, Name, Value),
	    assertz_fact(prev_config_value(Name, Bundle, Value)),
	    fail
	; true
	).

reset_config_fixpo :-
	retractall_fact(prev_config_value(_, _, _)).

% TODO: fix 'rule_default' and 'rule_set_value':
%  - allow 'undefined'?
%  - rule_default(_) is used for two different purposes (default values and
%    possible values)
%  - rule_set_value(_) does not allow interactive (i.e., only one possible value)
%  - rule_default(_) allow interactive (i.e., many possible values)

% Evaluate config rule for config value Name at Bundle.
eval_config_rule(Bundle, Name, _Seen) :-
	current_config_value(Bundle, Name, _),
	!,
	% Already computed
	true.
eval_config_rule(_Bundle, Name, Seen) :-
	member(Name, Seen),
	!,
	% Loop in rule evaluation
	throw(loop_in_bundle_config_rule(Name, Seen)).
eval_config_rule(Bundle, Name, Seen) :-
	( m_bundle_config_entry(Bundle, Name, _) ->
	    true
	; throw(undefined_bundle_config_parameter(Bundle, Name))
	),
	% Get rule body
	param_body(Bundle, Name,
	           NeededIfGoal,
		   SetValueGoal, Value,
		   DefaultValueGoal, DefaultValue),
	% Get config value domain
	( param_def(Bundle, Name, valid_values(ValidValues)) ->
	    true
	; true
	),
	% Get configuration value:
	%
	% (1) from a bundle parameter
	% TODO: use as input instead?
	( bundle_param_value(Bundle, Name, Value) ->
	    % TODO: Add option in bundle_param for this?
	    Explanation = parameter
	% (2) unique possible value, from SetValueGoal
	; eval_config_deps(NeededIfGoal, Name, Seen, Bundle),
	  eval_config_deps(SetValueGoal, Name, Seen, Bundle),
	  eval_config_goal(SetValueGoal, Bundle) ->
	    Explanation = propagated % (no other possible choice)
	% (3) choose one option (greedy or with user interaction)	
	; eval_config_deps(NeededIfGoal, Name, Seen, Bundle),
	  ( eval_config_deps(DefaultValueGoal, Name, Seen, Bundle),
	    eval_config_goal(DefaultValueGoal, Bundle) ->
	      ExplanationDefault = default % (default value)
	  ; ExplanationDefault = none,
	    show_message(warning, "No possible value for `~w', which is a required parameter.", [Name]),
	    fail
	  ),
	  % TODO: Default sometimes means a possible value
	  ( \+ param_def(Bundle, Name, noprevious),
	    prev_config_value(Name, Bundle, PrevValue) ->
	      ExplanationPrev = previous % (previous value)
	  ; ExplanationPrev = none,
	    PrevValue = DefaultValue
	  ),
	  % TODO: split in a different set of rules?
	  % TODO: explain only if not customizing?
	  % Explain default value
	  ( param_def(Bundle, Name, default_comment(Message)) ->
	      display_option(Message, DefaultValue)
	  ; true
	  ),
	  % Explain a particular default value
	  ( nonvar(DefaultValue),
	    param_def(Bundle, Name, default_value_comment(DefaultValue, MessageVal)) ->
	      warning($$(MessageVal))
	  ; true
	  ),
	  % Choose value
	  ( interactive_param(Bundle, Name) ->
	      % ask for a value, interactively from the user
	      param_help(Bundle, Name, Help),
	      ask_option_value(Help, Name, ValidValues, DefaultValue, PrevValue, Value)
	  ; % choose default value
	    Value = DefaultValue
	  ),
	  ( Value == DefaultValue ->
	      Explanation = using_default(ExplanationDefault)
	  ; Explanation = customized
	  )
	),
	!,
	% TODO: do proper type check and definition
	check_bundle_param_domain(Name, Value, ValidValues),
	% Set value
	%display(config_value(Bundle, Name, Value, Explanation)), nl,
	set_bundle_config_value(Bundle, Name, Value),
	% Show selected value
	( param_def(Bundle, Name, hidden),
	  \+ interactive_param(Bundle, Name) ->
	    % Do not show 'hidden' if we are not interactive
	    true
	; bundle_param_show(Bundle, Name, Value)
	).

% The parameter can be configured interactively
interactive_param(Bundle, Name) :-
	( param_def(Bundle, Name, interactive(ConfigLevels))
	; param_def(Bundle, Name, interactive(ConfigLevels, _))
	),
	get_config_level(Name, ConfigLevel),
	member(ConfigLevel, ConfigLevels),
	!.

eval_config_goal(true, _) :- !.
eval_config_goal(fail, _) :- !, fail.
eval_config_goal(param(_), _) :- !. % (ignored, evaluated in eval_config_deps)
eval_config_goal((A,B), Bundle) :- !,
	eval_config_goal(A, Bundle),
	eval_config_goal(B, Bundle).
eval_config_goal(Goal, Bundle) :- !,
	m_bundle_config_call(Bundle, Goal).

eval_config_deps(true, _, _, _) :- !.
eval_config_deps(fail, _, _, _) :- !, fail.
eval_config_deps(X, Name, Seen, Bundle) :- X = param(_), !,
	eval_config_lit(X, Name, Seen, Bundle).
eval_config_deps((A,B), Name, Seen, Bundle) :- !,
	eval_config_deps(A, Name, Seen, Bundle),
	eval_config_deps(B, Name, Seen, Bundle).
eval_config_deps(_, _Name, _Seen, _Bundle) :- !.

eval_config_lit(param(Param), MainName, Seen, ThisBundle) :-
	decomp_param(Param, ThisBundle, Bundle, Name, Value),
	eval_config_rule(Bundle, Name, [MainName|Seen]),
	( current_config_value(Bundle, Name, Value0) ->
	    true
	; fail
	),
	Value = Value0.

decomp_param(BundleNameVal, ThisBundle, Bundle, Name, Value) :- !,
	( BundleNameVal = Bundle:NameVal0 ->
	    NameVal = NameVal0
	; NameVal = BundleNameVal,
	  Bundle = ThisBundle
	),
	NameVal =.. [Name, Value].

% TODO: do proper type checking, do not halt/1
check_bundle_param_domain(Name, Value, ValidValues) :-
	( var(ValidValues) ->
	    true
	; ground(Value), member(Value, ValidValues) ->
	    true
	; display_list(
              ['Error: invalid value \'', Value, '\' for ',
	       Name, '. Valid values are: ', ValidValues]),
	  halt(1)
	).

bundle_param_show(Bundle, Name, Value) :-
	% TODO: Do not show values that are not configurable by the user?
	% This shows the value for the configuration option
	describe_param(Bundle, Name, ShowName),
	display_option(ShowName, Value).

% ---------------------------------------------------------------------------
% Pretty print parameters

describe_param(Bundle, Name, ShowName) :-
	( param_def(Bundle, Name, comment(Comment)) ->
	    ShowName = Comment
	; atom_codes(Bundle, BundleS),
	  atom_codes(Name, NameS),
	  list_concat([BundleS, ":", NameS], ShowName)
	).

display_option(ShowName, Value) :-
	length(ShowName, N),
	N2 is 30 - N,
	(N2 < 0 -> N3 is 0 ; N3 is N2),
	display('   '),
	display_string(ShowName),
	display(': '),
	repeated_display('.', N3),
	display_list([' ', Value, '\n']).

repeated_display(_,    0) :- !.
repeated_display(Term, Times) :-
	Times > 0,
	Times2 is Times - 1,
	display(Term),
	repeated_display(Term, Times2).

% ============================================================================

% TODO: Does it belong to this module?

:- use_module(library(bundle/bundle_config), [bundle_config_value/3]).
:- use_module(library(bundletool/config_common), [command_options/1]).
:- use_module(library(bundletool/third_party_install), [third_party_path/2]).

:- use_module(library(pathnames), [path_concat/3]).
:- use_module(library(system_extra), [find_executable/2]).
:- use_module(library(system), [file_exists/1]).

% (Support for GNU pkg-config based libraries)

:- export(foreign_config_var/3).
% The configuration for foreign library @var{Foreign} has value
% @var{Value} in variable @var{Var}.
foreign_config_var(Foreign, Var, Value) :-
	foreign_config_tool_path(_, Foreign, CfgToolPath),
	process_call(CfgToolPath, [~atom_concat('--', Var)],
	       [stdout(line(Value)) | ~command_options]).

% TODO: cache path
foreign_config_tool_path(Bundle, Foreign, CfgToolPath) :-
	m_bundle_foreign_config_tool(Bundle, Foreign, CfgTool),
	( % TODO: hack! make it nicer by unifying names of third_party,
	  % foreign libraries, and AUTO_INSTALL_* params (make it a functor)
	  toupper(Foreign, Foreign2),
	  atom_concat('AUTO_INSTALL_', Foreign2, AutoInstallOpt),
	  % (do not throw exception if does not exists)
	  ~current_config_value(Bundle, AutoInstallOpt) = 'yes' ->
	    % Look in third-party bin
	    third_party_path(bindir, ThirdPartyBinDir),
	    path_concat(ThirdPartyBinDir, CfgTool, CfgToolPath),
	    file_exists(CfgToolPath)
	; find_executable(CfgTool, CfgToolPath)
	).

:- export(foreign_config_version/2).
foreign_config_version(Foreign, Version) :-
	foreign_config_var(Foreign, 'version', Str),
	foreign_config_parse_version(Str, Version).

% from "Major.Minor" string to [Major,Minor]
foreign_config_parse_version(Str, L) :-
	( append(StrH, "." || StrT, Str) ->
	    L = [H|T],
	    number_codes(H, StrH),
	    foreign_config_parse_version(StrT, T)
	; L = [H],
	  number_codes(H, Str)
	).

% ---------------------------------------------------------------------------
% Set of set_prolog_flags from bundle config values (non-default values)

% TODO: Does it belong to this module or to bundle_config?
% TODO: Bundle is missing

:- use_module(library(aggregates), [findall/3]).

:- export(set_prolog_flag_from_bundle_config/1).
set_prolog_flag_from_bundle_config(TFlagCmds) :-
	findall(
	    set_prolog_flag(FlagName, CurrValue),
	    (
		config_flag_entry(FlagName, _Values, Default, CurrValue),
		(CurrValue \== Default)
	    ),
	    TFlagCmds).

% TODO: Store in saved configuration instead?
config_flag_entry(Flag, Values, Default, CurrValue) :-
	m_bundle_config_entry(Bundle, Name, Def0),
	member(config_prolog_flag(Flag, Values, Default), Def0),
	CurrValue = ~bundle_config_value(Bundle, Name).


