:- module(_, [], [fsyntax, assertions, regtypes, dcg]).

:- doc(title, "Bundletool (Prolog part)").
:- doc(author, "Jose F. Morales (inspired in previous code)").

:- doc(module, "
@section{Bootstrapping the build and installation process}

The complete build and installation process of Ciao has two
differentiated steps. First, a bootstrap binary is compiled (from a
shell-script). Then, this bootstrap binary is used to drive the rest
of the compilation and installation.

The bootstrap compiler comes in a pre-compiled bytecode form. This
bytecode together with the engine parts written in C can be executed
in most systems with a C compiler. 

Once this bootstrap compiler is available, the system compiles the
bootstrap 'bundletool', which drives the rest of the build and
installation process.

This step is transparent to the user and separated in a different
build directory. However, there are options to force the recompilation
and cleaning of that part.

@section{Building and installation}
 
This picture shows a detailed view all the elements of the build and
the installation process.

@begin{verbatim}
    Phase 1-
       SOURCE ---(configure)--> CONFIG
    Phase 2- (needs CONFIG)
       SOURCE ---(build)--> BUILD --(install)--> INSTALLED
@end{verbatim}

Where arrows specify actions and nodes are collections of files:

@begin{description}
@item{SOURCE}: Source code
@item{CONFIG}: Configuration of the source
@item{BUILD}: Binaries, compiled libraries, and generated
  documentation (.pdf, .html, etc.)
@item{INSTALLED}: Copies of BUILD that exclude temporary files (mostly
  caches for separate and incremental compilation).
@end{description}

The source distribution only contains SOURCE elements. The actions
'build', and 'install' create the BUILD, and INSTALLED elements.

NOTE: This process (except 'install') is the same for the bootstrap
'bundletool'. The big difference is that for bootstrapping the
process is driven by a simplified version written in (portable)
shell-script code (since there is no running Prolog).

@section{Maintenance (cleaning, uninstalling)}

Additionally, we provide several actions that undo the installation
('uninstall') and build processes ('clean', 'distclean'). The action
'realclean' cleans both the system and the bootstrap.

The following table summarizes the actions that @em{undo} each build
or install operation:

@begin{verbatim}
    ('undo' reverts the effect of each marked 'action') 
                  ,-------------------------------------.
   actions \\ undo | clean distclean realclean uninstall |
 .----------------+-------------------------------------+
 | boot-build     |                              x      |
 +----------------+-------------------------------------+
 | configure      |            x        x               |
 +----------------+-------------------------------------+
 | build          |   x        x        x               |
 +----------------+-------------------------------------+
 | install        |                              x      |
 `----------------+-------------------------------------+
@end{verbatim}
").

:- doc(bug, "Compilation of bootstrap leaves .po/.itf files next to
   the sources. Use a temporary CIAOCACHEDIR").

% ===========================================================================

:- use_module(library(bundletool/bundle_params),
	[root_bundle_name/1,
	 add_bundle_param_value/3,
	 bundle_param_value/3,
	 normalize_param_name/2]).
:- use_module(engine(internals),
	[po_filename/2,
	 itf_filename/2]).

:- use_module(library(read_from_string), [read_from_atom_atmvars/2]).

:- use_module(library(format),     [format/3]).

:- use_module(library(errhandle), [handle_error/2]).
:- use_module(library(lists),     [append/3, select/3]).
:- use_module(library(messages),  [error_message/2]).
:- use_module(library(aggregates), [findall/3]).

:- use_module(library(bundle/bundle_scan)).
:- use_module(library(bundle/bundle_config), [reset_all_bundle_config/0]).
:- use_module(library(bundletool/bundle_actions_aux),
	[bundle_do_hookcmd/2, ensure_load_bundleconfig_rules/0]).
:- use_module(library(bundletool/bundle_aux), [ensure_builddir/0]).
:- use_module(library(bundletool/bundle_aux), [root_bundle_source_dir/1]).

:- use_module(library(bundletool/ciaoc_aux),
	[copy_bootstrap/0,
	 emugen_absmach_def/0,
	 do_exec_header_for_builddir/0,
	 builddir_do_engine/1,
	 builddir_clean/1,
	 do_clean_dir_rec/1,
	 do_clean_mod/1,
	 clean_engine/0,
	 clean_platdep_mods/0,
	 % TODO: the following should not be exported from ciaoc_aux
	 do_delete_example_pl_files/1,
	 do_delete_file/1,
	 do_remove_dir/1
	]).

:- use_module(engine(internals), [reload_bundleregs/0, '$bundle_id'/1]).
:- use_module(library(bundletool/paths_extra), [fsR/2]).
:- use_module(library(bundle/bundle_info),
	[enum_sub_bundles/2,
	 enumrev_sub_bundles/2]).
:- use_module(library(process), [process_call/3]).

:- use_module(library(terms), [atom_concat/2]).

:- use_module(library(bundletool/messages_aux),
	[bold_message/2, normal_message/2]).

% ===========================================================================

% Some definitions:
%
% - bundle: a set of modules (which may depend on other bundles)
% - builddirs: areas (directory) where configuration and compilation takes place
% - installdirs: areas where installation takes place
%
% The bootstrap is compiled and run from its own 'builddir'
% ('build-boot/'), which is different from the builddir of the system
% ('build/').

% TODO: Document; we use a 2-phase with a separate bootstrap
%       [Idea: create bundletool_boot.sh]
%

% ===========================================================================
% The default bundle

% TODO: not necesarily root_bundle_name
default_bundle(Id) :- root_bundle_name(Id).

% ===========================================================================
:- doc(section, "Commands for 'ciaosetup'").

% TODO: (Here just for documentation)

:- regtype ciaosetup_cmd/1 # "Commands for 'ciaosetup' (SH side)".
ciaosetup_cmd(help).
ciaosetup_cmd(help_extra).
ciaosetup_cmd(emergency_clean).
ciaosetup_cmd(boot_build).
ciaosetup_cmd(boot_rebuild).
ciaosetup_cmd(boot_clean).
ciaosetup_cmd(boot_promote).
ciaosetup_cmd(X) :- bundle_cmd(X).

% ===========================================================================

% Invocation from the command-line interface
:- export(main/1).
main(Args) :-
	catch(main_(Args), E, handle_bundletool_error(E)),
	!.

% (debug)
% main_(Args) :-
% 	display(user_error, bundletool_aux_main(Args)), nl(user_error),
% 	fail.
main_(['help'|Args]) :- !,
	( Args = [] -> Cmd = '' ; Args = [Arg] -> Cmd = Arg ),
	do__help(Cmd).
main_(['help_extra'|Args]) :- !,
	( Args = [] -> Cmd = '' ; Args = [Arg] -> Cmd = Arg ),
	do__help_extra(Cmd).
main_(CmdArgs) :- !,
	parse_args(CmdArgs, Cmd, Target, Opts),
	bundle_do_cmd(Cmd, Target, Opts).

% ===========================================================================
:- doc(section, "Parsing of bundle actions from the CLI").

% The format is 'ACTION ARGS', where:
%
%  - 'ACTION' is a bundle action (e.g., 'build') 
%  - 'ARGS' is a sequence of options (e.g., --interactive) and bundle
%    targets (e.g., 'ciao')
%
% Most actions only accept one target. Some actions do not accept
% options.

% From options (term) to CLI arguments (atoms)
opts_to_args([param(Bundle, Name, Value)|Opts0], [Opt|Opts]) :- !,
	atom_concat(['--', Bundle, ':', Name, '=', Value], Opt),
	opts_to_args(Opts0, Opts).
opts_to_args([], []).

parse_args(CmdArgs, Cmd, Target, Opts) :-
	catch(parse_args_(CmdArgs, Cmd, Opts, Target),
              E, handle_bundletool_error(E)).

% Parse command, arguments, targets, and options
parse_args_([Cmd|Args], Cmd, Opts, Target) :- !,
	( bundle_cmd(Cmd) ->
	    true
	; throw(unknown_bundle_action(Cmd))
	),
	cmd_opts(Cmd, BundleParse, OptsParse),
	parse_args__(Args, Opts, Targets),
	% Select default target if none
	( BundleParse = arg_bundle ->
	    ( Targets = [] -> default_bundle(Target) 
	    ; Targets = [Target] -> true
	    ; throw(args_error("Multiple targets not accepted ('~w')", [Targets]))
	    )
	; % TODO: Implement opt_bundle, cwd_bundle; bundle may not be needed
	  default_bundle(Target)
	),
	( OptsParse = config_opts ->
	    % NOTE: checked later in configure
	    true
	; OptsParse = raw_opts ->
	    ( member(param(_, Name, _), Opts), % TODO: do not qualify in raw_opt
	      normalize_param_name(Name, Name2),
	      \+ raw_opt(Cmd, Name2) ->
	        throw(args_error("Unrecognized option '~w' for command '~w'", [Name, Cmd]))
	    ; true
	    )
	; OptsParse = no_opts ->
	    ( Opts = [] ->
	        true
	    ; opts_to_args(Opts, Opts2),
	      throw(args_error("Unrecognized options '~w' for command '~w'", [Opts2, Cmd]))
	    )
	; fail
	).
parse_args_([], _Cmd, _Opts, _Target) :-
	throw(args_error("No arguments were specified", [])).

parse_args__(['--interactive'|Args], Opts, Targets) :- % TODO: special arg parser?
	!,
	root_bundle_name(Root),
	Opts = [param(Root, 'INTERACTIVE_CONFIG', true)|Opts0],
	parse_args__(Args, Opts0, Targets).
parse_args__(['--list-options'|Args], Opts, Targets) :- % TODO: special arg parser?
	!,
	root_bundle_name(Root),
	Opts = [param(Root, 'LIST_OPTIONS', true)|Opts0],
	parse_args__(Args, Opts0, Targets).
parse_args__(['--describe-option', Param|Args], Opts, Targets) :- % TODO: special arg parser?
	!,
	root_bundle_name(Root),
	parse_param_name(Param, Bundle, Name),
	Targets = [Bundle|Targets0],
	Opts = [param(Root, 'DESCRIBE_OPTION', Name)|Opts0],
	parse_args__(Args, Opts0, Targets0).
parse_args__([Param0|Args], Opts, Targets) :-
	atom_concat('--', Param, Param0),
	parse_param_atom(Param, Bundle, Name, Value),
	!,
	Opts = [param(Bundle, Name, Value)|Opts0],
	parse_args__(Args, Opts0, Targets).
parse_args__([Target|Args], Opts, Targets) :- !,
	Targets = [Target|Targets0],
	parse_args__(Args, Opts, Targets0).
parse_args__([], [], []).

% cmd_opts(Cmd, BundleParse, OptsParse):
%
%   Parsing specification for command Cmd, where:
%
%   BundleParse = arg_bundle  % bundle is the non-option argument
%               | root_bundle % bundle is root_bundle
%   OptsParse = config_opts % configuration options
%             | raw_opts    % raw options
%             | no_opts     % no options

% TODO: extend BundleParse, implement opt_bundle (--bundle=BUNDLE) and
%   cwd_bundle (extract from working directory by looking at
%   registered bundles and scanning Manifest.pl in parent directories)

cmd_opts(user_install,          arg_bundle, config_opts) :- !.
cmd_opts(system_install,        arg_bundle, config_opts) :- !.
cmd_opts(user_install_paranoid, arg_bundle, config_opts) :- !.
cmd_opts(fullinstall,           arg_bundle, config_opts) :- !.
cmd_opts(configure,             arg_bundle, config_opts) :- !.
% (internal)
cmd_opts(scan_and_config,       arg_bundle, config_opts) :- !.
cmd_opts(config_noscan,         arg_bundle, config_opts) :- !.
cmd_opts(config_list_options,   arg_bundle, raw_opts) :- !.
cmd_opts(config_describe_option, arg_bundle, raw_opts) :- !.
%
cmd_opts(install,               arg_bundle, raw_opts) :- !.
cmd_opts(uninstall,             arg_bundle, raw_opts) :- !.
%
cmd_opts(environment_and_windows_bats, arg_bundle, raw_opts) :- !.
cmd_opts(gen_bundle_commit_info, arg_bundle, raw_opts) :- !.
% (any other command: no options)
cmd_opts(_,                     arg_bundle, no_opts).

% TODO: pass raw_opt as regular arguments (do not call restore_params on them)
raw_opt(environment_and_windows_bats, 'EMACS_PATH').
% NOTE: '--destdir' option is used in: 
%   Portfile.skel
%   pbundle_gen_mac.pl
%   CiaoDE.spec.skel
%   and some code in ciaobot
raw_opt(install, 'DESTDIR').
raw_opt(uninstall, 'DESTDIR').
raw_opt(config_list_options, 'LIST_OPTIONS').
raw_opt(config_describe_option, 'DESCRIBE_OPTION').
%   --git-repo-dir=Dir: location of the Git repository (if not using default)
raw_opt(gen_bundle_commit_info, 'GIT_REPO_DIR').

:- use_module(library(lists), [list_concat/2]).

% Parse a parameter name
parse_param_name(Param, Bundle, Name) :-
	atom_codes(Param, ParamS),
	list_concat([BundleS, ":", NameS], ParamS),
	!,
	atom_codes(Bundle, BundleS),
	atom_codes(Name, NameS).
parse_param_name(Param, Bundle, Name) :-
	atom_codes(Param, ParamS),
	!,
	% Use root_bundle_name if bundle is not specified
	root_bundle_name(Bundle),
	atom_codes(Name, ParamS).

% Parse a parameter assignment (as atom)
parse_param_atom(Param, Bundle, Name, Value) :-
	parse_param_string(Param, Bundle, Name, ValueS),
	atom_codes(Value, ValueS).

% Parse a parameter assignment (as string)
parse_param_string(Param, Bundle, Name, ValueS) :-
	atom_codes(Param, ParamS),
	list_concat([BundleS, ":", NameS, "=", ValueS], ParamS),
	!,
	atom_codes(Bundle, BundleS),
	atom_codes(Name, NameS).
parse_param_string(Param, Bundle, Name, ValueS) :-
	atom_codes(Param, ParamS),
	list_concat([NameS, "=", ValueS], ParamS),
	!,
	% Use root_bundle_name if bundle is not specified
	root_bundle_name(Bundle),
	atom_codes(Name, NameS).

% ===========================================================================
% Load bundle parameters

restore_params([param(Bundle, Name, Value)|Opts]) :- !,
	normalize_param_name(Name, Name2),
	( bundle_param_value(Bundle, Name2, _) ->
	    true % TODO: check
	; add_bundle_param_value(Bundle, Name2, Value)
	),
	restore_params(Opts).
restore_params([]).

% ===========================================================================
:- doc(section, "bundle_do_cmd commands").

% NOTE ABOUT 'clean' ACTIONS
% 
%   The meaning of 'clean' and 'distclean' is based on the Makefile
%   example at:
%
%       http://www.gnu.org/software/make/manual/html_node/Complex-Makefile.html
% 
%   However we do it in a slightly different way:
%
%       'realclean' = 'distclean' + 'boot-clean'
% 
% TODO: See cleaning targets for 'lpdoc' and fix it. See other clean
%     targets in bundle_actions__ciao.
% 
% TODO: improve output for foreign clean

% TODO: Proposal for changes of those names
%       (remember to update the documentation)
%
%   Allow modifiers for global installation options?
%
%   user_install => local_install
%   system_install => global_install
%
%   'fullinstall' and system-install are the same

% ---------------------------------------------------------------------------

:- regtype bundle_cmd/1 # "bundle_do_cmd commands (Prolog side)".
bundle_cmd(user_install).
bundle_cmd(system_install).
bundle_cmd(user_install_paranoid).
bundle_cmd(fullinstall).
bundle_cmd(configure).
bundle_cmd(config_noscan).
bundle_cmd(rescan_bundles).
bundle_cmd(scan_and_config). % (only in bundle_do_cmd)
bundle_cmd(build).
bundle_cmd(build_docs). % (for some cases, also in user_cmd)
bundle_cmd(build_nodocs). % (for some cases, also in user_cmd)
bundle_cmd(build_nolibs).
bundle_cmd(build_platdep).
bundle_cmd(clean).
bundle_cmd(clean_nodocs).
bundle_cmd(clean_norec).
bundle_cmd(clean_platdep).
bundle_cmd(clean_docs).
bundle_cmd(clean_docs_manuals).
bundle_cmd(clean_docs_readmes).
bundle_cmd(distclean).
bundle_cmd(configclean).
bundle_cmd(install). % (for some cases, also in user_cmd)
bundle_cmd(uninstall). % (for some cases, also in user_cmd)
bundle_cmd(register). % (new in bundle_cmd?)
bundle_cmd(unregister). % (new in bundle_cmd?)
bundle_cmd(X) :- bundle_hookcmd(X).

bundle_do_cmd_list(_Action, [], _Opts) :- !.
bundle_do_cmd_list(Action, [Target|Targets], Opts) :-
	bundle_do_cmd(Action, Target, Opts),
	bundle_do_cmd_list(Action, Targets, Opts).

bundle_do_cmd(user_install, Target, Opts) :- !,
	root_bundle_name(Root),
	Opts2 = [param(Root, 'instype', 'local')|Opts],
	bundle_do_cmd(fullinstall, Target, Opts2).
bundle_do_cmd(system_install, Target, Opts) :- !,
	root_bundle_name(Root),
	Opts2 = [param(Root, 'instype', 'global')|Opts],
	bundle_do_cmd(fullinstall, Target, Opts2).
bundle_do_cmd(user_install_paranoid, Target, Opts) :- !,
	root_bundle_name(Root),
	Opts2 = [param(Root, 'instype', 'local'),
		 param(Root, 'unused-pred-warnings', 'yes')|Opts],
	bundle_do_cmd(fullinstall, Target, Opts2).
bundle_do_cmd(fullinstall, Target, Opts) :- !,
	bundle_do_cmd(scan_and_config, Target, Opts),
	bundle_do_cmd(build, Target, []),
	bundle_do_cmd(install, Target, []).
bundle_do_cmd(configure, Target, Opts) :- !,
	root_bundle_name(Root),
	( member(param(Root, 'LIST_OPTIONS', true), Opts) ->
	    restore_params(Opts), % TODO: needed?
	    bundle_do_cmd(config_list_options, Target, Opts)
	; member(param(Root, 'DESCRIBE_OPTION', _Opt), Opts) ->
	    restore_params(Opts), % TODO: needed, pass arg instead
	    bundle_do_cmd(config_describe_option, Target, Opts)
	; bundle_do_cmd(scan_and_config, Target, Opts),
	  post_config_message(Target)
	).
bundle_do_cmd(boot_promote, Target, _Opts) :- !,
	do_boot_promote(Target).
bundle_do_cmd(scan_and_config, Target, Opts) :- !,
	% Note: scanning bundles must be done before configuration
	% Note: fsR cannot be used until bundles are scanned
	root_bundle_name(Root),
	( Target = Root ->
	    % TODO: avoid scanning for some options... allow scanning for other bundles
	    normal_message("Scanning bundles available in this source tree", []),
	    root_bundle_source_dir(CiaoDESrc),
	    bundle_scan(CiaoDESrc),
	    reload_bundleregs % (reload, bundles has been scanned)
	; true
	),
	%
	bundle_do_cmd(config_noscan, Target, Opts).
bundle_do_cmd(config_noscan, Target, Opts) :- root_bundle_name(Target), !,
	restore_params(Opts),
	ensure_builddir,
	root_bundle_name(Root),
	( bundle_param_value(Root, 'INTERACTIVE_CONFIG', true) ->
	    true % use saved config values
	; % TODO: do not reset, skip load instead
	  reset_all_bundle_config
	),
	do_config_noscan.
bundle_do_cmd(config_noscan, Target, _Opts) :- !,
	% TODO: implement configuration for individual bundles
	format(user_error, "ERROR: Cannot configure bundle `~w'.~n", [Target]),
	halt(1).
bundle_do_cmd(rescan_bundles, Target, _Opts) :- !,
        normal_message("Re-scanning (sub-)bundles for '~w'", [Target]),
	root_bundle_name(Root),
	( Target = Root -> % Allow scanning without previous 'configure'
	    root_bundle_source_dir(BundleDir)
	; BundleDir = ~fsR(bundle_src(Target)) ->
	    true
	; format(user_error, "ERROR: bundle '~w' is unknown (did you run 'configure'?)~n", [Target]),
	  halt(1)
	),
	( root_bundle_name(Target) ->
	    bundle_scan(BundleDir)
	; % TODO: implement rescan_bundles for individual bundles
	  format(user_error, "ERROR: sub-bundle scan not supported yet~n", []),
	  halt(1)
	).
bundle_do_cmd(build, Target, Opts) :- !,
	% TODO: Make sure that lpdoc is build before creating the documentation
	%       (in general, 'build' should delay tasks)
	bundle_do_cmd(build_nodocs, Target, Opts),
	bundle_do_cmd(build_docs, Target, Opts).
bundle_do_cmd(build_nodocs, ciaode, Opts) :- !,
	ensure_ready_for_build(build_nodocs, ciaode),
	bold_message("Building (no docs) the Ciao System", []),
	normal_message("(an engine with the new settings will be compiled)", []),
	% Separated process, because 'build_nodocs ciao' breaks installer in Windows
	% TODO: Why?
	bundle_do_cmd_list(build_nodocs, [engine, ciao, extra], Opts),
	bold_message("Ciao System build (no docs) completed", []).
% TODO: Implement really fast build: reuse the bootstrap engine if the config and code was the same
bundle_do_cmd(build_nodocs, engine, _Opts) :- !,
	ensure_ready_for_build(build_nodocs, engine),
	emugen_absmach_def,
	builddir_do_engine(build),
	% TODO: not sure if this should be done here or as precompilation of ciaoc (this is simply a cached text file for the active engine!)
	do_exec_header_for_builddir.
bundle_do_cmd(build_nodocs, static_engine, _Opts) :- !, % (Invoked from ciaobot)
	ensure_ready_for_build(build_nodocs, static_engine),
	% TODO: do exec_header_for_builddir here?
	builddir_do_engine(build_static).
bundle_do_cmd(build_nodocs, extra, Opts) :- !,
	bundle_do_cmd(build_extra, ciaode, Opts).
bundle_do_cmd(build_nodocs, libraries, Opts) :- !,
	bundle_do_cmd(build_libraries, ciao, Opts),
	bundle_do_cmd(build_extra_libraries, ciaode, Opts).
bundle_do_cmd(build_nodocs, emacs_mode, Opts) :- !,
	bundle_do_cmd(build_emacs_mode, ciao, Opts).
bundle_do_cmd(build_nodocs, shell, Opts) :- !,
	bundle_do_cmd(build_shell, ciao, Opts).
bundle_do_cmd(build_nodocs, ciaoc, Opts) :- !,
	bundle_do_cmd(build_ciaoc, ciao, Opts).
bundle_do_cmd(build_nodocs, profiler, Opts) :- !,
	bundle_do_cmd(build_profiler, ciao, Opts).
bundle_do_cmd(build_docs, emacs_mode, _Opts) :- !.
bundle_do_cmd(build_docs, shell, _Opts) :- !.
bundle_do_cmd(build_docs, ciaoc, _Opts) :- !.
bundle_do_cmd(build_docs, profiler, _Opts) :- !.
bundle_do_cmd(build_docs, engine, _Opts) :- !.
%
bundle_do_cmd(build_nolibs, ciaode, Opts) :- !,
	ensure_ready_for_build(build_nolibs, ciaode),
        % TODO: why is 'ciaoc' and 'shell' not included in build_nolibs of 'ciao'?
	bundle_do_cmd_list(build_nodocs, [engine, ciaoc, shell], Opts),
	bundle_do_cmd(build_nolibs, ciao, Opts),
	bundle_do_cmd(build_nolibs_extra, ciaode, Opts).
%
bundle_do_cmd(build_platdep, ciaode, Opts) :- !,
	ensure_ready_for_build(build_platdep, ciaode),
	% TODO: 'platdep' is an overapproximation, it also includes
	%       binaries that are platform independent!!!
        % TODO: why is 'ciaoc' and 'shell' not included in build_platdep of 'ciaode'?
	bundle_do_cmd_list(build_nodocs, [engine, ciaoc, shell], Opts),
	bundle_do_cmd(build_platdep, ciao, Opts),
	bundle_do_cmd(build_platdep_extra, ciaode, Opts).
%
% Deletes all intermediate files, but leaves configuration settings
% NOTE: Only 'clean' and 'clean_platdep' are necessary for 'ciaobot' at this moment
bundle_do_cmd(clean, Target, Opts) :- !,
	bundle_do_cmd(clean_docs, Target, Opts),
	bundle_do_cmd(clean_nodocs, Target, Opts).
%
% Like 'clean', but keeps documentation targets.
% (This reverses the 'build_nodocs' and part of 'build_docs' actions)
bundle_do_cmd(clean_nodocs, Target, Opts) :- !,
	ensure_ready_for_build(clean_nodocs, Target),
	normal_message("Cleaning (nodocs) ~w", [Target]),
	% TODO: ad-hoc list, use bundle deps
	( root_bundle_name(Target) ->
	    findall(P, enumrev_sub_bundles(Target, P), Ps),
	    bundle_do_cmd_list(clean_nodocs, Ps, Opts)
	; Target = ciao ->
	    Parts = [emacs_mode, java, timingmodel],
	    bundle_do_cmd_list(clean_nodocs, Parts, Opts)
	; Target = ciaopp ->
	    bundle_do_cmd(clean_nodocs, ilciao, Opts)
	; true
	),
	bundle_do_cmd(clean_norec, Target, Opts).
%
% Like 'clean_nodocs' but non-recursive on sub-bundles
bundle_do_cmd(clean_norec, Target, Opts) :- !,
	ensure_ready_for_build(clean_norec, Target),
	%
	( root_bundle_name(Target) ->
	    do_clean_dir_rec(~fsR(bundle_src(Target)/'Manifest')),
            % TODO: Clean Manifest/... in each bundle too
	    % TODO: clean all builddir except configuration?
	    builddir_clean(pbundle),
	    builddir_clean(bin),
	    normal_message("Cleaning engine", []),
	    clean_engine
	; Target = ciao ->
            % Clean some files from ciao_etc sub-bundle
	    % TODO: improve
	    do_delete_file(~fsR(bundle_src(ciao)/etc/'DOTprofile')),
	    do_delete_file(~fsR(bundle_src(ciao)/etc/'DOTcshrc')),
            % TODO: make sure that clean_dir_rec is not used in directories that contains other sub-bundles (otherwise we are cleaning things twice or more)
	    do_clean_dir_rec(~fsR(bundle_src(ciao)))
	; Target = website ->
	    do_delete_file(~fsR(bundle_src(website)/download_cgi/'download_config.pl')), % TODO: add hook
	    do_delete_file(~fsR(bundle_src(website)/'SETTINGS.pl')),
	    do_delete_file(~fsR(bundle_src(website)/bundletool_bot)),
	    do_clean_dir_rec(~fsR(bundle_src(website)))
	; Target = java ->
	    bundle_do_cmd(clean_javalibs, ciao, Opts)
	; Target = ilciao ->
            % TODO: Write in Prolog or use build.xml
	    do_delete_example_pl_files(~fsR(bundle_src(ciaopp)/ilciao/examples)),
	    do_remove_dir(~fsR(bundle_src(ciaopp)/ilciao/classes))
	; Target = emacs_mode ->
	    % TODO: Check if this is already done from Prolog
	    do_clean_dir_rec(~fsR(bundle_src(ciaode)/'emacs-mode')),
            % TODO: fix the Prolog part
	    bundle_do_cmd(clean_emacs_mode, ciao, Opts)
	; Target = timingmodel ->
	    bundle_do_cmd(clean_timingmodel, ciao, Opts)
	; Target = engine ->
	    clean_engine
	; Target = profiler ->
	    % (already included in 'ciao')
	    do_clean_dir_rec(~fsR(bundle_src(ciao)/contrib/profiler))
	; '$bundle_id'(Target) ->
	    do_clean_dir_rec(~fsR(bundle_src(Target)))
	; format(user_error, "ERROR: unknown bundle '~w'.~n", [Target]),
	  halt(1)
	).
%
bundle_do_cmd(clean_platdep, Target, _Opts) :- !,
	ensure_ready_for_build(clean_platdep, Target),
	( root_bundle_name(Target) ->
	    normal_message("Cleaning engine", []),
	    clean_engine,
	    normal_message("Cleaning binaries", []),
	    builddir_clean(bin),
	    normal_message("Cleaning platform-dependent modules", []),
	    clean_platdep_mods
	; true
	).
%
% Clean the final documentation targets (READMEs and manuals)
bundle_do_cmd(clean_docs, Target, Opts) :- !,
	ensure_ready_for_build(clean_docs, Target),
	% TODO: refine targets
	( root_bundle_name(Target) ->
	    findall(P, enum_sub_bundles(Target, P), Ps),
	    bundle_do_cmd_list(clean_docs, Ps, Opts)
	; true
	),
	bundle_do_cmd(clean_docs_manuals, Target, Opts),
	bundle_do_cmd(clean_docs_readmes, Target, Opts).
bundle_do_cmd(clean_docs_manuals, Target, _Opts) :- !,
	% TODO: use Manifest
	% TODO: this cleans all manuals at once; use lpdoc instead?
	ensure_ready_for_build(clean_docs_manuals, Target),
	( root_bundle_name(Target) ->
	    do_remove_dir(~fsR(builddir(build)/doc))
	; true
	).
bundle_do_cmd(clean_docs_readmes, Target, _Opts) :- !,
	% TODO: use Manifest
	ensure_ready_for_build(clean_docs_readmes, Target),
	( Target = ciaode ->
	    % TODO: Should this file be there?
            do_delete_file(~fsR(bundle_src(Target)/'DEVEL'))
	; Target = ciao ->
            % TODO: Is this cleaned from bundle_actions__ciao.pl too?
            do_delete_file(~fsR(bundle_src(Target)/'NewUser')),
       	    do_delete_file(~fsR(bundle_src(Target)/'README')),
       	    do_delete_file(~fsR(bundle_src(Target)/'INSTALLATION')),
       	    do_delete_file(~fsR(bundle_src(Target)/'INSTALLATION_Win32'))
	; Target = lpdoc ->
            % TODO: Is this cleaned from bundle_actions__lpdoc.pl too?
            do_delete_file(~fsR(bundle_src(Target)/'README')),
       	    do_delete_file(~fsR(bundle_src(Target)/'INSTALLATION'))
	; Target = ciaopp ->
            % TODO: Is this cleaned from bundle_actions__ciaopp.pl too?
            do_delete_file(~fsR(bundle_src(Target)/'README')),
       	    do_delete_file(~fsR(bundle_src(Target)/'INSTALLATION'))
	; true
	).
%
% Like 'clean' but also removes configuration settings.
bundle_do_cmd(distclean, Target, Opts) :- !,
	ensure_ready_for_build(distclean, Target),
	bundle_do_cmd(clean, Target, Opts),
	%
	% Clean config (this must be the last step)
	normal_message("Cleaning configuration for '~w'", [Target]),
	bundle_do_cmd(configclean, Target, Opts),
	( root_bundle_name(Target) ->
	    % TODO: make sure that all binaries are removed in 'clean'
	    builddir_clean(bundlereg),
	    builddir_clean(all)
	; true
	).
%
% Clean the bundle configuration
% TODO: split in a configclean for each bundle (i.e., engine, etc.)
% Warning! configclean is the last cleaning step. If you clean all
%          the configuration files then many scripts will not run.
bundle_do_cmd(configclean, Target, _Opts) :- !,
	ensure_ready_for_build(configclean, Target),
	( root_bundle_name(Target) ->
	    builddir_clean(config)
	; true
	).
% ----------
bundle_do_cmd(register, emacs_mode, Opts) :- !,
	bundle_do_cmd(register_emacs_mode, ciao, Opts).
bundle_do_cmd(unregister, emacs_mode, Opts) :- !,
	bundle_do_cmd(unregister_emacs_mode, ciao, Opts).
% ----------
bundle_do_cmd(install, emacs_mode, Opts) :- !,
	bundle_do_cmd(install_emacs_mode, ciao, Opts).
bundle_do_cmd(uninstall, emacs_mode, Opts) :- !,
	bundle_do_cmd(uninstall_emacs_mode, ciao, Opts).
% ----------
bundle_do_cmd(Action, Bundle, Opts) :-
	restore_params(Opts),
	bundle_do_hookcmd(Bundle, Action).

% ===========================================================================
:- doc(section, "Bundle commands that may require hooks (bundle_do_hookcmd)").

:- regtype bundle_hookcmd/1 # "Bundle commands that may require hooks".
bundle_hookcmd(X) :- user_cmd(X).
bundle_hookcmd(X) :- wrapped_cmd(X).
bundle_hookcmd(X) :- internal_cmd(X).

% same name as targets for bundle_do
user_cmd(build_docs). % common
user_cmd(build_nodocs). % common
user_cmd(install). % common
user_cmd(uninstall). % common
%
user_cmd(show_bundles). % ciaode
%
user_cmd(gen_bundle_commit_info). % ciaode
user_cmd(gen_pbundle__app). % ciaode
user_cmd(gen_pbundle__bin). % ciaode
user_cmd(gen_pbundle__bin_tbz). % ciaode
user_cmd(gen_pbundle__bin_tgz). % ciaode
user_cmd(gen_pbundle__descfile). % ciaode
user_cmd(gen_pbundle__macport). % ciaode
user_cmd(gen_pbundle__noa). % ciaode
user_cmd(gen_pbundle__noa_tbz). % ciaode
user_cmd(gen_pbundle__noa_tgz). % ciaode
user_cmd(gen_pbundle__pkg). % ciaode
user_cmd(gen_pbundle__rpm). % ciaode
user_cmd(gen_pbundle__src). % ciaode
user_cmd(gen_pbundle__tbz). % ciaode
user_cmd(gen_pbundle__tgz). % ciaode
user_cmd(gen_pbundle__win32). % ciaode
% Called from CiaoDE.iss.skel
% Uses this parameter:
%   --emacs_path=EmacsDir (as a windows path)
user_cmd(environment_and_windows_bats). % ciao

% temporary subtargets
wrapped_cmd(build_emacs_mode). % ciao (wrapped)
wrapped_cmd(install_emacs_mode). % ciao (wrapped)
wrapped_cmd(uninstall_emacs_mode). % ciao (wrapped)
wrapped_cmd(register_emacs_mode). % ciao (wrapped)
wrapped_cmd(unregister_emacs_mode). % ciao (wrapped)
%
wrapped_cmd(build_shell). % ciao (wrapped)
wrapped_cmd(build_ciaoc). % ciao (wrapped)
wrapped_cmd(build_profiler). % ciao (wrapped)
%
wrapped_cmd(config_list_options). % ciaode (wrapped, for 'configure' with some options)
wrapped_cmd(config_describe_option). % ciaode (wrapped, for 'configure' with some options)

% targets from DO invoked from do__internal (sh)
internal_cmd(build_extra). % ciaode
internal_cmd(build_extra_libraries). % ciaode
internal_cmd(build_nolibs_extra). % ciaode
internal_cmd(clean_emacs_mode). % ciao
internal_cmd(clean_javalibs). % ciao
internal_cmd(clean_timingmodel). % ciao
%
internal_cmd(build_libraries). % common
internal_cmd(build_nolibs). % common
internal_cmd(build_platdep). % common
internal_cmd(build_platdep_extra). % ciaode
internal_cmd(build_docs_manuals). % common
internal_cmd(build_docs_readmes). % common
%
% TODO: separate as special hooks?
internal_cmd(calibrateciaopp). % common
internal_cmd(save_calibrateciaopp). % common
internal_cmd(gen_timingmodel). % common
%
internal_cmd(install_docs). % common
internal_cmd(uninstall_docs). % common
internal_cmd(prebuild_libraries). % common
internal_cmd(runbenchmarks). % common
internal_cmd(runtests). % common
internal_cmd(create_tags). % common
internal_cmd(delete_tags). % common

% ---------------------------------------------------------------------------
:- doc(section, "Invoking the Configuration").

:- use_module(library(bundle/bundle_config),
	[bundle_config_value/3,
	 restore_all_bundle_config/0]).
:- use_module(library(bundletool/bundle_configure), [config_noscan/0]).

% Invoke configuration (do not scan bundles)
do_config_noscan :-
	ensure_load_bundleconfig_rules, % TODO: make it more fine grained (not always needed, not all bundles needed)
 	restore_all_bundle_config, % TODO: Necessary? bundle(bundle_config) contains an initialization directive
	config_noscan.

% ---------------------------------------------------------------------------
:- use_module(library(system_extra), [using_tty/0]).

% Show a help message after (only when run from a TTY)
post_config_message(Target) :-
	( using_tty ->
	    show_post_config_message(Target)
	; true
	).

show_post_config_message(Target) :- root_bundle_name(Target), !,
	normal_message(
"Please check that all the configuration values above are correct. If
not, you can change or customize the configuration using the command
line or --interactive configure option.

To continue, execute:
   \"./ciaosetup build\"    to compile and generate the documentation
                          (if not in distribution), then
   \"./ciaosetup install\"  to install the system.", []).
show_post_config_message(_).

% ---------------------------------------------------------------------------
:- doc(section, "Promote bootstrap compiler (dangerous!)").

% TODO: Create a backup/ directory with a NODISTRIBUTE flag?
% TODO: This is an operation of ciao/ciaoc bundle
%
% NOTE: It only saves the static ciaoc (but it could save other
% bootstrapping data like documentation in the future).

do_boot_promote(Target) :-
	( root_bundle_name(Target) ->
	    true
	; % TODO: implement configuration for individual bundles
	  format(user_error, "ERROR: Cannot promote bundle '~w'.~n", [Target]),
	  halt(1)
	),
	%
	ensure_ready_for_build(boot_promote, Target),
	ask_copy_bootstrap.

ask_copy_bootstrap :-
	bold_message("Promoting bootstrap compiler", []),
	normal_message("The current compiler will become the next bootstrap compiler", []),
	normal_message("(Warning: do not call it unless you are completely sure)", []),
	normal_message("", []),
	% TODO: I refuse to add an ad-hoc list of files here -- wait until I can extract the list from absmach_def.pl
	normal_message("+-------------------------------------------------------------------------+", []),
	normal_message("|        WARNING!!! Auto-generated C files are not being copied.          |", []),
	normal_message("|                     Remember to copy them manually.                     |", []),
	normal_message("+-------------------------------------------------------------------------+", []),
	normal_message("", []),
	( ask_yesno("Are you sure?"),
	  ask_yesno("Really?") ->
	    copy_bootstrap,
	    normal_message("Bootstrap compiler promoted", [])
	; normal_message("Promotion canceled", [])
	).

% ===========================================================================

% Check that there exist a configuration. It also implies that there
% exist a running bootbundletool and that bundles has been scanned.

:- use_module(library(system), [file_exists/1]).

% TODO: Check another file instead (e.g., 'CONFIG_saved')?
ensure_ready_for_build(Action, Target) :-
	( % TODO: fsR/1 will fail if bundles are not scanned (that also means that there is no configuration)
	  ConfigSH = ~fsR(builddir(build)/'CONFIG_sh'),
	  file_exists(ConfigSH) ->
	    true
	; format(user_error, "ERROR: Cannot do '~w' on bundle '~w' without a configuration. Please run 'configure' before.~n", [Action, Target]),
	  halt(1)
	).

% ===========================================================================
:- doc(section, "Handle errors in bundletool").

handle_bundletool_error(args_error(Format, Args)) :-
	append("ERROR: ", Format, T1),
	append(T1, "~n", T2),
	format(user_error, T2, Args),
	halt(1).
handle_bundletool_error(make_error(Format, Args)) :- % TODO: used?
	error_message(Format, Args),
	halt(1).
handle_bundletool_error(unknown_bundle(Bundle)) :-
	format(user_error, "ERROR: Unknown bundle '~w'.~n", [Bundle]),
	halt(1).
handle_bundletool_error(unknown_bundle_action(Action)) :-
	format(user_error, "ERROR: Unknown command '~w'.~n", [Action]),
	halt(1).
handle_bundletool_error(bundle_do_failed(Bundle, Target)) :-
	format(user_error, "ERROR: Command '~w' on bundle '~w' failed.~n", [Target, Bundle]),
	halt(1).
handle_bundletool_error(error(Error, Where)) :-
	handle_error(Error, Where).
handle_bundletool_error(Error) :-
	format(user_error, "ERROR: Unknown error '~w'.~n", [Error]),
	halt(1).

% ===========================================================================
:- doc(section, "Auxiliary for I/O").

:- use_module(library(strings), [get_line/1, write_string/1]).

% Prompt Question to the user, wait until 'yes' or 'no' is written
ask_yesno(Question) :-
	write_string(Question), write_string(" (y/n) "),
	ask_yesno_loop.

ask_yesno_loop :-
	get_line(Answer),
	( Answer = "y" ->
	    true
	; Answer = "n" ->
	    fail
	; normal_message("Unrecognized answer. Please type 'y' for 'yes' or 'n' for 'no'.", []),
	  ask_yesno_loop
	).

% ===========================================================================
:- doc(section, "Help (for the command-line)").
% TODO: Move most of this code to an auxliary library (CLI-help?)
% TODO: Transform predicates into facts (later, extract from assertions)

% Shows the help message
show_help('') --> !,
	show_help_banner,
	%
	show_help_group(help),
	show_help_group(bootstrap),
	show_help_group(quickstart),
	show_help_group(configure),
	show_help_group(bootstrap_opts),
	show_help_group(bundle_actions),
	show_help_group(promotion),
	%
	show_help_group(emacs_mode),
	show_help_group(ciaopp).
show_help(_) -->
	"No help for command\n".

show_help_banner -->
	"usage: ciaosetup <cmd> [<args>]\n",
	"\n".

% Help for help
show_help_group(help) -->
	show_help_section("Help"),
	showcmd(help, "[<cmd>]", [
          %1_______________________________________________
	  "Show help (about CMD if specified)"
        ]),
	showcmd(help_extra, "[<cmd>]", [
          %1_______________________________________________
	  "Show advanced help"
        ]),
	show_help_sep.

show_help_group(bootstrap_opts) --> { extrahelp }, !,
	% show_help_section("Options for bootstrapping")
        showcmdtext([
          %2........................________________________________________________
          "These special options are recognized during the bootstrapping process:",
	  "",
	  "  --ciao:custom-cc=CC         Select a custom C compiler",
	  "  --ciao:custom-ld=LD         Select a custom C linker (optional, uses",
	  "                              custom C compiler if not provided)",
	  "  --ciao:extra-cflags=FLAGS   Use additional flags for C compiler",
	  "  --ciao:extra-ldflags=FLAGS  Use additional flags for C linker"
        ]),
	show_help_sep.
show_help_group(bootstrap_opts) --> [].

show_help_group(bootstrap) --> { extrahelp }, !,
	show_help_section("Bootstrapping"),
	showcmd(boot_build, "", [
          %1_______________________________________________
	  "Build the bootstrap (if it does not exist)"
        ]),
	showcmd(boot_rebuild, "", [
          %1_______________________________________________
          "Force rebuild of the bootstrap"
        ]),
	showcmd(boot_clean, "", [
          %1_______________________________________________
	  "Clean the bootstrap"
	]),
	show_help_sep.
show_help_group(bootstrap) --> [].

% TODO: Proposal for changes of those names
%       (remember to update the documentation)
%
%   Allow modifiers for global installation options?
%
%   user_install => local_install
%   system_install => global_install

% TODO: fullinstall and system-install are the same

show_help_group(quickstart) -->
	show_help_section("Quick start commands"),
	showcmd(fullinstall, "[<opts>]", [
          %1_______________________________________________
          "Configure, build, and install"
        ]),
	showcmd(user_install, "[<opts>]", [
          %1_______________________________________________
          "Full install locally (in your home directory)"
        ]),
	showcmd(system_install, "[<opts>]", [
          %1_______________________________________________
          "Full install globally (system wide)"
        ]),
	show_help_sep,
        showcmdtext([
          %2........................________________________________________________
          "See 'configure' for accepted configuration options."
        ]),
	show_help_sep,
	( { extrahelp } ->
	    show_help_section("Quick start commands (advanced)"),
	    showcmd(user_install_paranoid, "[<opts>]", [
	      %1_______________________________________________
	      "Like 'user-install' but enable static checking",
	      "of some possible code defects (like unused",
	      "predicates, modules, etc.)."
	    ]),
	    show_help_sep
	; []
	).

show_help_group(promotion) --> { extrahelp }, !,
	show_help_section("Bootstrap promotion (for system developers)"),
	showcmd(boot_promote, "", [
          %1_______________________________________________
          "Promote current the compiler (bytecode) and ",
          "engine (auto-generated C files) for bootstrap",
	  "",
	  "NOTE: This is a dangerous operation.",
          "      Use 'build engine' and 'build ciaoc'",
	  "      (or a whole 'build' process) to prepare",
          "      the current compiler and engine C files."
        ]),
	show_help_sep.
show_help_group(promotion) --> [].

show_help_group(configure) -->
	% TODO: add BUNDLE as a parameter
	% TODO: configure help is also in bundle_configure.pl
	show_help_section("Configuration (before build)"),
	showcmd(configure, "[<bundle>] [<args>] [<opts>]", [
          %1_______________________________________________
	  "Configure using the specified options",
	  "(discarding any previous selection)"
        ]),
	showcmdtext([
          %2........................________________________________________________
          "",
	  "The additional arguments for configure are accepted:",
	  "",
	  "  --interactive           Configure the system interactively",
	  "                          (existing previous configuration is consulted)",
	  "  --list-options          List the available configuration options",
	  "  --describe-option <opt> Describe the configuration option"
        ]),
	show_help_sep,
        % TODO: really use BUNDLE as a parameter
	show_help_group(configure_extra).

show_help_group(configure_extra) --> { extrahelp }, !,
	showcmd(rescan_bundles, "[<bundle>]", [
          %1_______________________________________________
          "Rescan the (sub)bundles (when source changes)"
        ]),
	showcmd(configclean, "[<bundle>]", [
          %1_______________________________________________
          "Clean the configuration"
        ]),
	show_help_sep.
show_help_group(configure_extra) --> [].

show_help_group(bundle_actions) -->
	show_help_section("Build"),
	showcmd(build, "[<bundle>]", [
          %1_______________________________________________
	  "Build the specified bundle"
	]),
	show_help_sep,
	( { extrahelp } ->
	    showcmdtext([
              %2........................________________________________________________
    	      "Additionally, 'build_nodocs' omits documentation, while",
	      "'build_docs' builds just the documentation.",
	      "",
	      "In addition to all bundles, other valid targets are:",
	      % TODO: this should not be needed
	      "",
              "  extra                  All bundles except 'ciao'"
	    ]),
	    show_help_sep
	; []
	),
	( { extrahelp } ->
	    showcmdtext([	    
              %2........................________________________________________________
	      "For convenience, the following are subtargets of 'ciao' are also",
	      "available:",
	      "",
	      % TODO: Do not enumerate bundles; add an option to show bundles and subbundles!
              "  engine                 The engine (build is incremental)",
              "  ciaoc                  The 'ciaoc' compiler",
	      "  shell                  The 'ciaosh' toplevel (or shell)",
              "  emacs_mode             The Emacs based IDE for Ciao",
              "",
              "  libraries              [not documented]",
              "",
	      "  profiler               (already part of 'ciao')"
            ]),
	    show_help_sep
	; []
	),
	( { extrahelp } ->
	    showcmd(build_nolibs, "", [
              %1_______________________________________________
              "Like 'build_nodocs' (excludes libraries)"
            ]),
	    showcmd(build_platdep, "", [
              %1_______________________________________________
              "Like 'build_nodocs' (only platdep targets)"
            ]),
	    show_help_sep
	; []
	),
	( { extrahelp } ->
	    showcmd(build_docs_readmes, "[<bundle>]", [
              %1_______________________________________________
	      "Generate documentation (only READMEs)"
            ]),
	    show_help_sep
	; []
	),
	%
	show_help_section("Installation"),
	showcmd(install, "[<bundle>]", [
          %1_______________________________________________
          "Install libraries, binaries, and documentation"
        ]),
	showcmd(uninstall, "[<bundle>]", [
          %1_______________________________________________
          "Uninstall"
        ]),
	show_help_sep,
        ( { extrahelp } ->
	    showcmdtext([
              %2........................________________________________________________
	      "Install and uninstall accepts the options:",
	      "",
	      "  --destdir=DESTDIR      Prepend DESTDIR to each (un)install target",
	      "                         (used for packaged bundle creation)"
	    ])
	; []
	),
	% showcmd(register, "[<bundle>]", [
        %   %1_______________________________________________
        %   "Register in the system (locally or globally)"
        % ]),
	% showcmd(unregister, "[<bundle>]", [
        %   %1_______________________________________________
        %   "Unregister from the system"
        % ]),
	show_help_section("Cleaning"),
	showcmd(clean, "[<bundle>]", [
          %1_______________________________________________
          "Clean intermediate files (keeps configuration)"
        ]),
	showcmd(distclean, "[<bundle>]", [
          %1_______________________________________________
	  "Like 'clean' but also removes configuration"
        ]),
	showcmd(realclean, "", [
          %1_______________________________________________
          "Like 'distclean' but also clean the bootstrap"
        ]),
	show_help_sep,
	( { extrahelp } ->
	    showcmdtext([
              %2........................________________________________________________
	      "In addition to all bundles, other available targets for 'clean' are:",
	      "",
	      % TODO: Do not enumerate bundles; add an option to show bundles and subbundles!
	      "  ciao",
	      "    engine",
	      "    emacs_mode",
	      "    java",
	      "    timingmodel",
	      "    profiler (already part of 'ciao')",
	      "  ciaopp",
	      "    ilciao"
            ]),
	    show_help_sep
	; []
	),
	% NOTE: clean_docs is NOT equivalent to 'clean' for docs; it just
	% remove the final targets, not the temporary files.
	( { extrahelp } ->
	    showcmd(clean_nodocs, "[<bundle>]", [
              %1_______________________________________________
              "Like 'clean', but keep documentation targets"
            ]),
	    showcmd(clean_docs, "[<bundle>]", [
              %1_______________________________________________
              "Delete documentation targets (temporary files are",
	      "removed in 'clean_nodocs')"
            ]),
	    showcmd(clean_docs_readmes, "[<bundle>]", [
              %1_______________________________________________
              "Delete documentation (README) targets"
            ]),
	    showcmd(clean_docs_manuals, "[<bundle>]", [
              %1_______________________________________________
	      "Delete documentation (manuals) targets"
            ]),
	    show_help_sep
	; []
	),
	( { extrahelp } ->
	    showcmd(clean_platdep, "[<bundle>]", [
              %1_______________________________________________
              "Like 'clean_nodocs' but limited to architecture",
	      "and platform dependent output files"
            ]),
	    show_help_sep
	; []
	),
	( { extrahelp } ->
	    showcmd(emergency_clean, "", [
              %1_______________________________________________
              "Emergency clean (for unstable source trees)"
            ]),
	    show_help_sep,
	    showcmdtext([
              %2........................________________________________________________
	      "NOTE: Use this option when the bootstrap is unable to compile properly and",
	      "no cleaning action work."
            ]),
	    show_help_sep
	; []
	),
	show_help_group(bundle_actions_extra).

show_help_group(bundle_actions_extra) --> { extrahelp }, !,
	show_help_section("Benchmarking and Testing"),
	showcmd(runbenchmarks, "[<bundle>]", [
          %1_______________________________________________
          "Execute all benchmarks available in the system"
        ]),
	showcmd(runtests, "[<bundle>]", [
          %1_______________________________________________
          "Execute all tests (regression, etc.)"
        ]),
	show_help_sep,
	%
	show_help_section("Querying Bundles"),
	showcmd(show_bundles, "", [
          %1_______________________________________________
          "Show all available bundles"
        ]),
	show_help_sep,
	showcmd(gen_bundle_commit_info, "", [
          %1_______________________________________________
          "[Not documented, internal]"
        ]),
	show_help_sep,
	%
	show_help_section("Generation of Packaged Bundles (pbundle)"),
	showcmd(gen_pbundle__TYPE, "", [
          %1_______________________________________________
          "Generates a packaged bundle of the specified",
	  "TYPE (see gen_pbundle_hook/3)"
        ]),
	showcmd(gen_pbundle__descfile, "", [
          %1_______________________________________________
          "Generate the pbundle_meta file"
        ]),
	show_help_sep.
show_help_group(bundle_actions_extra) --> [].

show_help_group(ciaopp) --> { extrahelp }, !,
	show_help_section("CiaoPP"),
	showcmd(calibrateciaopp, "", [
          %1_______________________________________________
          "Calibrate for resource analysis (for 'ciaopp')"
        ]),
	showcmd(save_calibrateciaopp, "", [
          %1_______________________________________________
          "Save the current calibration results (for 'ciaopp')"
        ]),
	show_help_sep,
	show_help_section("Timingmodel (a simple WAM for time analysis)"),
	showcmd(gen_timingmodel, "", [
          %1_______________________________________________
          "Benchmark and estimate timing model for",
	  "'miniprolog'"
        ]),
	show_help_sep.
show_help_group(ciaopp) --> [].

show_help_group(emacs_mode) --> { extrahelp }, !,
	show_help_section("Emacs mode"),
	showcmd(create_tags, "[<bundle>]", [
          %1_______________________________________________
          "Build the TAGS file (emacs) for the given bundle"
        ]),
	showcmd(delete_tags, "[<bundle>]", [
          %1_______________________________________________
          "Delete the TAGS file (emacs)"
        ]),
	show_help_sep.
show_help_group(emacs_mode) --> [].

% ---------------------------------------------------------------------------
:- doc(subsection, "Auxiliary for help").

:- use_module(library(lists), [length/2]).

do__help(Cmd) :-
	set_extrahelp(no),
	show_help(Cmd, Str, []),
	write_string(Str).

do__help_extra(Cmd) :-
	set_extrahelp(yes),
	show_help(Cmd, Str, []),
	write_string(Str).

dashify(X, Y) :-
	atom_codes(X, Xs),
	dashify_(Xs, Ys),
	atom_codes(Y, Ys).

dashify_([], []) :- !.
dashify_([0'_|Xs], [0'-|Ys]) :- !, dashify_(Xs, Ys).
dashify_([X|Xs], [X|Ys]) :- dashify_(Xs, Ys).

string([]) --> [].
string([X|Xs]) --> [X], string(Xs).

% TODO: Use ANSI colors?
%
% % # Setup ANSI color if the output is a TTY and the terminal does support them
% % 
% % #if tty -s <&1 && ( [ x"${TERM}" = x"xterm" ] || [ x"${EMACS}" = x"t" ] ) ; then
% % if tty -s <&1 && [ x"${TERM}" = x"xterm" ] ; then
% %     BOLD_B="\033[1m"
% %     BOLD_E="\033[0m"
% %     UNDER_B="\033[4m"
% %     UNDER_E="\033[0m"
% % else
% %     BOLD_B=""
% %     BOLD_E=""
% %     UNDER_B=""
% %     UNDER_E=""
% % fi
% % 
% % # Echo in 'bold' style
% % bold_echo() {
% %     echo "${BOLD_B}${1}${BOLD_E}"
% % }
% % # Echo in 'under' style
% % under_echo() {
% %     echo "${UNDER_B}${1}${UNDER_E}"
% % }

bold_echo(X) --> string(X).

under_echo(X) --> string(X).

blanks(N) --> { N =< 0 }, !, [].
blanks(N) --> " ", { N1 is N - 1 }, blanks(N1).

marginsize(2).
col1size(27).

% Command-line name of the action (replace '_' by '-')
showcmd(Action, Extra, Desc) -->
	% Margin
	{ marginsize(Margin) },
	blanks(Margin),
	% Command name
	{ dashify(Action, Action1) },
	{ atom_codes(Action1, Action2) },
	bold_echo(Action2),
	% Command arguments (optional)
	{ Extra = "" -> Extra2 = ""
	; Extra2 = " "||Extra
	},
	string(Extra2),
	% Compute size for alignment
	{ length(Action2, La) },
	{ length(Extra2, Le) },
	{ L is La + Le + Margin },
	{ col1size(Col) },
	{ L < Col -> Pad is Col - L ; Pad = 0 },
	( { Pad = 0 } -> "\n", blanks(Col)
	; blanks(Pad)
	),
	aligntext(Desc, Col).

% Additional description of a command
showcmdtext(Text) -->
	% Margin (twice)
	{ marginsize(Margin), Margin2 is 2 * Margin },
	blanks(Margin2),
	aligntext(Text, Margin2).

% Display an aligned text. The text is given as a list of lists.
% (assume that the alignment padding for the first line is already
% printed)
aligntext([Row|Rows], Col) --> !,
	string(Row), "\n",
	( { Rows = [] } -> []
	; % More rows, align and print
	  blanks(Col),
	  aligntext(Rows, Col)
	).
aligntext([], _Col) --> [].

show_help_section(Text) -->
	under_echo(Text),
	":\n",
	"\n".

show_help_sep -->
	"\n".

:- data extrahelp_/0.
set_extrahelp(Bool) :-
	retractall_fact(extrahelp_),
	( Bool = yes ->
	    assertz_fact(extrahelp_)
	; true
	).
extrahelp :- extrahelp_.

