:- module(_, [], [assertions, fsyntax, hiord]).

:- doc(title,  "Interface to Ciao compilation (separate process)").
:- doc(subtitle, "(compiler, doc generator, etc.)").

:- doc(author, "Jos@'{e} F. Morales").
:- doc(author, "Ciao Deveveloper Team").

% TODO: Implement batch execution
%  - The build compiler could be optim_comp
%  - Collect all compilation queries and build a plan.
%  - Make sure that .po/.itf files of bundletool (due to dynamic compilation)
%    are not mixed with the build .po/.itf files.

% TODO: Do not use a dummy file in gen_asr_file_main_rel/1 (assrt_lib
%   cannot document main modules!)

:- use_module(library(lists), [append/3, length/2]).
:- use_module(library(aggregates), [findall/3]).
%
:- use_module(library(system), [cd/1, working_directory/2, file_exists/1]).
:- use_module(library(system_extra), [using_tty/0]).
%
:- use_module(library(pathnames), [path_concat/3, path_split/3]).
:- use_module(library(bundletool/paths_extra), [fsR/2]).
%
:- use_module(library(bundletool/messages_aux),
	[bold_message/2, normal_message/2]).
:- use_module(library(messages), [show_message/3]).
%
:- use_module(library(bundletool/bundle_aux),
	[b_filebuild/4,
	 b_link/3,
	 n_and_props/3,
	 n_output/3,
	 n_name/2]).
:- use_module(library(bundletool/bundle_aux), [root_bundle_source_dir/1]).

% ===========================================================================

% TODO: duplicated in config-and-dump.sh
:- export(ciaoc/1). % (EXPORTED)
ciaoc := ~b_filebuild(plexe, ciao, 'ciaoc').

% TODO: duplicated in config-and-dump.sh
:- export(bootstrap_ciaoc/1). % (EXPORTED)
bootstrap_ciaoc := ~fsR(bundle_src(ciao)/'bootstrap'/'ciaoc.sta').

% ===========================================================================

:- use_module(library(process), [process_call/3]).
:- use_module(library(bundletool/config_common), [
	localciao_env/2, command_options/1, stop_if_error/1]).

:- use_module(library(bundletool/bundle_configure), [
    set_prolog_flag_from_bundle_config/1
]).

:- export(invoke_lpdoc/2).
invoke_lpdoc(ConfigFile, Args) :-
	atom_concat('stop_if_error=', ~stop_if_error, StopIfError),
	invoke_lpdoc(['-d', StopIfError,
	              '-f', ConfigFile|Args]).

lpdoc_exec := ~fsR(concat_verk(lpdoc, plexe, builddir_bin(build)/('lpdoc'))).

% TODO: make lpdoc verbose message more descriptive, remove message
%   here (e.g., show basename of directory containing SETTINGS)
:- export(invoke_lpdoc/1).
invoke_lpdoc(Args) :-
	% TODO: Why stop_if_error is not propagated here?
	Cmd = ~lpdoc_exec,
	Dir = ~fsR(builddir_doc(build)),
	Env = ~localciao_env(build),
	bold_message("Executing `lpdoc' with arguments ~w", [Args]),
	Options = [cwd(Dir), env(Env) | ~command_options],
	process_call(Cmd, Args, Options).

ciaosh_exec := ~fsR(concat_verk(ciao, plexe, builddir_bin(build)/('ciaosh'))).

:- export(invoke_ciaosh/1).
invoke_ciaosh(Input) :-
	Env = ~localciao_env(build),
	process_call(~ciaosh_exec,
	       ['-q', '-f'],
	       [env(Env), stdin(file(Input)) | ~command_options]).

:- export(invoke_ciaosh_batch/1).
% Batch execution of queries using ciaosh and current config prolog flags
% TODO: Create a module instead? (remember clean_mod/1)
invoke_ciaosh_batch(Cmds) :-
	add_config_prolog_flags(Cmds, Cmds2),
	Env = ~localciao_env(build),
	process_call(~ciaosh_exec,
	       ['-q', '-f'],
	       [env(Env), stdin(terms(Cmds2)) | ~command_options]).

add_config_prolog_flags(Cmds, CmdsR) :-
	set_prolog_flag_from_bundle_config(SetPrologFlags),
	append(SetPrologFlags, Cmds, CmdsR).

:- export(invoke_ciaoc/1).
invoke_ciaoc(Args) :-
	Env = ~localciao_env(build),
	process_call(~ciaoc, Args, [env(Env)]).

% Build the executable for InDir/InFile, store in the build directory
% and select the current version (add a symbolic link).
%
% @var{Opts} indicate the compilation options. The options indicate
% the compiler iteration that is ued:
%
%  - 'bootstrap_ciaoc': initial bootstrap ciaoc
%  - 'final_ciaoc': (default)
%      final ciaoc compiler (fixpoint when self-compiling)
%
% Other options are:
%
%  - 'static': build a static executable (self-contained, so that
%     changes in the compiled system libraries does not affect it)
%
b_make_exec(Bundle, InDir, InFile, OutFile, Opts) :-
	b_filebuild(plexe, Bundle, OutFile, FileBuild),
	( member(static, Opts) ->
	    Static = ['-s']
	; Static = []
	),
	( member(final_ciaoc, Opts) ->
	    Dir = ~fsR(InDir),
	    In = ~fsR(concat_k(ext('.pl'), InFile)),
	    Args = ['-x' | ~append(Static, ['-o', FileBuild, In])],
	    Env = ~localciao_env(build),
	    process_call(~ciaoc, Args, [cwd(Dir), env(Env)])
	; member(bootstrap_ciaoc, Opts) ->
	    Dir = ~fsR(InDir),
	    In = ~fsR(concat_k(ext('.pl'), InFile)),
	    Args = ['-x' | ~append(Static, ['-o', FileBuild, In])],
	    Env = ~localciao_env(bootbuild),
	    process_call(~bootstrap_ciaoc, Args, [cwd(Dir), env(Env)])
	; throw(bad_opts_in_b_make_exec(Opts))
	),
	b_link(plexe, Bundle, OutFile).

% TODO: Fix using CIAOCACHEDIR. We have a problem here: we build some
%       libraries with the bootstrap compiler and others with the
%       compiler that is built into 'bootlpmake'. That may give us a
%       lot of problems.
%
%       If we manage to have .po/.itf in a separate directory, we
%       could just have a toplevel as our bootstrap (that is, a system
%       where dynamically loaded code is possible).

%%     eval ${SETLOCALCIAO} ${BOOTSTRAP_CIAOC} -s -x ${CIAOCOPTS} -o ${CIAOC} ${CIAOCNAME}

% TODO: Merge with b_make_exec (this generates the C code for ciaoc
%   using the boostrap compiler; in this branch of Ciao this code is
%   reused for all other Ciao static executables)
:- export(emugen_absmach_def/0).
emugen_absmach_def :-
	InDir = bundle_src(ciao)/engine,
        InFile = absmach_def,
        %
	Env = ~localciao_env(bootbuild),
	Dir = ~fsR(InDir),
	In = ~fsR(concat_k(ext('.pl'), InFile)),
	% TODO: emugen should be part of the bootstrap ciaoc. Since it
        %   is not yet, this command may fail. In that case, try using
        %   FAKE_EMUGEN (and do boot-rebuild).
	process_call(~bootstrap_ciaoc, ['-c', In], [cwd(Dir), env(Env)]).
%	display('***USING FAKE_EMUGEN***'), nl.

% FAKE_EMUGEN:
%   mkdir -p build/objs/DARWINi86
%   cp ciao/bootstrap/Makefile-absmach_def build/objs/DARWINi86/Makefile-absmach_def
%   cp ciao/bootstrap/instrdefs.h build/objs/DARWINi86/instrdefs.h 
%   cp ciao/bootstrap/wamloop.c build/objs/DARWINi86/wamloop.c
%   cp ciao/bootstrap/absmachdef.h build/objs/DARWINi86/absmachdef.h

% TODO: build_standalone_list/3 does not receive the kind of utility (K)
%       It can only build Prolog applications. Add plug-ins? Or rewrite
%       the special applications in some that that 'b_make_exec' understand?
%       (see 'Use packages to generate script files' in my TODO list)

% TODO: show the same kind of messages that are used when compiling libraries
:- export(build_standalone_list/3).
build_standalone_list(Bundle, Dir, Ps) :-
	( % (failure-driven loop)
	  member(P0, Ps),
	    build_standalone(Bundle, Dir, P0),
	    fail
	; true
	).

build_standalone(Bundle, Dir, P0) :-
	n_and_props(P0, P, Props),
	n_name(Props, Name),
	n_output(P, Props, Output),
	( member(static, Props) ->
	    % make static executable
	    Opts = [static]
	; Opts = []
	),
	( member(shscript, Props) ->
	    true % TODO: invoke custom build predicate
	; member(bootstrap_ciaoc, Props) ->
	    % use bootstrap ciaoc
	    % TODO: it should use 'build' configuration
	    normal_message("Building '~w' (~s) using bootstrap compiler", [Output, Name]),
	    b_make_exec(Bundle, Dir, P, Output, [bootstrap_ciaoc|Opts])
	; % member(final_ciaoc, Props) ->
	  % use final_ciaoc (default)
	  % TODO: it should use 'build' configuration
	  % TODO: Add options for other ciaoc iterations
	  normal_message("Building '~w' (~s)", [Output, Name]),
	  b_make_exec(Bundle, Dir, P, Output, [final_ciaoc|Opts])
	).
	    
:- export(build_mods/2).
% CompActions: list of compiler actions (gpo, gaf) 
% (IT SHOULD BE INTERNALLY EXPORTED)
build_mods(BaseDir0, CompActions) :-
	path_concat(BaseDir0, '', BaseDir),
	bold_message("Compiling ~w libraries", [BaseDir]),
	compile_modules(compilable_module, BaseDir, CompActions).

compile_modules(Filter, BaseDir, CompActions) :-
	findall(m(Dir, File, FileName),
	  compilable_modules(Filter, BaseDir, Dir, File, FileName), ModulesU),
	sort(ModulesU, Modules),
	show_duplicates(Modules),
	compile_module_list(Modules, BaseDir, CompActions).

% TODO: (put a list of files instead?)
compilable_modules(platdep_compilable_module, _BaseDir, Dir, File, FileName) :- !,
	% TODO: BaseDir ignored
	platdep_module(Dir, File, FileName),
	file_exists(FileName).
compilable_modules(Filter, BaseDir, Dir, File, FileName) :-
	current_file_find(Filter, BaseDir, FileName),
	path_split(FileName, Dir, File).

show_duplicates(Modules) :-
	sort_by_file(Modules, Modules2),
	dump_duplicates(Modules2).

% TODO: This can be improved!
dump_duplicates(Modules) :-
	dump_duplicates_(Modules, [_|Modules]).

dump_duplicates_([], _).
dump_duplicates_([m(PrevFile, _, PrevFileName)|PrevModules], [m(File, _, FileName)|Modules]) :-
	( File == PrevFile ->
	    show_message(error, "Module ~w already defined in ~w", [FileName, PrevFileName])
	; true
	),
	dump_duplicates_(PrevModules, Modules).

sort_by_file(Modules, Modules2) :-
	sort_by_file_(Modules, Modules1),
	sort(Modules1, Modules2).

sort_by_file_([], []).
sort_by_file_([m(Dir, File, FileName)|Modules],
	    [m(File, Dir, FileName)|Modules2]) :-
	sort_by_file_(Modules, Modules2).

:- export(compile_module_list/3).
% TODO: integrate ciaoc_batch_call.pl in ciaoc (or create another executable)
compile_module_list(Modules, BaseDir, CompActions) :-
	tty(UsingTTY),
	working_directory(OrigDir, OrigDir),
	(BaseDir == '' -> true ; cd(BaseDir)),
	invoke_ciaosh_batch([
	  use_module(library(bundletool/ciaoc_batch_call), [compile_mods/4]),
	  compile_mods(Modules, CompActions, BaseDir, UsingTTY)
	]),
	cd(OrigDir).

tty(X) :-
	( using_tty ->
	    X = using_tty
	; X = no_tty
	).

% ===========================================================================

:- use_module(library(system), [copy_file/3]).

:- export(copy_bootstrap/0).
:- pred copy_bootstrap # "Backup the bootstrap @tt{ciaoc} and replace
   it by @tt{ciaoc.sta}".

copy_bootstrap :-
        BootstrapCiaocName = 'ciaoc.sta',
	BootstrapCiaoc = ~fsR(bundle_src(ciao)/bootstrap/BootstrapCiaocName),
	Ciaoc = ~ciaoc,
	%
	backup_name(BootstrapCiaoc, BootstrapCiaocBak),
	%
        copy_file(BootstrapCiaoc, BootstrapCiaocBak, [overwrite]),
	copy_file(Ciaoc, BootstrapCiaoc, [overwrite]).

:- use_module(library(process), [process_call/3]).

% Obtain a backup name by appending the date in YmdHMS format
% TODO: implement in Prolog
backup_name(File, FileBak) :-
	process_call(path(date), ['+%Y%m%d%H%M%S'],
                     [stdout(line(DateStr)), verbose_error, status(_)]),
	atom_codes(DateAtm, DateStr),
	atom_concat([File, '-', DateAtm], FileBak).

% ===========================================================================

:- doc(section, "Collect platform specific modules").

:- data platdep_module/3.

:- use_module(library(read), [read/2]).

% TODO: This MUST be in build/, not here; compute only when needed
% File where platdep_modules/3 is stored
platdep_modules_file(F) :-
	F = ~fsR(bundle_src(ciao)/'Manifest'/'platdep_modules.pl').

% Read platdep_module/3 from file
read_platdep_mods :-
	platdep_modules_file(PlatdepModulesFile),
	open(PlatdepModulesFile, read, Stream),
	read_platdep_mods_(Stream, Xs),
	close(Stream),
	%
	retractall_fact(platdep_module(_, _, _)),
	( % (failure-driven loop)
	  member(platdep_module(Dir, File, FileName), Xs),
	    assertz_fact(platdep_module(Dir, File, FileName)),
	    fail
	; true
	).

read_platdep_mods_(Stream, Out) :-
	read(Stream, R),
	!,
	( R = end_of_file ->
	    Out = []
	; Out = [R|Rs],
	  read_platdep_mods_(Stream, Rs)
	).

:- use_module(library(file_utils), [output_to_file/2]).

%% platdep_module('contrib/difference_constraints/','difference_constraints_rt.pl','contrib/difference_constraints/difference_constraints_rt.pl').
%% platdep_module('contrib/difference_constraints/difference_constraints_tab/','difference_constraints_tab.pl','contrib/difference_constraints/difference_constraints_tab/difference_constraints_tab.pl').
%% platdep_module('contrib/hashtable/','hashtable.pl','contrib/hashtable/hashtable.pl').
%% platdep_module('contrib/hrtime/','hrtime.pl','contrib/hrtime/hrtime.pl').
%% platdep_module('contrib/profiler/','profiler_rt.pl','contrib/profiler/profiler_rt.pl').
%% platdep_module('contrib/profiler/','profiler_utils_base.pl','contrib/profiler/profiler_utils_base.pl').
%% platdep_module('contrib/profiler/','profiler_utils_native.pl','contrib/profiler/profiler_utils_native.pl').
%% platdep_module('contrib/tabling/','tabling_rt.pl','contrib/tabling/tabling_rt.pl').
%% platdep_module('library/concurrency/','concurrency.pl','library/concurrency/concurrency.pl').
%% platdep_module('library/random/','random.pl','library/random/random.pl').
%% platdep_module('library/sha1/','sha1.pl','library/sha1/sha1.pl').
%% platdep_module('library/sockets/','sockets.pl','library/sockets/sockets.pl').

% Scan platdep modules
% TODO: document why it is collected (and when)
% TODO: This file should be in other place (builddir?) (it is autogenerated)
:- export(scan_platdep_modules/1).
scan_platdep_modules(LibDirs) :-
	platdep_modules_file(F),
	output_to_file(dump_platdep_modules(LibDirs), F).

:- use_module(library(sort), [sort/2]).
:- use_module(library(write), [portray_clause/1]).

dump_platdep_modules(LibDirs) :-
	% Show the platform dependent modules
	% TODO: kludge! this collects the modules after they have been compiled
	% TODO: why portray?
	findall(Mod,
	    (
		member(LibDir, LibDirs),
		current_platdep_module(LibDir, Mod)
	    ),
	    PML0),
	sort(PML0, PML),
	list(PML, portray_clause).

:- use_module(library(source_tree), [current_file_find/3]).
:- use_module(library(terms), [atom_concat/2]).

% TODO: Implement in a completely different way: ask the compiler for
%   that information (it is in the .itf file)
current_platdep_module(BaseDir, platdep_module(Dir, File, FileName)) :-
	% TODO: coupled with the foreign interface...
	current_file_find(compilable_module, BaseDir, FileName),
	%
	path_split(FileName, Dir, File),
	%
	atom_concat(FileBase, '.pl', FileName),
	FileSO = ~fsR(concat_k(ext(~get_so_ext), ~atom_concat([FileBase, '_', ~get_platform]))), % TODO: full path?
	file_exists(FileSO).

% ---------------------------------------------------------------------------

:- use_module(engine(internals),
	[po_filename/2,
	 itf_filename/2,
	 asr_filename/2,
	 ast_filename/2]).

:- export(compile_platdep_modules/1).
compile_platdep_modules(CompActions) :-
	read_platdep_mods,
	compile_modules(platdep_compilable_module, '', CompActions).

% TODO: This is a horrible hack! where are platdep modules collected?
% TODO: The current method to detect platdep_modules is incomplete (e.g., chat_tabling.c is not detected); this should talk with the foreign interface

:- export(enum_platdep_mod/1).
enum_platdep_mod(FileName) :-
	read_platdep_mods,
	% (nondet)
	platdep_module(_,_,File),
	FileName = ~fsR(bundle_src(ciao)/File).

% Exclude precompiled files to enforce platdep compilation
% TODO: this is a hack, force recompile instead
:- export(enum_platdep_noarch_file/1).
enum_platdep_noarch_file := File :-
	platdep_module(_, _, ModulePl),
	atom_concat(Module, '.pl', ModulePl),
	Base = ~fsR(bundle_src(ciao)/Module),
	( po_filename(Base, File)
	; itf_filename(Base, File)
	; asr_filename(Base, File)
	; ast_filename(Base, File)
	).

% ===========================================================================
:- doc(section, "Miscellanea (some implemented in SH)").
% TODO: reimplement in Prolog when possible

execSH_internal(Command, Args) :-
	root_bundle_source_dir(CiaoDESrc),
	ciao_lib_dir(CiaoLibDir),
	Env = ['CIAODESRC' = CiaoDESrc,
	       'DEFAULT_CIAOLIB' = CiaoLibDir],
	% TODO: hardwired path
	atom_concat(CiaoDESrc, '/ciao/lib/compiler/scripts/builddir_trampoline.sh', InternalSH),
	process_call(InternalSH, [Command|Args], [env(Env), verbose_error, status(0)]).

% TODO: reimplement in Prolog, do not export from this module
% Delete a file
:- export(do_delete_file/1).
do_delete_file(File) :-
	execSH_internal(do_delete_file, [File]).

% TODO: reimplement in Prolog, do not export from this module
% Delete a directory (recursively)
:- export(do_remove_dir/1).
do_remove_dir(Dir) :-
	execSH_internal(do_remove_dir, [Dir]).

% TODO: reimplement in Prolog
:- export(do_exec_header_for_builddir/0).
do_exec_header_for_builddir :-
	execSH_internal(do_exec_header_for_builddir, [~fsR(builddir(build))]).

% Special targets for the engine (for CIAOBUILDDIR)
:- export(builddir_do_engine/1).
builddir_do_engine(Target) :-
	execSH_internal(builddir_do_engine, [~fsR(builddir(build)), Target]).

% Special clean targets for builddir
:- export(builddir_clean/1).
builddir_clean(Target) :-
	execSH_internal(builddir_clean, [~fsR(builddir(build)), Target]).

% Clean (compilation files in) a directory
:- export(do_clean_dir_rec/1).
do_clean_dir_rec(Dir) :-
	execSH_internal(do_clean_dir_rec, [Dir]).

% Clean compilation files for a individual module (Base is file without .pl suffix)
:- export(do_clean_mod/1).
do_clean_mod(Base) :-
	execSH_internal(do_clean_mod, [Base]).

% (temporary, for ilciao)
% TODO: Write in Prolog or use build.xml
:- export(do_delete_example_pl_files/1).
do_delete_example_pl_files(Dir) :-
	execSH_internal(do_delete_example_pl_files, [Dir]).

:- use_module(engine(internals),
	[po_filename/2,
	 itf_filename/2]).

:- export(clean_engine/0).
clean_engine :-
	% Ensure absmach_def file is recompiled on next build
	% TODO: generalize
	Base = ~fsR(bundle_src(ciao)/engine/absmach_def),
	do_delete_file(~po_filename(Base)),
	do_delete_file(~itf_filename(Base)),
	%
	builddir_do_engine(clean).

:- export(clean_platdep_mods/0).
clean_platdep_mods :-
	( % (failure-driven loop)
	  enum_platdep_mod(FileName),
	    atom_concat(NoExt, '.pl', FileName),
	    do_clean_mod(NoExt),
	    fail
	; true
	).

