:- module(config_common, [], [assertions, fsyntax]).

:- doc(title, "Common definitions based on bundle configuration values").
:- doc(author, "Jose F. Morales").
:- doc(author, "Ciao Development Team").

:- use_module(library(terms), [atom_concat/2]).

:- use_module(library(bundletool/paths_extra), [fsR/2]).
:- use_module(library(bundletool/bundle_params),
	[root_bundle_name/1,
	 bundle_param_value/3]).
:- use_module(library(bundle/bundle_config), [bundle_config_value/3]).

% ===========================================================================
% The URL and directory for our main distribution site

% (Not configurable)

:- export(home_url_str/1). % (EXPORTED)
home_url_str := "http://ciao-lang.org/".
% TODO: Wrong
:- export(packages_dir_str/1). % (EXPORTED)
packages_dir_str := "packages/master/".

% ===========================================================================

:- export(rootprefix/1). % (EXPORTED)
:- pred rootprefix(DestDir) # "@var{DestDir} is the prefix that is
   prepended to each (un)install target (useful for packaged bundle
   creation)".

% TODO: rename to install_destdir? or rename DESTDIR to rootprefix?
rootprefix(R) :-
	root_bundle_name(Root),
	( bundle_param_value(Root, 'DESTDIR', Value) ->
	    R = Value
	; R = ''
	).

% ===========================================================================

:- export(perms/1). % (EXPORTED)
% Define this to be the permissions for installed execs/dirs and data files:
perms(perms(rwX, rwX, rX)).

% ===========================================================================

:- export(stop_if_error/1). % (EXPORTED)
stop_if_error := ~bundle_config_value(ciaode, 'STOP_IF_ERROR').

:- export(with_docs/1). % (EXPORTED)
with_docs := ~bundle_config_value(ciaode, 'WITH_DOCS').

:- export(command_options/1). % (EXPORTED)
command_options := ~stop_command_options(~stop_if_error).
:- use_module(library(bundletool/bundle_aux), [stop_command_options/2]).

:- export(instype/1). % (EXPORTED)
instype := ~bundle_config_value(ciaode, 'INSTYPE').

% Prefix dir for installation (or builddir for 'local' instype).
% (e.g., if /usr/local, binaries will be installed in /usr/local/bin)
:- export(installdir_prefix/1). % (EXPORTED)
installdir_prefix := ~bundle_config_value(ciaode, 'INSTALLDIR_PREFIX').

% Directory for installation of binaries
:- export(installdir_bin/1). % (EXPORTED)
installdir_bin := ~bundle_config_value(ciaode, 'INSTALLDIR_BIN').

% TODO: Distinguish cases for build and installation
:- export(installdir_libroot/1). % (EXPORTED)
installdir_libroot := ~bundle_config_value(ciaode, 'INSTALLDIR_LIBROOT').

% Relative path (for symbolic links) to reallib_dir from lib_dir
% TODO: find a better solution
:- export(relinstalldir_lib/2). % (EXPORTED)
relinstalldir_lib(Bundle) := ~relinstalldir_lib_(~instype, Bundle).
relinstalldir_lib_(local, _Bundle) := ''.
relinstalldir_lib_(global, Bundle) := ~fsR(concat_ver(Bundle, Bundle)). % (see installdir)

:- export(installdirbase_lib/2). % (EXPORTED)
installdirbase_lib(Bundle) := ~installdirbase_lib_(~instype, Bundle).
installdirbase_lib_(local, Bundle) := ~fsR(bundle_src(Bundle)).
installdirbase_lib_(global, Bundle) := ~fsR(installdirbase(Bundle)).

:- export(get_installdir_lib/2). % (EXPORTED)
get_installdir_lib(Bundle) := ~get_installdir_lib_(~instype, Bundle).
get_installdir_lib_(local, Bundle) := ~fsR(bundle_src(Bundle)).
get_installdir_lib_(global, Bundle) := ~fsR(installdir_lib(Bundle)).

:- export(name_ciaoengine/1).
name_ciaoengine := 'ciaoengine'.
:- export(name_ciaoengine_arch/1). % (EXPORTED)
name_ciaoengine_arch := ~atom_concat([~name_ciaoengine, '.', ~get_platform]).

:- export(installdir_engine/1). % (EXPORTED)
installdir_engine := D :-
	( instype(local) ->
	    D = ~fsR(builddir(build)/objs)
	; Bundle = ciao,
	  D = ~fsR(installdir_lib(Bundle)/'engine')
	).

:- export(installdir_include/1).
installdir_include := D :-
	( instype(local) ->
	    D = ~fsR(builddir(build)/include)
	; installdir_prefix(Prefix),
	  D = ~fsR(Prefix/include)
	).

% TODO: duplicated in config-and-dump.sh, bundletool scripts
% NOTE: The problem here is to use the proper value for
%   get_debug/1 (we need the 'builddirid')
:- export(get_platformdeb/2). % (EXPORTED)
% TODO: This fix for cross-compilation is incomplete
get_platformdeb(BuildDirId) := OsArchDebug :-
	get_debug(RunningDebug),
	( RunningDebug = '-install', BuildDirId = build ->
	    % We are running 'bootbuild' and BuildDirId is different,
	    % get Debug from the configuration
	    Debug = ~get_debug_config
	; % Otherwise, get default debug
	  Debug = ~get_debug
	),
	OsArchDebug = ~atom_concat(~get_platform, Debug).

% get_debug, using current configuration settings
% (it may differ from get_debug from the running engine)
% TODO: duplicated in config-and-dump.sh
get_debug_config(Debug) :-
	debug_level(DebugLevel),
	( DebugLevel = 'nodebug' ->
	    Debug = ''
	; atom_concat('-', DebugLevel, Debug)
	).

debug_level := ~bundle_config_value(ciao, 'DEBUG_LEVEL').

local_engine(BuildDirId) := ~fsR(concat_k(exec, builddir(BuildDirId)/objs/(~get_platformdeb(BuildDirId))/(~name_ciaoengine))).
local_hdir(BuildDirId) := ~fsR(builddir(BuildDirId)/include/(~get_platformdeb(BuildDirId))).

:- export(localciao_env/2). % (EXPORTED)
:- pred localciao_env(BuildDirId, Env) # "@var{Env} is the environment
   (used by @pred{process_call/3}) that activates the Ciao build at
   @var{BuildDirId}".

localciao_env(BuildDirId) := Env :-
	Env = ['CIAOALIASPATH' = '',
	       'CIAOLIB' = ~fsR(bundle_src(ciao)),
	       'CIAOHDIR' = ~local_hdir(BuildDirId),
	       'CIAOENGINE' = ~local_engine(BuildDirId)].

% TODO: Maybe it is not the place to put this file
:- export(build_mod_options/2).
build_mod_options(Bundle, C) :- build_mod_options_(Bundle, C, [do_gpo]).

build_mod_options_(Bundle, Opts, Opts0) :-
	bundle_gen_asr(Bundle, GCA),
	asr_option(GCA, Opts, Opts0).

% Generate 'asr' files in the compilation of this bundle?
% TODO: why?
bundle_gen_asr(ciao, GCA) :- !, GCA = ~bundle_config_value(ciao, 'GEN_CIAO_ASR').
bundle_gen_asr(ciaopp, GCA) :- !, GCA = ~bundle_config_value(ciaopp, 'GEN_CIAOPP_ASR').
bundle_gen_asr(_, no).

asr_option(yes, [do_gaf|T], T).
asr_option(no, Opts, Opts).

:- export(gmake/1). % (EXPORTED)
gmake := ~atom_concat(~bundle_config_value(ciaode, 'MAKEDIR'),
	              ~bundle_config_value(ciaode, 'MAKENAME')).
:- export(gtar/1). % (EXPORTED)
gtar := ~atom_concat(~bundle_config_value(ciaode, 'TARDIR'),
	             ~bundle_config_value(ciaode, 'TARNAME')).

% ===========================================================================
% Options from LPdoc bundle (defined here since they are needed for
% installation in bundletool)

:- use_module(library(bundle/bundle_config), [bundle_config_value/3]).

% TODO: do the changes required to remove '/'
:- export(docdir/1). % (EXPORTED)
docdir := ~atom_concat(~fsR(rootprefix(~bundle_config_value(lpdoc, 'DOCDIR'))), '/').

% TODO: do the changes required to remove '/'
:- export(htmldir/1). % (EXPORTED)
htmldir := ~atom_concat(~fsR(rootprefix(~bundle_config_value(lpdoc, 'HTMLDIR'))), '/').

% TODO: do the changes required to remove '/'
:- export(mandir/1). % (EXPORTED)
mandir := ~atom_concat(~fsR(rootprefix(~bundle_config_value(lpdoc, 'MANDIR'))), '/').

% TODO: do the changes required to remove '/'
:- export(infodir/1). % (EXPORTED)
infodir := ~atom_concat(~fsR(rootprefix(~bundle_config_value(lpdoc, 'INFODIR'))), '/').

%% .bib files now in a predefined location (JFMC)
%%bibfile := ~decompose(~bundle_config_value(lpdoc, 'BIBFILES'), ',').
% Bibliography files for documentation (clip.bib and general.bib)
:- export(bibfile/1). % (EXPORTED)
bibfile := ~fsR(bundle_src(ciaode)/'doc'/'bibtex'/'clip')
	 | ~fsR(bundle_src(ciaode)/'doc'/'bibtex'/'general').

% decompose(Text0, Separator, Element) :-
%	( atom_concat([Element0, ',', Text], Text0) ->
%	    ( Element = Element0
%	    ; decompose(Text, Separator, Element)
%	    )
%       ; Element = Text0
%       ).

:- export(docformatdir/2). % (EXPORTED)
docformatdir(html, Dir) :- !, htmldir(Dir).
docformatdir(manl, Dir) :- !, mandir(Dir).
docformatdir(info, Dir) :- !, infodir(Dir).
docformatdir(_,    Dir) :- docdir(Dir).

:- export(docformat/1). % (EXPORTED)
docformat := pdf|manl|info|html. % | ps.
