:- module(pbundle_gen_win32, _, [fsyntax, assertions]).

:- doc(title,  "pbundle Generation as Windows Installer").
:- doc(author, "Edison Mera").

:- doc(module, "This module provides the definitions used to generate
   installers for the Windows platform. Currently the installer is
   based on the @em{Inno Setup Compiler} (ISCC). Thus, ISCC is required
   in order to build a graphical installer. To get the latest version
   visit @href{http://www.jrsoftware.org/isdl.php} and download the
   @em{QuickStart Pack}.").

% ===========================================================================

:- use_module(library(aggregates)).
:- use_module(library(streams)).
:- use_module(library(lists)).
:- use_module(library(llists)).
:- use_module(library(messages)).
:- use_module(library(terms), [atom_concat/2]).
:- use_module(library(file_utils)).
:- use_module(library(process), [process_call/3, process_pipe/2]).
:- use_module(library(system), [winpath/3, winpath/2]).
:- use_module(library(source_tree), [current_file_find/3]).

:- use_module(library(bundletool/config_common), [
	get_platformdeb/2,
	name_ciaoengine/1,
	command_options/1]).
:- use_module(engine(internals), ['$bundle_prop'/2]).
:- use_module(library(bundletool/paths_extra), [fsR/2]).
:- use_module(library(bundle/bundle_info), [
	enum_sub_bundles/2,
	bundle_version/2,
	bundle_version_patch/2,
	bundle_versioned_packname/2,
	bundle_manual_base/2,
	bundle_commit_info/3]).
:- use_module(library(bundletool/pbundle_gen_common)).
:- use_module(library(bundletool/bundle_aux),
	[wr_template/4, ensure_builddir/0, ensure_build_doc_dirs/0]).

% (hooks for gen_pbundle)
:- include(library(bundletool/pbundle_gen_hookdefs)).

% ===========================================================================

:- doc(iscc/1, "Specifies the path where the @em{Inno Setup Compiler} is
   installed.").

iscc := '/cygdrive/c/Program\\ Files/Inno\\ Setup\\ 5/iscc.exe'.

% TODO: Use Bundle to customize more definitions

% (hook)
gen_pbundle_hook(win32, Bundle, _Options) :- !,
	simple_message("Creating Windows installer for ~w, please be patient... ",
	    [~bundle_versioned_packname(Bundle)]),
	simple_message("Creating ISS scripts... "),
	FileIss = 'CiaoDE.iss',
	FileListName = 'file_list.iss',
	create_iss_file(Bundle, FileIss, FileListName),
	compute_file_list_iss(FileListName),
	simple_message("Compiling ISS scripts... "),
	invoke_iscc(FileIss),
	simple_message("Creation of Windows installer finished").

invoke_iscc(FileIss) :-
	process_call(~iscc, [FileIss], ~command_options),
	% TODO: should it appear before?
	create_pbundle_output_dir.

% TODO: too many definitions here are hardwired
create_iss_file(Bundle, FileIss, FileListName) :-
	OutputBaseFileName = ~atom_codes(~bundle_versioned_packname(Bundle)),
	% TODO: see PrettyCommitDesc in pbundle_download.pl
	CommitId = ~bundle_commit_info(Bundle, id),
	AppVerName = ~atom_codes(~atom_concat(['Ciao-', ~bundle_version_patch(Bundle), ' (', CommitId, ')'])),
	%
	wr_template(cwd, ~bundletool_dir/'win32', FileIss, [
	    'MyAppName' = "Ciao", % TODO: extract from bundle
	    'MyAppVerName' = AppVerName,
	    'OutputBaseFileName' = OutputBaseFileName,
	    'MyAppPublisher' = "The CLIP Laboratory", % TODO: extract from bundle
	    'LicenseFile' = ~license_file,
 	    'MyAppExeName' = ~fsR(concat_verk(ciao, plexe, 'ciaosh')), % TODO: extract from bundle
	    'CiaoVersion' = ~bundle_version(ciao), % TODO: extract from bundle
	    'SourceDir' = ~source_dir,
	    'MyRelBuildDir' = ~relciaodir(~fsR(builddir(build))),
	    'OutputDir' = ~output_dir,
	    'ManualIcons' = ~get_manual_icons(Bundle),
	    'DefaultDirName' = ~default_dir_name(Bundle),
	    'CiaoEngineExec' = ~winpath(relative, ~fsR(concat_k(exec, ~engine_bin_dir/(~name_ciaoengine)))),
	    'FileListName' = ~winpath(full, ~fsR(bundle_src(ciaode)/FileListName))
	]).

default_dir_name(Bundle) := D :-
	'$bundle_prop'(Bundle, packname(Packname)),
	D = ~atom_codes(Packname).

get_manual_icons(Bundle, S) :-
	findall(Str, get_manual_icons_(Bundle, Str), L),
	flatten(L, S).

get_manual_icons_(ParentBundle, Str) :-
	enum_sub_bundles(ParentBundle, Bundle),
	%
	DocFormat = pdf,
	RelBuildDir = ~relciaodir(~fsR(builddir(build))),
	'$bundle_prop'(Bundle, packname(PackName)),
	FileMain = ~atom_concat([~bundle_manual_base(Bundle), '.', DocFormat]),
	Str = ["Name: {group}\\" || (~atom_codes(PackName)), " Manual in PDF; ",
	       "Filename: {app}\\" || (~atom_codes(RelBuildDir)), "\\doc\\" || (~atom_codes(FileMain)), "; ",
	       "WorkingDir: {app}\\" || (~atom_codes(RelBuildDir)), "\\doc\n"].

% Compute the list of files that will be included in the ISS file
compute_file_list_iss(FileName) :-
	ensure_builddir, % TODO: needed?
 	ensure_build_doc_dirs, % TODO: needed?
	open_output(FileName, Output),
	output_file_list,
	close_output(Output).

output_file_list :-
	( % (failure-driven loop)
	  current_file(Source, DestDir),
	    display_file_entry(Source, DestDir),
	    fail
	; true
	).

% TODO: include into library(source_tree) logic
current_file(Source, DestDir) :-
	( current_file_find(distributable_precomp(bin), './', FileName),
	  % TODO: Disabled (check again if it is needed)
	  % \+ exclude_win_subdir(FileName),
	  atom_concat('./', BaseFile, FileName),
	  fullwinname(BaseFile, Source),
	  fullwinpath(BaseFile, DestDir)
	; extra_file(Source, DestDir)
	).

% TODO: why? (it was not really, './' prefix does not match)
% exclude_win_subdir('./ciao/doc').
% exclude_win_subdir('./ciao/engine').
% exclude_win_subdir('./lpdoc/doc').
% exclude_win_subdir('./ciaopp/doc').
% exclude_win_subdir('./doc').
% exclude_win_subdir('./ciao/optim_comp').

engine_bin_dir := ~relciaodir(~fsR(builddir(build)/objs/(~get_platformdeb(build)))).

% TODO: Wrong name!!
extra_file(Source, DestDir) :-
	Source = ~winpath(~extra_system_file), % (nondet)
	DestDir = ~atom_concat(~winpath(relative, ~engine_bin_dir), '\\').

each_line(Lines0, Line) :-
	( append(Line0, [0'\n|Lines], Lines0) ->
	    ( Line = Line0
	    ; each_line(Lines, Line)
	    )
        ; Line = Lines0
        ).

bin_sh('/usr/bin/sh.exe').

extra_system_file(A) :- bin_sh(A).
extra_system_file(A) :-
	% Get shared library dependencies for /usr/bin/sh.exe
	% TODO: generalize, move to library, "otool -L" in macosx
	bin_sh(BinSh),
	process_pipe([
          process_call(path(ldd), [BinSh], [status(_)]),
          process_call(path(grep), ['/usr/bin'], [status(_)]),
          process_call(path(sed), ['-e', 's:.* => \\(.*\\) (0.*:\\1:g'])
        ], [stdout(string(Lines)), verbose_error, status(_)]),
	each_line(Lines, Line),
	Line \= [],
	atom_codes(A, Line).
% TODO: hardwired, why?
extra_system_file('/usr/bin/cyggsl-0.dll').
extra_system_file('/usr/bin/cyggslcblas-0.dll').
extra_system_file('/usr/lib/lapack/cygblas-0.dll').
% TODO: hardwired, why?
extra_system_file := ~fsR(concat_k(exec, ~engine_bin_dir/(~name_ciaoengine))).
extra_system_file := ~fsR(concat_k(ext('.dll'), ~engine_bin_dir/'libciao')).

display_file_entry(Source, DestDir) :-
	display_list(['Source: ', Source, '; DestDir:{app}\\', DestDir, '\n']).

license_file := ~atom_codes(~winpath(relative, ~fsR(bundle_src(ciaode)/'LGPL'))).

source_dir := ~atom_codes(~winpath(relative, ~fsR(bundle_src(ciaode)))).
output_dir := ~atom_codes(~winpath(relative, ~pbundle_output_dir)).

fullwinname(File, WinName) :-
	winpath(relative, File, WinName).

fullwinpath(File, WinPath) :-
	extract_unix_filepath(File, Path),
	winpath(relative, Path, WinPath).

decompose_win_filename_string(FullName, Dir, Name, Extension) :-
	decompose_filename_string(FullName, 0'\\, Dir, Name, Extension).

decompose_unix_filename_string(FullName, Dir, Name, Extension) :-
	decompose_filename_string(FullName, 0'/, Dir, Name, Extension).

decompose_filename_string(FullName, PathSeparator, Dir, Name, Extension) :-
	append(Main, [PathSeparator|SubPath], FullName),
	!,
	decompose_filename2_string(Main, PathSeparator, Dir, SubPath, Name,
	    Extension).
decompose_filename_string(FullName, PathSeparator, "", Name, Extension) :-
	append(Name, [PathSeparator|Extension], FullName),
	!.
decompose_filename_string(Name, _, "", Name, "").

decompose_filename2_string(Main, PathSeparator, Dir, SubPath, Name, Extension) :-
	decompose_filename_string(SubPath, PathSeparator, Dir2, Name, Extension),
	append(Main, [PathSeparator|Dir2], Dir).

extract_win_filepath(FullName, Path) :-
	extract_filepath(FullName, 0'\\, Path).

extract_unix_filepath(FullName, Path) :-
	extract_filepath(FullName, 0'/, Path).

extract_filepath(FullName, PathSeparator, Path) :-
	decompose_filename_string(~atom_codes(FullName), PathSeparator, PathC, _, _),
	atom_codes(Path, PathC).

% ===========================================================================

% % TODO: this target is probably not necessary
% gen_pbundle__win32_test <- :-
% 	ciaode_iss('CiaoDE_test.iss', 'file_list_test.iss',
% 	    ~atom_codes(~bundle_versioned_packname(~enum_bundle(whole)))),
% 	file_list_test_iss('file_list_test.iss'),
% 	gen_pbundle__win32_test.

% TODO: targets that are not necessary
% 'CiaoDE.iss' <- ['Manifest/CiaoDE.iss.skel'] :: FileName :-
% 	OutputBaseFileName = ~atom_codes(~bundle_versioned_packname(~enum_bundle(whole))),
% 	ciaode_iss(FileName, 'file_list.iss', OutputBaseFileName).
% 
% 'CiaoDE_test.iss' <- ['Manifest/CiaoDE.iss.skel'] :: FileName :-
% 	OutputBaseFileName = ~append(~atom_codes(~bundle_versioned_packname(~enum_bundle(whole))),
% 	    "-test"),
% 	ciaode_iss(FileName, 'file_list_test.iss', OutputBaseFileName).
% 
% 'file_list.iss' <- [] :: FileName :-
% 	compute_file_list_iss(FileName).
% 
% 'file_list_test.iss' <- [] :: FileName :-
% 	file_list_test_iss(FileName).

% 'ciao/Win32/wsetup.cpx' <- ['ciao/Win32/wsetup.pl'] :-
% 	make_exec(['ciao/Win32/wsetup.pl'], 'ciao/Win32/wsetup.cpx').

% % TODO: this target is probably not necessary
% gen_pbundle__win32_test :-
% 	gen_pbundle__win32_('CiaoDE_test.iss').

% file_list_test_iss(FileName) :-
% 	file_list_type_iss(test, FileName).
